#!/usr/bin/perl
#========================================================================
# orantstat
#========================================================================
# Env:
#  - ORACLE_HOME and ORACLE_SID must be set
#  - PATH must include ORACLE_HOME/bin
#
# Usage:
#  - Run 'orantstat --' to view usage table
#
#
# Modification History: (Change version number above)
#
#  Ver  Date    Programmer   Desc
#  ==== ======= ============ ============================================
#  1.32 27Nov00 Dean Riggio  Synched up with Unix version. Functionality
#                            matches orastat v1.32
#  1.35 25Jul01 Dean Riggio  Synched up with unix version (through -iv).
#  1.36 26Jul01 Dean Riggio  Synched up with unix version (through -vs).
#  1.37 30Jul01 Dean Riggio  Synched up with unix version 1.45
#  1.38 15Oct01 Dean Riggio  Eliminated all calls that depend on the NT
#                            Resource Kit.
#  1.30 04Dec01 Bill Border  Fixed -m and added -ma
#  2.16 18Dec01 Craig Geier  Converted to use mostly ODBC. Synched with
#                            unix version 2.16
#  2.29 07Mar02 Craig Geier  Synched with unix version 2.29 except for -az 
#  2.30 16Jun04 Larry Rush   Restructured Code And Synced Up With Unix
#                            Version V2.74 And Implemented A Task Table
#                            To Drive Tasks
#  2.31 28Sep04 Larry Rush   Added code to support some servers in Germany
#                            that do not fit are standard configs in the
#                            registry...  Hopefully, this fix will not 
#                            break things
#  2.32 29Sep04 Larry Rush   Added -rg
#  2.33 22Nov04 Larry Rush   Added more logic to determine real alert log filename
#  2.34 10Jan05 Larry Rush   Modified -cb routine and added code to make sure
#                            $root_fs\\temp exists 
#  2.35 21Jan05 Larry Rush   Removed the Oracle version criteria for -hd and -ht
#  2.36 02Feb05 Larry Rush   Had to add logic to accept a backslash at end of the path
#                            Also added the getSIDS subroutine to determine $oh and $os
#  2.37 02Jun05 Larry Rush   Had to double escape the backslash path in $os
#  2.38 13Oct05 Larry Rush   Synched up with Unix version.
#  2.39 28Nov05 Larry Rush   28Nov05 Bill Border  Rewrote -tt - It is now ACCURATE (it used v$sort_usage)
#========================================================================

#use diagnostics;
use Cwd;                                                          # V02.34
use Time::Local;                                                  # V02.38

BEGIN { eval "use Win32::Registry"; eval "use Win32::Service"; };

#------------------------------------------------------------------------
#                            G L O B A L S       
#------------------------------------------------------------------------
$version        = "2.39";
$P              = "orantstat";
$|              = 1; 
$pid            = $$;
$arg            = $ARGV[0] if defined $ARGV[0];
$arg2           = $ARGV[1] if defined $ARGV[1];
$arg3           = $ARGV[2] if defined $ARGV[2];
$system_root    = $ENV{"SYSTEMROOT"};
$root_fs        = substr($system_root,0,2);
$tmp_dir        = "$root_fs\\temp";
$sqlin          = "$tmp_dir\\orantstat_sqlin_${pid}.tmp";
$tmpfile        = "$tmp_dir\\orantstat_tmpfile_${pid}.tmp";
$default_editor = "notepad";
$os             = ( exists $ENV{ORACLE_SID} ) ? $ENV{ORACLE_SID} : undef;
$os_scmd        = "";
$os_conn        = "";
@oraSQLout      = ();
$oraCmdrc       = "";                   # V02.34

## Used To Call Specific Task And Used By Usage To List Current Tasks Available
#
%TaskTbl = (
 "--"            => { TASK => \&Usage,             DESC => "View This Help File" },
 "-"             => { TASK => \&dash_task,         DESC => "Check for PMON Running, Show Oracle Version, Instance Status" },
 "-ab"           => { TASK => \&ab_task,           DESC => "Archive Log RMAN Backups [Option Archive Log Sequence Number]" },
 "-ad"           => { TASK => \&ad_task,           DESC => 'List Archive Destinations (v$archive_dest)' },
 "-af"           => { TASK => \&af_task,           DESC => "List Archive Log Directory Contents" },
 "-al"           => { TASK => \&al_task,           DESC => "Show All Archived Logs" },
 "-an"           => { TASK => \&an_task,           DESC => "Analyze Table COMPUTE STATISTICS - '$P -an TABLE-OWNER.TABLE-NAME'" },
 "-ar"           => { TASK => \&ar_task,           DESC => "Show Current DB Activity Rate" },
 "-au"           => { TASK => \&au_task,           DESC => "Show DB Audit Status" },
 "-av"           => { TASK => \&av_task,           DESC => "Archive Log Volume - [D/H] [# of days]" },
 "-az"           => { TASK => \&az_task,           DESC => "Show Current DB Activity Rate - Log to /opt/oracle/adm/cronlog/db_activity_rate_<SID>.txt" },
 "-ba"           => { TASK => \&ba_task,           DESC => "List Contents of dbamon.backup_age Table" },
 "-bb"           => { TASK => \&bb_task,           DESC => "List Summary of all RMAN backup from v\$backup_piece" },
 "-bc"           => { TASK => \&bc_task,           DESC => "List Contents of buffer pool" },
 "-bd"           => { TASK => \&bd_task,           DESC => "Datafiles Backup Mode Status" },
 "-bi"           => { TASK => \&bi_task,           DESC => "List RMAN backup inventory -- [ALL|SINCE|BETWEEN]" },
 "-bm"           => { TASK => \&bm_task,           DESC => 'List Tablespaces and Datafiles in Backup Mode (from v$backup)' },
 "-bo"           => { TASK => \&bo_task,           DESC => 'Buffer Cache Overview (Many Reports)' },
 "-bp"           => { TASK => \&bp_task,           DESC => "List running RMAN backup sessions (if any)" },
 "-br"           => { TASK => \&br_task,           DESC => "Backups: List all media written to since the last RMAN LVL0 backup" },
 "-bs"           => { TASK => \&bs_task,           DESC => "Create DDL to take datafiles that are in backup mode out of backup mode" },
 "-c"            => { TASK => \&c_task,            DESC => "Configuration: Cat (type) init.ora File" }, 
 "-cb"           => { TASK => \&cb_task,           DESC => "Controlfile Backup: Trace and Binary" },
 "-ce"           => { TASK => \&ce_task,           DESC => "Configuration: Edit 'vi' init.ora File" },
 "-cf"           => { TASK => \&cf_task,           DESC => "List Control Files" },
 "-ck"           => { TASK => \&ck_task,           DESC => 'List Time of Last Checkpoint (from v$datafile_headers)' },
 "-cm"           => { TASK => \&cm_task,           DESC => 'Configuration: View Contents of v$parameter with modifiable flags' },
 "-cp"           => { TASK => \&cp_task,           DESC => 'Configuration: View Contents of v$parameter' },
 "-ct"           => { TASK => \&ct_task,           DESC => "Coalesce TEMP tablespace" },
 "-cv"           => { TASK => \&cv_task,           DESC => "Generate syntax to recompile INVALID Objects" },
 "-dc"           => { TASK => \&dc_task,           DESC => "List Datafiles in 'cp' commands to copy all datafiles elsewhere" },
 "-dd"           => { TASK => \&dd_task,           DESC => "Reproduce Object DDL (9i+ Only)\n"                  .
                                                           "           Table:      $P -dd TABLE SCOTT DEPT\n"       .
                                                           "           Index:      $P -dd INDEX SCOTT DEPT_X\n"     .
                                                           "           Tablespace: $P -dd TABLESPACE \"\" TSNAME\n" .
                                                           "           User:       $P -dd USER \"\" SCOTT\n"        .
                                                           "           DBLink:     $P -dd USER \"\" SCOTT" },
 "-de"           => { TASK => \&de_task,           DESC => "List All Datafiles, Online Redo Logs and Control Files (For Destroying a DB)" },
 "-df"           => { TASK => \&df_task,           DESC => "List Datafiles - Optional FSCHECK 2nd parm to bdf filesystem" },
 "-df2"          => { TASK => \&df2_task,          DESC => "List Datafiles By Filesize Descending" },
 "-df3"          => { TASK => \&df3_task,          DESC => "List Datafiles - from v\$datafile" },
 "-dh"           => { TASK => \&dh_task,           DESC => "List Datafiles and their 'Used Blocks' High-Water-Mark - Creates RESIZE commands" },
 "-dn"           => { TASK => \&dn_task,           DESC => "A Better and Faster version of -dh" },
 "-dp"           => { TASK => \&dp_task,           DESC => "Database properties (from DATABASE_PROPERTIES)" },
 "-du"           => { TASK => \&du_task,           DESC => "List Datafiles where UNRECOVERABLE_CHANGE# is not null" },
 "-ec"           => { TASK => \&ec_task,           DESC => "Configuration: Edit 'vi' config.ora File" },
 "-er"           => { TASK => \&er_task,           DESC => "Display contents of DBA_ERRORS" },
 "-ev"           => { TASK => \&ev_task,           DESC => 'Select from v$system_event' },
 "-ex"           => { TASK => \&ex_task,           DESC => "Select from plan_table" },
 "-fm"           => { TASK => \&fm_task,           DESC => "Display hours since file modification time - Requires ARG" },
 "-fs"           => { TASK => \&fs_task,           DESC => "Free Space in Each Datafile" },
 "-ft"           => { TASK => \&ft_task,           DESC => "List 'Fast Start' Transactions being rolled back" },
 "-hd"           => { TASK => \&hd_task,           DESC => "Hot Tables - With Datafile Name" },
 "-ht"           => { TASK => \&ht_task,           DESC => "Hot Tables" },
 "-in"           => { TASK => \&in_task,           DESC => "List Indexes" },
 "-iv"           => { TASK => \&iv_task,           DESC => "List INVALID Objects" },
 "-l"            => { TASK => \&l_task,            DESC => "List Archive Log Status" },
 "-la"           => { TASK => \&la_task,           DESC => 'Latch Details (from v$latch)' },
 "-lf"           => { TASK => \&lf_or_rl_task,     DESC => "List Redo Log Files" },
 "-lg"           => { TASK => \&lg_task,           DESC => "Longops: Long Operation Progress" },
 "-lh"           => { TASK => \&lh_task,           DESC => "Lock Holders/Waiter" },
 "-li"           => { TASK => \&li_task,           DESC => 'List Resource Limits (v$resource_limit)' },
 "-lk"           => { TASK => \&lk_task,           DESC => "List Current TX (Non-Row) Locks" },
 "-ln"           => { TASK => \&ln_task,           DESC => "List DB Links" },
 "-lo"           => { TASK => \&lo_task,           DESC => "List Current Table Locks" },
 "-ls"           => { TASK => \&ls_task,           DESC => "Listener Status" },
 "-lv"           => { TASK => \&lv_task,           DESC => "List LVOL's / Usage" },
 "-lw"           => { TASK => \&lw_task,           DESC => "List LVOL's / Usage - Create ln -s Commands" },
 "-m"            => { TASK => \&m_or_ma_task,      DESC => "View Last 20 Lines Of Alert Log" },
 "-ma"           => { TASK => \&m_or_ma_task,      DESC => "Cat entire Alert Log" },
 "-mf"           => { TASK => \&mf_task,           DESC => "Tail -90f Alert Log" },
 "-mr"           => { TASK => \&mr_task,           DESC => "Alert log - reformatted\n"
                                                         . "           Last 20 lines   : $P -mr (DEFAULT)\n"
                                                         . "           Entire file     : $P -mr a\n"
                                                         . "           'vi' entire file: $P -mr v" },
 "-mt"           => { TASK => \&mt_task,           DESC => "MTS Overview" },
 "-mv"           => { TASK => \&mv_task,           DESC => "'vi' (read-only mode) Of Alert Log" },
 "-nf"           => { TASK => \&nf_task,           DESC => "List Objects whose NEXT EXTENT will not fit in the tablespace" },
 "-nl"           => { TASK => \&nl_task,           DESC => "List NOLOGGING Tables and Indices" },
 "-ob"           => { TASK => \&ob_task,           DESC => "Objects - With Decimal and Hex Object ID" },
 "-op"           => { TASK => \&op_task,           DESC => 'OPS: View V$PING - Lock Conversions' },
 "-pb"           => { TASK => \&pb_task,           DESC => "Performance: Current Data Block Buffer Hit Ratio - RIGHT NOW" },
 "-pc"           => { TASK => \&pc_task,           DESC => "Password Check: Check for accounts where username=password" },
 "-pd"           => { TASK => \&pd_task,           DESC => "Performance: View Data Block Buffer Hit Ratio" },
 "-pf"           => { TASK => \&pf_task,           DESC => "Performance: View Total Cumulative Free List Waits" },
 "-ph"           => { TASK => \&ph_task,           DESC => "Performance: Hot Blocks - Block With Latch Sleeps" },
 "-pi"           => { TASK => \&pi_task,           DESC => "Performance: View Histogram of Datafile I/O" },
 "-pj"           => { TASK => \&pj_task,           DESC => "Performance: I/O Distribution By Filesystem" },
 "-pk"           => { TASK => \&pk_task,           DESC => "Performance: I/O Distribution By Tablespace" },
 "-pp"           => { TASK => \&pp_task,           DESC => "Performance: I/O Service Times" },
 "-ps"           => { TASK => \&ps_task,           DESC => "OPS (Parallel Server) Status" },
 "-pw"           => { TASK => \&pw_task,           DESC => "Performance: Show segment names for tables with buffer waits" },
 "-qs"           => { TASK => \&qs_task,           DESC => "SQL Statment Count" },
 "-ra"           => { TASK => \&ra_task,           DESC => "List Rollback Segment Activity" },
 "-rb"           => { TASK => \&rb_task,           DESC => "List Rollback Segments" },
 "-rc"           => { TASK => \&rc_or_rd_task,     DESC => "List And SHRINK All Rollback Segments" },
 "-rd"           => { TASK => \&rc_or_rd_task,     DESC => "List And ALTER MAXEXTENTS UNLIMITED All Rollback Segments" },
 "-rf"           => { TASK => \&rf_task,           DESC => 'View Timestamps from V$RECOVER_FILE' },
 "-rg"           => { TASK => \&rg_task,           DESC => 'Contents of DBA_REGISTRY - DB Component Status' },
 "-rh"           => { TASK => \&rh_task,           DESC => "List REDO Logs - History" },
 "-rl"           => { TASK => \&lf_or_rl_task,     DESC => "List REDO Logs - Files" },
 "-ro"           => { TASK => \&ro_task,           DESC => "List Roles" },
 "-rp"           => { TASK => \&rp_task,           DESC => "RMAN Long Operation Progress" },
 "-rr"           => { TASK => \&rr_task,           DESC => "REDO Log Recovery Report - CURRENT Online Redo Log Info" },
 "-rs"           => { TASK => \&rs_task,           DESC => "List REDO Logs - Status" },
 "-ru"           => { TASK => \&ru_task,           DESC => "Rollback Usage - By Session" },
 "-sa"           => { TASK => \&sa_task,           DESC => "Session Statistics - All Sessions" },
 "-sb"           => { TASK => \&sb_task,           DESC => "Standby DB - Show log gaps" },
 "-sc"           => { TASK => \&sc_task,           DESC => "Sessions - By Session CPU Time" }, 
 "-sd"           => { TASK => \&sd_or_se_task,     DESC => "Sessions - Details - Sessions, Running SQL and Waits" },
 "-se"           => { TASK => \&sd_or_se_task,     DESC => "Sessions" },
 "-sg"           => { TASK => \&sg_task,           DESC => "List SGA Usage" },
 "-sh"           => { TASK => \&sh_task,           DESC => "Shared Pool Usage & Tuning Recommendations" },
 "-sl"           => { TASK => \&sl_task,           DESC => "SELECT * from table - name supplied as 2nd parameter" },
 "-sn"           => { TASK => \&sn_task,           DESC => "SNAPSHOT - Run systemstate trace 3 times (for Oracle diagnostics)" },
 "-so"           => { TASK => \&so_task,           DESC => "List sorts by user" },
 "-sp"           => { TASK => \&sp_task,           DESC => "List StatsPack Snapshot Data" },
 "-sq"           => { TASK => \&sq_task,           DESC => "Run SQL - Pass SQL as argument in DOUBLE QUOTES" },
 "-sr"           => { TASK => \&sr_task,           DESC => "View Running SQL - '$P -sr SESSION_NUMBER'" },
 "-ss"           => { TASK => \&ss_task,           DESC => "List default storage clause for all TABLESPACES" },
 "-st"           => { TASK => \&st_task,           DESC => "System Statistics" },
 "-su"           => { TASK => \&su_task,           DESC => "System Utilization" },
 "-sw"           => { TASK => \&sw_task,           DESC => "Session Wait Statistics" },
 "-sy"           => { TASK => \&sy_task,           DESC => "List All Granted System Privileged" },
 "-ta"           => { TASK => \&ta_task,           DESC => "List All Tables - From DBA_SEGMENTS (Name, TS, Size, Extents, Maxextents)" },
 "-tb"           => { TASK => \&tb_task,           DESC => "Count Tables - By Schema" },
 "-tc"           => { TASK => \&tc_task,           DESC => "Create SQL to count all rows in all permanent tables" },
 "-td"           => { TASK => \&td_task,           DESC => "Describe table - name of table supplied as 2nd parameter" },
 "-ti"           => { TASK => \&ti_task,           DESC => "Table Info - '$P -ti TABLE-OWNER.TABLE-NAME'" },
 "-tj"           => { TASK => \&tj_task,           DESC => "Datafile Space Allocated - By Filesystem" },
 "-tm"           => { TASK => \&tm_task,           DESC => "List Temporary Segments" },
 "-tp"           => { TASK => \&tp_task,           DESC => "List Objects (from DBA_SEGMENTS) By BLOCKS and BUFFER_POOL - Show buffer pool stats" },
 "-tq"           => { TASK => \&tq_task,           DESC => "Tablespaces - Details and Defaults" },
 "-tr"           => { TASK => \&tr_task,           DESC => "List Transactions" },
 "-ts"           => { TASK => \&ts_task,           DESC => "List Tablespaces" },
 "-tt"           => { TASK => \&tt_task,           DESC => "List 8i+ TEMPORARY Locally Managed Tablespaces" },
 "-tu"           => { TASK => \&tu_task,           DESC => "Temp space usage by user" },
 "-tz"           => { TASK => \&tz_task,           DESC => "List Tables - From DBA_TABLES" },
 "-ua"           => { TASK => \&ua_task,           DESC => "List USERS with incorrect default/temp tablespace - Creates SQL to ALTER USERS" },
 "-ub"           => { TASK => \&ub_task,           DESC => "List the byte count of data, by User" },
 "-ud"           => { TASK => \&ud_task,           DESC => "List Users With ORADBA" },
 "-ug"           => { TASK => \&ug_task,           DESC => "List Users Table Grants (Much Output) - 2nd Parm is optional grantee name" },
 "-un"           => { TASK => \&un_task,           DESC => "Display 9i UNDO Statistics" },
 "-us"           => { TASK => \&us_task,           DESC => "List All Users" },
 "-ut"           => { TASK => \&ut_task,           DESC => "List the byte count of data, by User and Tablespace" },
 "-v"            => { TASK => \&v_task,            DESC => "List Oracle version and whether it is 32-bit or 64-bit" },
 "-vs"           => { TASK => \&vs_task,           DESC => "List All Views" },
 "-vw"           => { TASK => \&vw_task,           DESC => "Count Views - By Schema" },
 "-wr"           => { TASK => \&wr_task,           DESC => "Wait Reasons - By User and Wait Type" },
 "-ws"           => { TASK => \&ws_task,           DESC => 'Wait stats - from v$waitstat' },
 "-zr"           => { TASK => \&zr_task,           DESC => 'List contents of DBAMON.STANDBY_REBUILD' }
);

#------------------------------------------------------------------------------
#
#                                    MAIN
#
#------------------------------------------------------------------------------

%sids = &getSIDS();                                            # V02.36

#
## If Nothing Specified... Tell User What Could Have Been Specified
#
&Usage() if not defined $arg;

&msg( "" );
if ( not defined $os )
{
  print "\n$P | Env var ORACLE_SID must be set\n";
  exit 1;
}

if ( exists $sids{ $os } )                                  #V02.36
{                                                           #V02.36
    $oh = $sids{ $os };                                     #V02.36
}                                                           #V02.36
elsif ( exists $sids{ "\U$os\E" } )                         #V02.36
{                                                           #V02.36
    $oh = $sids{ "\U$os\E" };                               #V02.36
}                                                           #V02.36
elsif ( exists $sids{ "\L$os\E" } )                         #V02.36
{                                                           #V02.36
    $oh = $sids{ "\L$os\E" };                               #V02.36
}                                                           #V02.36
else                                                        #V02.36
{                                                           #V02.36
    &msg( "Available Oracle SIDS:" );                       #V02.36
    &msg( join( "\n", sort keys %sids ) );                  #V02.36
    die "\nMatching ORACLE_SID Not Found Winthin Registry: $os\n"; # V02.36
}                                                           #V02.36

print STDERR "\$ENV{ORACLE_SID}=$ENV{ORACLE_SID}\n" if exists $ENV{ORACLE_SID} and 
                                                       exists $ENV{LEARN};

$oh =~ s/\\/\\\\/g;                                         #V02.37
$ENV{ORACLE_HOME} = $oh;
print STDERR "\$ENV{ORACLE_HOME}=$ENV{ORACLE_HOME}\n" if exists $ENV{ORACLE_HOME} and 
                                                       exists $ENV{LEARN};

#                                                              # V02.34
## Create $root_fs\\temp If It Does Not Exist                  # V02.34
#                                                              # V02.34
if ( ! -d $tmp_dir )                                           # V02.34
{                                                              # V02.34
    &msg( "Creating Non-Existent Directory: $tmp_dir" );       # V02.34
    system( "mkdir $tmp_dir 2>&1" );                           # V02.34
}                                                              # V02.34

#
## Get Hostname and Company From Domain
#
%values=();
$Register = "System\\CurrentControlset\\Services\\Tcpip\\Parameters";
$HKEY_LOCAL_MACHINE->Open($Register,$hkey)|| die "Cannot open registry $Register - $!";
$hkey->GetValues(\%values);
foreach $value (keys(%values))
{
   $RegType      = $values{$value}->[1];
   $RegValue     = $values{$value}->[2];
   $RegKey       = $values{$value}->[0];
   if ( $RegKey eq "Hostname" )
   {
       $host = $RegValue;
   }

   if ( $RegKey eq "Domain" )
   {
       $company = $RegValue;
       ( undef, $company, undef ) = split( /\./, $company, 3 );
   }
}
$hkey->Close();

print STDERR "\$host=$host\n" if defined $host and 
                                 exists $ENV{LEARN};

print STDERR "\$company=$company\n" if defined $company and 
                                       exists $ENV{LEARN};
#
## Add Agilent Specific Commands If That Is The Determined Company
#
if ($company eq "agilent") # Add This Task To Task Table If Company Is Agilent
{
  $TaskTbl{ "-cfbl" } = { TASK => \&cfbl_task, DESC => "\n--------------------------------------------------------------------\n"             .
                                                       "Agilent-Specific Commands:\n"                                                         .
                                                       "   -cfbl List trace controlfile backup with create timestamp > [numeric timestamp]\n" .
                                                       "--------------------------------------------------------------------" };
}

$date = &get_date("Y/M/D-H:M:S");                                  #V02.36
print "$date $P | ORACLE_SID=$os ORACLE_HOME=$oh\n";               #V02.36
print "$date $P | Version=$version Host=$host Company=$company\n"; #V02.36

#
## Determine Version Of Oracle We Are Connecting To Via SQLPLUS To BootStrap Ourselves
#
unless( -e "$oh\\bin\\sqlplus.exe" or -e "$oh\\bin\\plus80.exe" )
{
    print "\"SQLPLUS.EXE\" Could Not Be Found...\n";
    exit 1;
}

$os_scmd = ( -e "$oh\\bin\\sqlplus.exe" ) ? "$oh\\bin\\sqlplus.exe" : "$oh\\bin\\plus80.exe";
$sqlpluscmd = $os_scmd;
@banner  = `$os_scmd -?`;
$oraver  = undef;
foreach $line ( @banner )
{
    chomp( $line );
    next if $line !~ /^SQL/;

    ( @words ) = split( /\s+/, $line );
    $oraver    = $words[ 2 ];
    if ( $oraver =~ /^8.1|^9|^10/ )
    {
        $os_scmd .= " /NOLOG";
        $os_conn  = "connect / as sysdba";
    }
    elsif ( $oraver =~ /^8.0/ )
    {
        $os_scmd = "$oh\\bin\\svrmgr30.exe";
        unless( -e $os_scmd )
        { 
            print "\"$os_scmd\" could not be found.\n";
            exit 1;
        }
        $os_conn  = "connect internal";
    }
    else
    {
      print "Version $oraver is not yet supported, but we will give it a try\n";
      #exit 1;
    }
    last;
}

$oraver_num = scalar( substr( $oraver, 0, 5 ) );
$linesize   = ( $os_scmd =~ /svrmgr/ ) ? "SET CHARWIDTH 999"   : "SET LINESIZE 999";

#
## Get pfile location 
#
$pfile = undef;                                                 #V02.36
for( $i = 0; $i < 10; $i++ )
{
    eval
    {
        %values=();
        $Register = "SOFTWARE\\ORACLE\\HOME$i";
        $HKEY_LOCAL_MACHINE->Open($Register,$hkey) || die "Cannot open registry $Register -$!";
        $hkey->GetValues(\%values);
        $hkey->Close();
    };

    next if $@ or join( ' ', sort keys %values ) !~ /$os/i;      #V02.31
                                                                 #V02.31
    $key = 'ORA_' . uc( $os) . '_PFILE';                         #V02.31
    if ( exists $values{$key} and $values{$key}->[ 2 ] =~ /$oh/ ) #V02.36
    {                                                            #V02.31
        $pfile = $values{$key}->[ 2 ];                           #V02.31
        last;                                                    #V02.36
    }                                                            #V02.31
    elsif ( exists $values{RDBMS_CONTROL} )
    {
        $pfile  = $values{RDBMS_CONTROL}->[ 2 ];
        $pfile .= ( $oraver_num < 9 ) ? "\\init$os.ora" : "\\spfile$os.ora";
        last if -e $pfile;                                       #V02.36
    }
} 

if ( defined $pfile )                                                 #V02.36
{                                                                     #V02.36
    $msg  = "Pfile loc: $pfile";                                      #V02.36
    $msg .= " And Exists" if -e $pfile;                               #V02.36
    &msg( "$msg" );                                                   #V02.36
}                                                                     #V02.36
else                                                                  #V02.36
{                                                                     #V02.36
    &msg( "Warning: Unable to locate pfile location via registry!" ); #V02.36
}                                                                     #V02.36

#
## Get Init File Location, ASSUMES our Standard Location
#
$initfile = "$oh\\database\\init$os.ora";                             #V02.38
# Check for the tricky re-direct situation                            #V02.38
chomp( $line = `findstr /I /C:ifile $initfile` );                     #V02.38
if ( defined $line and length( $line ) > 0 )                          #V02.38
{                                                                     #V02.38
    #$line     =~ s/^\s+\s+$//g;                                       #V02.38
    $line     =~ s/\s+//g;                                            #V02.38
    @words    = split( /=/, $line );                                  #V02.38
    $initfile = pop @words;                                           #V02.38
    $initfile =~ s/'//g;                                              #V02.38
}                                                                     #V02.38

if ( ! -e $initfile )                                                     #V02.38
{                                                                         #V02.38
    &msg( "Warning: Standard loc for INIT.ORA: $initfile - NOT FOUND!" ); #V02.38
    if ( defined $pfile and -e $pfile )                                   #V02.38
    {                                                                     #V02.38
        &msg( "Warning: Using PFILE loc: $pfile for INIT.ORA Loc!" );     #V02.38
        $initfile = $pfile                                                #V02.38
    }                                                                     #V02.38
    else                                                                  #V02.38
    {                                                                     #V02.38
        &msg( "Warning: Unable to determine INIT.ORA location!" );        #V02.38
    }                                                                     #V02.38
}                                                                         #V02.38
else                                                                      #V02.38
{                                                                         #V02.38
    $msg  = "Ifile loc: $initfile";                                        #V02.38
    $msg .= " And Exists" if -e $initfile;                                #V02.38
    &msg( "$msg" );                                                       #V02.38
}                                                                         #V02.38

#
## Get Alert Log Location
#
$alrtdir = undef;                                                   #V02.36
if ( defined $initfile )                                            #V02.36
{                                                                   #V02.36
    chomp( $line = `findstr /I /C:background_dump_dest $initfile` ); #V02.36
    $line     =~ s/\s+//g;                                            #V02.38
    @words = split( /=/, $line );                                 #V02.36
    $alrtdir = pop @words;                                          #V02.38
    $alrtdir =~ s/^'//; $alrtdir =~ s/'$//;                         #V02.36
}                                                                   #V02.36
                                                                    #V02.36
if ( not defined $alrtdir )                                         #V02.36
{                                                                   #V02.36
    $sqlstmt = "SELECT 'X', VALUE " .                               #V02.36
               'FROM V$PARAMETER '  .                               #V02.36
               "WHERE NAME='background_dump_dest'";                 #V02.36
                                                                    #V02.36
    &oraSQL( $sqlstmt );                                            #V02.36
                                                                    #V02.36
    foreach ( @oraSQLout )                                          #V02.36
    {                                                               #V02.36                         
        next if $_ !~ /^X/;                                         #V02.36
                                                                    #V02.36
        ( undef, $alrtdir ) = split;                                #V02.36
        $alrtdir =~ s/^'//; $alrtdir =~ s/'$//;                     #V02.36
        last;                                                       #V02.36
    }                                                               #V02.36
}                                                                   #V02.36
                                                                    #V02.36
if ( defined $alrtdir and -d $alrtdir )                             #V02.36
{                                                                   #V02.36
    ### Currently The Predominant Version Is "SIDALRT.LOG"          #V02.33
    $alrtfile = "$alrtdir\\$os" . "ALRT.LOG";                       #V02.33
    $alrtfile =~ s/^\s+//g;                                         #V02.33
    $alrtfile =~ s/'//g;                                            #V02.33
                                                                    #V02.36
    if ( ! -e $alrtfile )                                           #V02.33
    {                                                               #V02.33
        ### Try Version "alert_SID.log"                             #V02.33
        $alrtfile = "$alrtdir\\alert_" . "$os" . '.log';            #V02.33
        $alrtfile =~ s/^\s+//g;                                     #V02.33
        $alrtfile =~ s/'//g;                                        #V02.33
                                                                    #V02.36
        if ( ! -e $alrtfile )                                       #V02.33
        {                                                           #V02.33
            $alrtfile = "$alrtfile - alert log not found!";         #V02.33
        }                                                           #V02.33
    }                                                               #V02.33
}                                                                   #V02.36
                                                                    #V02.36
if ( defined $alrtfile )                                            #V02.36
{                                                                   #V02.36
    $alrtfile =~ s/\?/$oh/;                                         #V02.36
    $msg  = "Alertlog loc: $alrtfile";                              #V02.36
    $msg .= " And Exists" if -e $alrtfile;                          #V02.36
    &msg( "$msg" );                                                 #V02.36
}                                                                   #V02.36
else                                                                #V02.36
{                                                                   #V02.36
    &msg( "Warning: Unable to locate Alertfile!" );                 #V02.36
}                                                                   #V02.36

#
## Already To Go...
#

#
## Make Sure That Task User Requested Exists In Task Table
#
if ( not exists $TaskTbl{$arg} )
{
    print "'$arg' - Not Supported\n";
    &Usage();
}

#
## Run It...
#
$codeRef = $TaskTbl{$arg}->{TASK};
&{$codeRef}();

#
## All Done
#
exit 0;

#------------------------------------------------------------------------------
#
# Subroutines  
#
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
sub Usage()
{
    print "\n$P | Please Set Environment Variable ORACLE_SID!\n" if not defined $os;

    print "$P | Usage:\n";
    foreach my $key ( sort keys %TaskTbl )
    {
        $refTask = $TaskTbl{$key};
        if ( length( $key ) < 7 )
        {
            #printf( "%-3s %-6s %s\n", ' ', $key, %{$refTask}->{DESC} );
        }
        else
        {
            #print "   $key  " . %{$refTask}->{DESC} . "\n";
        }
    }

    exit 0;
}

#------------------------------------------------------------------------------
sub getSIDS
#------------------------------------------------------------------------------
{
    my %sids  = ();
    @Services = ();
    %Values   = ();
    $Key      = "";
    $Path     = "SYSTEM\\CurrentControlSet\\Services";
    $HKEY_LOCAL_MACHINE->Open( $Path, $Key )
                         or
                         die "Can't open registry $Path: $!";
    $Key->GetKeys( \@Services );
    $Key->Close();

    foreach $service ( sort @Services )
    {
        next if $service !~ /OracleService/i;

        $sid    = substr( $service, 13 );
        %Values = ();
        $Path1  = "$Path\\$service";
        $HKEY_LOCAL_MACHINE->Open( $Path1, $Key )
                             or
                             die "Can't open registry $Path1: $!";
        $Key->GetValues( \%Values );
        $Key->Close();

        $oh = $Values{ ImagePath }->[ 2 ];
        @split_path   = split( /\\/, $oh );
        $oh           = pop @split_path; # Remove ORACLE.EXE SID
        $oh           = pop @split_path; # Remove bin
        $oh           = join( '\\', @split_path ); # Set path
        chop( $oh ) if $oh =~ /\\$/;
        $sids{ $sid } = "$oh";
    }
   
    return( %sids );
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
sub msg
{
  $msgmsg = $_[0];
  $date = &get_date("Y/M/D-H:M:S");
  print "$date ${P} | $msgmsg\n";
}

#------------------------------------------------------------------------------
#
#------------------------------------------------------------------------------
sub get_date()
{
  # Return date in format supplied as argument. Valid formats are:
  #   Y/M/D-H:M:S
  #   YMD_HMS

  $format = @_[0];

  @sysdate = localtime;
  for ($i = 0; $i <= 5; $i++)
  {
    if (@sysdate[$i] < 10)
    {
      @sysdate[$i] = 0 . @sysdate[$i];
    }
  }

  if ($format eq "Y/M/D-H:M:S")
  {
    $sysdate = @sysdate[5]+1900 . "/" . ++@sysdate[4] . "/" . @sysdate[3] . "-" . @sysdate[2] . ":" . @sysdate[1] . ":" . @sysdate[0];
  }
  elsif ($format eq "YMD_HMS")
  {
    $sysdate = @sysdate[5]+1900 . ++@sysdate[4] . @sysdate[3] . "_" . @sysdate[2] . @sysdate[1] . @sysdate[0];
  }
  else
  {
    &msg("Invalid format passed to get_date. Must be Y/M/D-H:M:S or YMD_HMS");
  }
  return $sysdate;
}

#------------------------------------------------------------------------------
sub timeToText
#------------------------------------------------------------------------------
{
  $zz_ts = $_[0];
  ($sec,$min,$hour,$mday,$mon,$year) = localtime($zz_ts);
  $year += 1900;
  $mon += 1;

  # Pad with 0's
  $yy = sprintf("%04d",$year);
  $mm = sprintf("%02d",$mon);
  $dd = sprintf("%02d",$mday);
  $hh = sprintf("%02d",$hour);
  $mi = sprintf("%02d",$min);
  $ss = sprintf("%02d",$sec);

  $date = "$yy/$mm/$dd-$hh:$mi:$ss";
  return $date;
}

#------------------------------------------------------------------------------
sub timeToNum
#------------------------------------------------------------------------------
{
  $zz_text = $_[0];
  $zz_text =~ s/-/\//g;
  $zz_text =~ s/:/\//g;
  ##print "zz_text=$zz_text\n";
  @words = split(/\//,$zz_text);

  if ($#words != 5)
  {
    &error("Input Target_End_Time ($zz_text) should be in the format: YYYY/MM/DD-HH:MM:SS","DOC116");
  }

  $zz_yyyy = $words[0];
  $zz_mon  = $words[1];
  $zz_dd   = $words[2];
  $zz_hh   = $words[3];
  $zz_mm   = $words[4];
  $zz_ss   = $words[5];
  ##print "zz_yyyy=$zz_yyyy zz_mon=$zz_mon zz_dd=$zz_dd zz_hh=$zz_hh zz_mm=$zz_mm zz_ss=$zz_ss\n";

  &error("Input Target_End_Time - YYYY must be 2002-2099","DOC116") if ($zz_yyyy < 2002 || $zz_yyyy > 2099);
  &error("Input Target_End_Time - MON  must be 1-12     ","DOC116") if ($zz_mon  <    1 || $zz_mon  >   12);
  &error("Input Target_End_Time - DD   must be 1-31     ","DOC116") if ($zz_dd   <    1 || $zz_dd   >   31);
  &error("Input Target_End_Time - HH   must be 0-23     ","DOC116") if ($zz_hh   <    0 || $zz_hh   >   23);
  &error("Input Target_End_Time - MM   must be 0-59     ","DOC116") if ($zz_mm   <    0 || $zz_mm   >   59);
  &error("Input Target_End_Time - SS   must be 0-59     ","DOC116") if ($zz_ss   <    0 || $zz_ss   >   59);

  @zz_date=($zz_ss,$zz_mm,$zz_hh,$zz_dd,--$zz_mon,$zz_yyyy);
  $zz_result = timelocal(@zz_date);

  return $zz_result;
}

#------------------------------------------------------------------------------
sub error
#------------------------------------------------------------------------------
{
  $emsg = $_[0];
  &msg("ERROR: $emsg");
  exit 1;

  $ENV{'ORACLE_SID'} = "$instance";
  $str1="${backup_type}: ($errnum) - $msg (${webserver}/oberr.html#${errnum})";
  $str2="DB: $instance";
  $str3="Backup Level: $backup_type";
  $str4="For more information see: ${webserver}/processes/oracle/ora_backup_errors.html#$errnum";
  &msg("Sending BLT ticket $str1 $str2 $str3 $str4");
  $progstring = "${P}[${errnum}]";
  $ENV{'BLT_SUBJECT'} = "${host}-${instance}|${backup_type} [$errnum] - $msg";

  system("/usr/local/dba/tools/db_create_blt.pl $progstring \"$str1\" \"$str2\" \"$str3\" \"$str4\"");

  if ($errnum == -86 || 
      $errnum == -88 ||
      $errnum == -90 ||
      $errnum == -91)
  {
    #&blt_dcit;   comment this for now - the tickets aren't working anyway bb 09aug01
  }

  &msg("=============================================================");
  &msg("Fatal Error: $msg");
  &msg("=============================================================");
  &msg("Exiting rc=1");
  exit 1;
}

#------------------------------------------------------------------------------
# executes a SQL statement and puts results in @oraSQLout 
#------------------------------------------------------------------------------
sub oraSQL
{
  my $sql    = $_[0];
  @sql_stmts = split( '\;', $sql );

  open( SO, ">$sqlin" );
  print SO "$os_conn;\n";
  print SO "$linesize;\n";
  print SO "set echo on;\n";
  print SO "whenever sqlerror exit 1\n";

  foreach $statement (@sql_stmts)
  {
     $statement =~ s/^\s+//;                     # V02.30
     $statement =~ s/\n//;                       # V02.30
     $statement =~ s/;$//;                       # V02.30
     $statement =~ s/ v\\\$/ v\$/g;              # V02.38
     $statement =~ s/ x\\\$/ x\$/g;              # V02.38
     $statement =~ s/ obj\\\$/ obj\$/g;          # V02.38

     next if length( $statement ) == 0;          # V02.30

     $statement .= ';';

     if (exists $ENV{'LEARN'})
     {
         print "\n#===========================================================#\n";
         &msg("LEARN Mode: SQL Being Executed: \n $statement");
         print "#===========================================================#\n";
     }

     print SO "$statement\n";
  }
  print SO "exit;\n";
  close( SO );

  @oraSQLout = `$os_scmd \@$sqlin`;
  $oraCMDrc  = $?;                             # V02.34

  #
  ## Report Any ORA Errors Before Returning
  #
  if ( grep( /^ORA/, @oraSQLout ) && ! grep( /^ORA-00942/, @oraSQLout ) )
  {
      print "\n#SQL Error Start >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>#\n";
      print "SQL Error >>> $sqlin <<< Commands\n";
      system( "type $sqlin" );
      print "\n";
      print "SQL Error >>> Listing <<<\n";
      foreach $line ( @oraSQLout ) { print "$line"; }
      print "\n#SQL Error End <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<#\n";
  }
   
  unlink "$sqlin" if -f "$sqlin";              # V02.30

  return;
}

#------------------------------------------------------------------------------
# executes a SQL statement and puts results in @oraSQLout 
#------------------------------------------------------------------------------
sub oraSQLold
{
  my $sql      = @_[0];
  my $conn_str = @_[1];

  if (not defined $conn_str)
  {
     print "Connection string not passed. Check syntax for call to \"oraSQLold\".\n";
     exit 1;
  }

  if ($os_scmd =~ "sqlplus")
  {
    $os_set1 = "set linesize 9999
                set echo on
                set long 2000000000
                set pagesize 50000";
    $os_set2 = "set numwidth 20";
    $os_end  = "exit";
  }
  elsif ($os_scmd =~ "plus80")
  {
    $os_set1 = "set long 2000000000
                set echo on
                set pagesize 50000";
    $os_set2 = "set numwidth 20";
    $os_end  = "exit";
  }
  else
  {
    $os_set1 = "set numwidth 20";
    $os_set2 = "set echo on";
    $os_end  = "exit";
  }
  ##&msg("os_scmd=$os_scmd os_conn=$os_conn");

  $sql =~ s/\\\$/\$/g;                               # V02.30
  if ($ENV{'LEARN'} ne "")
  {
      print "\n#===========================================================#\n";
      &msg("LEARN Mode: SQL Being Executed: \n $sql");
      print "#===========================================================#\n";
  }

  open (SO,">$sqlin");
  print SO ("$conn_str;
             $os_set1;
             $os_set2;
             whenever sqlerror exit 1;
             $sql
             $os_end;");
  close(SO);
  
  @oraSQLout = `$os_scmd \@$sqlin`;
  $oraCMDrc  = $?;                             # V02.34

  #
  ## Report Any ORA Errors Before Returning
  #
  if ( grep( /^ORA/, @oraSQLout ) && ! grep( /^ORA-00942/, @oraSQLout ) )
  {
      print "\n#SQL Error Start >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>#\n";
      print "SQL Error >>> $sqlin <<< Commands\n";
      system( "type $sqlin" );
      print "\n";
      print "SQL Error >>> Listing <<<\n";
      foreach $line ( @oraSQLout ) { print "$line"; }
      print "\n#SQL Error End <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<#\n";
  }
   
  unlink "$sqlin" if -f "$sqlin";              # V02.30
}

sub subEMerr # Send Failures To Me                                  # V02.34
{                                                                   # V02.34
  my( $to, $subject, $emsg ) = @_;                                  # V02.34
  &subEMail( $to, $subject, $emsg );                                # V02.38
  exit;                                                             # V02.38
} # End Of Subroutine subEMerr                                      # V02.34

sub subEMail # Send Failures To Me                                  # V02.34
{                                                                   # V02.34
  my( $to, $subject, $emsg ) = @_;                                  # V02.34
  &msg( "$subject" );                                               # V02.34
  $emsg        = ( not defined $emsg or length $emsg == 0 ) ?       # V02.38
                 "???"                                      :       # V02.38
                 $emsg;                                             # V02.38
  my $mailcmd  = "perl C:\\dbatools\\socketmail.pl " .              # V02.34
                 'larry_rush@agilent.com';                          # V02.38
  my $cmd      = "$mailcmd $to \"$subject\" $emsg";                 # V02.34
  system( "$cmd" );                                                 # V02.34
} # End Of Subroutine subEMail                                      # V02.34

# --------------------------------------------------------------------
# - 
# --------------------------------------------------------------------
sub dash_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
    
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  # Check for Oracle services

  $svc1 = "OracleService$os";
  %status = "";
  Win32::Service::GetStatus('', $svc1, \%status);

  if ($status{CurrentState} == 4)
  {
    $status = "Running";
  }
  else
  {
    $status = "Not Running";
  }

  print "$P | $svc1: $status\n\n";

  if ($status ne "Running") { exit 1 };

  $shu = "?";
  $sta = "?";
  $ver = "?";
  $sts = "?";
  $ins = "?";
  $par = "?";
  $log = "?";
  $cha = "?";
  $scn = "?";
  $asc = "?";

  # Determine Oracle version

  $sqlstmt = "select 'V', banner from v\$version;";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @lines = split("\n");
    @words = split;
    $verb  = $words[0];
    if ($verb eq "ORA-01034:")
    {
      $sta = "(Oracle Not Available)";
    }
    elsif ($verb =~ "ORA-01012")
    {
      print "";
    }
    elsif ($verb =~ "ORA-")
    {
      print "\nERROR: ${lines[0]}\n\n";
    }
    if ($verb eq "V")
    {
      if (/10g/)
      {
        $ver = $words[7];
        if ($_ =~ /64b/)
        {
          $ver = $ver . " (64-bit)";
        }
      }
      if (substr($words[1],0,7) eq "Oracle8")
      {
        $ver = $words[5];
        $z = substr($ver,length($ver)-1,1);
        if ($z eq "," || $_ =~ /64bit/)
        {
          chop($ver) if $z eq ",";
          $ver = $ver . " (64-bit)";
        }
      }
      if (substr($words[1],0,7) eq "Oracle9")
      {
        $ver = $words[5];
        $z = substr($ver,length($ver)-1,1);
        if ($z eq "," || $_ =~ /64bit/)
        {
          chop($ver) if $z eq ",";
          $ver = $ver . " (64-bit)";
        }
      }
      elsif (substr($words[1],0,7) eq "Oracle8" &&
          $words[5] eq "Production")
      {
        $ver = $words[3];
      }
      elsif (substr($words[1],0,7) eq "Oracle7")
      {
        $ver = $words[4];
      }
    }
  }

  if (substr($ver,0,1) eq "7")
  {
    $sqlstmt = "select 'X',
      name, checkpoint_change#, archive_change# from v\$database;
    select 'A', value from v\$instance where key = 'RESTRICTED MODE';
    select 'B', value from v\$instance where key = 'SHUTDOWN PENDING';
    select 'Z', value from v\$nls_parameters where parameter = 'NLS_CHARACTERSET';
    select 'C',
      to_char(to_date(value,'J'),'YYYY/MM/DD'),
      sysdate - to_date(value,'J')
      from v\$instance where key = 'STARTUP TIME - JULIAN';";
  }
  elsif ($ver =~ /^8.0/)
  {
    $sqlstmt = "select 'X', 
    instance_name, 
    to_char(startup_time,'YYYY/MM/DD-HH24:MI:SS') startup_time, 
    version, 
    status, 
    shutdown_pending, 
    nvl(parallel,'?'), 
    logins,
    sysdate - startup_time
    from v\$instance;
    select 'W', CHECKPOINT_CHANGE#, ARCHIVE_CHANGE#, CONTROLFILE_TYPE, dbid, name from v\$database;
    select 'Z', value from  v\$nls_parameters where parameter = 'NLS_CHARACTERSET';";
  }
  elsif ($ver =~ /^8.1.[4-7]/ )
  {
    $sqlstmt = "select 'X', 
    instance_name, 
    to_char(startup_time,'YYYY/MM/DD-HH24:MI:SS') startup_time, 
    version, 
    status, 
    shutdown_pending, 
    nvl(parallel,'?'), 
    logins,
    sysdate - startup_time
    from v\$instance;
    select 'V', 
      to_char(created,'YYYY/MM/DD-HH24:MI:SS'),
      log_mode,
      '?' from v\$database;
    select 'W', CHECKPOINT_CHANGE#, ARCHIVE_CHANGE#, CONTROLFILE_TYPE, dbid, name, open_mode from v\$database;
    select 'Z', value from  v\$nls_parameters where parameter = 'NLS_CHARACTERSET';";
  }
  else
  {
    $sqlstmt = "select 'X', 
    instance_name, 
    to_char(startup_time,'YYYY/MM/DD-HH24:MI:SS') startup_time, 
    version, 
    status, 
    shutdown_pending, 
    nvl(parallel,'?'), 
    logins,
    sysdate - startup_time
    from v\$instance;
    select 'V', 
      to_char(created,'YYYY/MM/DD-HH24:MI:SS'),
      log_mode,
      force_logging from v\$database;
    select 'W', CHECKPOINT_CHANGE#, ARCHIVE_CHANGE#, CONTROLFILE_TYPE, dbid, name, open_mode from v\$database;
    select 'Z', value from  v\$nls_parameters where parameter = 'NLS_CHARACTERSET';";
  }

  ##print "sqlstmt=$sqlstmt\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "ORA-01034:")
    {
      $sta = "(Oracle Not Available)";
    }
    if ($verb eq "V")
    {
      $created = $words[1];
      $logmode = $words[2];
      $forclog = $words[3];
    }
    if ($verb eq "W")
    {
      $scn = $words[1];
      $asc = $words[2];
      $cfm = $words[3];
      $dbid= $words[4];
      $dbnm= $words[5];
      $open= $words[6] . $words[7];
    }
    if ($verb eq "X")
    {
      if (substr($ver,0,1) eq "7")
      {
        $ins = $words[1];
        $scn = $words[2];
        $asc = $words[3];
        $dbid= $words[4];
        $sta = "(PMON Running)";
      }
      else
      {
        $ins = $words[1];
        $sts = $words[2];
        $sta = $words[4];
        $shu = $words[5];
        $par = $words[6];
        $log = $words[7];
        $upt = $words[8];
        $upt = $upt * 24;
      }
    }

    if ($verb eq "A")
    {
      if ($words[1] eq "0") { $log = "ALLOWED"; }
                       else { $log = "RESTRICTED"; }
    }

    if ($verb eq "B")
    {
      if ($words[1] eq "0") { $shu = "NO"; }
                       else { $shu = "YES"; }
    }

    if ($verb eq "C")
    {
       $sts = $words[1];
       $upt = $words[2];
       if ($sta eq "?" && $upt ne "?")
       {
          $sta="NOMOUNT";
       }
       elsif ($sta eq "(PMON Running)" && $upt ne "?")
       {
          $sta = "MOUNTED";
          $sqlstmt = "select 'P',user_id from dba_users;";
	  &oraSQL($sqlstmt);
	  foreach (@oraSQLout)
          {
             ##print;
             @words = split;
             $verb    = $words[0];
             if ($verb =~ "ORA-")
             {
                print "$verb\n";
             }
             if ($verb eq "P") 
             {
                $sta = "OPEN";
             }
          }
          #$cur_time=`date +%H:%M:%S`;
          chomp( @lines = `pslist ORACLE` );
          $lastline     = pop @lines; 
          $lastline     =~ s/^\s+//;
          $lastline     =~ s/\s+$//;
          @words        = split( /\s+/, $lastline );
          $elapsedtime  = pop @words;
          ( $hr,
            $min,
            $sec )      = split( ':', $elapsedtime, 3 );
          $numsecs      = ( $hr * 60 * 60 ) + ( $min * 60 ) + $sec;
          $upt          = ( $numsecs / ( 60 * 60 ) );
        }

    }
    if ($verb eq "Z")
    {
      $cha = $words[1];
    }
  }

  $daysup = sprintf( "%.2f", ( $upt / 24.0 ) );

  print "orastat | Status:       $sta\n";
  print "orastat | OpenMode:     $open\n" if $open ne "";
  print "orastat | Started:      $sts\n";
  print "orastat | HoursUp:      $upt\n";
  print "orastat | DaysUp:       $daysup\n";
  print "orastat | Instance:     $ins\n";
  print "orastat | OracleVer:    $ver\n";
  print "orastat | ShutdownMode: $shu\n";
  print "orastat | Parallel:     $par\n";
  print "orastat | Logins:       $log\n";
  print "orastat | Characterset: $cha\n";
  print "orastat | CkptSCN:      $scn\n";
  print "orastat | ArcSCN:       $asc\n";
  print "orastat | CF Mode:      $cfm\n";
  print "orastat | DBID:         $dbid\n";
  print "orastat | DBName:       $dbnm\n";
  print "orastat | Created:      $created\n" if $created;
  print "orastat | LogMode:      $logmode\n" if $logmode;
  print "orastat | ForceLogging: $forclog\n" if $forclog;

  exit;
}

# --------------------------------------------------------------------
# -ab
# --------------------------------------------------------------------
sub ab_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ab" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($arg2 ne "")
  {
    $where = "and r.sequence# = $arg2";
    &msg("For Archive Log #$arg2");
  }
  else
  {
    $where = "";
  }

  $sqlstmt = "
    select
      \'X\',
      r.thread#,
      r.sequence#,
      r.blocks*r.block_size archive_size,
      to_char(r.first_time,\'YYYY/MM/DD-HH24:MI:SS\') first_time,
      to_char(r.next_time,\'YYYY/MM/DD-HH24:MI:SS\') next_time,
      (sysdate - r.next_time)*24 archive_age_hrs,
      s.backup_type,
      to_char(s.completion_time,\'YYYY/MM/DD-HH24:MI:SS\'),
      (sysdate - s.completion_time)*24 backup_age_hrs,
      s.elapsed_seconds
    from
      v\$backup_redolog r,
      v\$backup_set s
    where
      r.set_stamp = s.set_stamp and
      r.set_count = s.set_count
      $where order by 5;
  ";
  ##print "sqlstmt=$sqlstmt\n";

  print "\n";
  print "Arc Thread \n";
  print "  Sequence       Size Redo Start Time     Redo End Time       ArcAge Backup End Time     BckAge BckTim\n";
  print "---------- ---------- ------------------- ------------------- ------ ------------------- ------ ------ \n";

  &oraSQL($sqlstmt);
  $arccount = 0;
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    if ($verb eq "X")
    {
      $thread     = $words[1];
      $seq        = $words[2];
      $size       = $words[3];
      $redostart  = $words[4];
      $redoend    = $words[5];
      $arcage     = $words[6];
      #$bcktype    = $words[7];
      $bckend     = $words[8];
      $bckage     = $words[9];
      $bckelapsed = $words[10];

      printf "%-1s %8s %10s %19s %19s %6.2f %19s %6.2f %6.0f \n",$thread,$seq,$size,$redostart,$redoend,$arcage,$bckend,$bckage,$bckelapsed;
      $arccount++;
    }
  }
  print "\n";
  &msg("$arccount Archive Log Backup(s) Listed");
  exit;
}

# --------------------------------------------------------------------
# -ad
# --------------------------------------------------------------------
sub ad_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ad" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\', 
    DEST_ID      ,
    STATUS       ,
    BINDING      ,
    NAME_SPACE   ,
    REOPEN_SECS  ,
    TARGET       ,
    DESTINATION  ,
    to_char(FAIL_DATE,\'YYYY/MM/DD-HH24:MI:SS\') FAIL_DATE, 
    sysdate - fail_date FAIL_AGO,
    FAIL_SEQUENCE,
    ERROR
    from v\$archive_dest order by 2;';

  #&msg("sqlstmt=$sqlstmt");
  &msg("Archive Destinations (from v\$archive_dest)\n");
  print "Num  Status   Binding    NameSpc  Reopen  Target   Destination/Errors\n";
  print "==== ======== ========== ======== ======= ======== ====================================\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb    = $words[0];
    $dest    = $words[1];
    $status  = $words[2];
    $binding = $words[3];
    $name    = $words[4];
    $reopen  = $words[5];
    $target  = $words[6];
    $dest2   = $words[7];
    $faildt  = $words[8];
    $failago = $words[9] * (24);
    #$failsq  = $words[10];
    $error1  = $words[11];
    $error2  = $words[12];
    $error3  = $words[13];
    $error4  = $words[14];
    $error5  = $words[15];

    $dest2   .= " " . $error1 . " " . $error2 . " " . $error3 . " " . $error4 . " " . $error5;
    $dest2   .= "\n                                                   Failed At: $faildt" if $faildt ne "0" && $faildt ne "";
    $dest2   .= "\n                                                   Failed $failago Hours Ago" if $failago ne "0" && $failago ne "";

    if ($verb eq "X")
    {
      printf "%-4s %-8s %-10s %-8s %-7s %-8s %-30s\n",
              $dest,$status,$binding,$name,$reopen,$target,$dest2;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -af
# --------------------------------------------------------------------
sub af_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-af" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &oraSQLold("archive log list;",$os_conn);
      
  foreach (@oraSQLout)
  {
    @words = split( /\s+/, $_ );
    $verb    = $words[0];
    if ($_ =~ /^Database log mode/)
    {
      $arclogmode = $words[3];
    }
    elsif ( $_ =~ /^Automatic archival/ )
    {
      $arcauto = $words[2];
    }
    elsif ( $_ =~ /^Archive destination/ )
    {
      $arcdir  = $words[2];
      last;
    }
  }

  &msg("arclogmode: $arclogmode");
  &msg("arcauto   : $arcauto");
  &msg("arclogdir : $arcdir");

  opendir(ARCDIR,"$arcdir");
  @filenames = readdir(ARCDIR);
  @filenamessorted = sort @filenames;
  closedir(ARCDIR);
  print "\n";

  print "Archive Log File Name                                  MTime      Age(Secs) Size      MTime-Formatted     CTime-Formatted\n";
  print "------------------------------------------------------ ---------- --------- --------- ------------------- -------------------\n";
  $numlogs = 0;
  $totsize = 0;
  $age     = -1;
  foreach $file (@filenamessorted)
  {
    next if ($file !~ /^arc/i);
    $arcmtime = ( stat "$arcdir/$file" ) [9];
    $arcsize  = ( stat "$arcdir/$file" ) [7];
    $arcctime = ( stat "$arcdir/$file" ) [10];
    ($sec,$min,$hour,$mday,$mon,$year) = localtime($arcmtime);
    $arcmtime_fmt = sprintf( "%4d/%02d/%02d-%02d:%02d:%02d",
                             $year + 1900, $mon + 1, $mday, $hour, $min, $sec );
    ($sec,$min,$hour,$mday,$mon,$year) = localtime($arcctime);
    $arcctime_fmt = sprintf( "%4d/%02d/%02d-%02d:%02d:%02d",
                             $year + 1900, $mon + 1, $mday, $hour, $min, $sec );

    $age = time - $arcmtime;
    $oldest = $age if $age > $oldest;

    printf "%-54s %10s %9s %9s %s %s\n","$arcdir/$file",$arcmtime,$age,$arcsize,$arcmtime_fmt,$arcctime_fmt;
    $numlogs++;
    $totsize += $arcsize;
  }

  $totsize /= 1000000;
  $agedays = ( $age / ( 24 * 60 * 60 ) < 0 ) ? 0 : $age / ( 24 * 60 * 60 );
  $agedays = sprintf("%8.3f",$agedays);
  $olddays = $oldest / (24*60*60);
  $olddays = sprintf("%8.3f",$olddays);

  print "\n";
  &msg("Total number of archive logs: $numlogs");
  &msg("Total size   of archive logs: $totsize (MB)");
  &msg( sprintf( "%s %4d %s", "Most recent archive log age: ", ( $age < 0 ) ? 0 : $age,  " (Seconds)" ) );
  &msg("Most recent archive log age:  $agedays (Days)");
  &msg("Oldest archive log age:       $olddays (Days)");

  exit;
}

# --------------------------------------------------------------------
# -al
# --------------------------------------------------------------------
sub al_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-al" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $where = "";
  $where = "and sequence# = $arg2" if $arg2 ne "";
  $where = "and first_time >= (sysdate - $arg3)" if $arg3 ne "";

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - arg2=$arg2 arg3=$arg3 arg4=$arg4");

  $sqlstmt = "select \'X\', 
    THREAD#, 
    SEQUENCE#, 
    name,
    FIRST_CHANGE#, 
    to_char(first_time,\'YYYY/MM/DD-HH24:MI:SS\'),
    blocks,
    block_size,
    to_char(completion_time,\'YYYY/MM/DD-HH24:MI:SS\')
  from
    v\\\$archived_log 
  where
    creator != 'RMAN' $where
  order by 
    6  ;";

  ##print "sqlstmt=$sqlstmt\n";

  &msg("Archived Logs History - Sorted by 1stTimestamp");
  print "Thread   Seq. Archived Log File Name                                  1st SCN 1stTimestamp           File Size Completion Timestamp      \n";
  print "------ ------ ----------------------------------------------- --------------- ------------------- ------------ --------------------------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb   = $words[0];
    $thread = $words[1];
    $seq    = $words[2];
    $name   = $words[3];
    $fstchg = $words[4];
    $fsttime= $words[5];
    $blks   = $words[6];
    $blksize= $words[7];
    $cmptime= $words[8];

    if ($verb eq "X")
    {
      if ($arg4 eq "FINDIT")
      {
        next if $name =~ /STANDBY/;

        ##&msg("Looking for $name and $name.gzRMAN/bkRMAN");
        next if ! -e "$name" && ! -e "$name.gzRMAN" && ! -e "$name.bkRMAN";
      }

      $ctime_num = &timeToNum($cmptime);
      $logsize = ($blks * $blksize) + 1024;
      ##print "fsttime=$fsttime\n";
      printf "%6d %6d %-47s %15s %-19s %12d %-19s %12d \n",
        $thread,$seq,$name,$fstchg,$fsttime,$logsize,$cmptime,$ctime_num;
      #last if ++$counter > 500;
      ++$counter;
    }
  }
  close(RO);

  if ($counter < 1)
  {
    print "$P | ERROR: No rows found\n";
  }
  exit;
}

# --------------------------------------------------------------------
# -an
# --------------------------------------------------------------------
sub an_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-an" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  @words = split(/\./,$arg2);
  $owner = uc $words[0];
  $table = uc $words[1];
  if ($owner eq "" || $table eq "")
  {
    print "$P | Usage: $P -an TABLE-OWNER.TABLE-NAME\n";
    exit 1;
  }
  print "Running ANALYZE TABLE for table $owner.$table\n\n";
  open(RO,"$os_scmd <<! 2>&1
$os_conn
analyze table $owner.$table compute statistics for table;
select \'Y\', avg_row_len, initial_extent, next_extent, num_rows 
   from dba_tables 
   where table_name = '$table' and owner = '$owner';
!|") || die "Cannot select from $table";
  while (<RO>)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $rows = $words[1];
    }
    if ($verb eq "Y")
    {
      $rowsize = $words[1];
      $initial = $words[2];
      $next    = $words[3];
      $rows    = $words[4];
      print "Current Row Count:              $rows\n";;
      print "Average Row Size (Bytes):       $words[1]\n";
      print "Initial Extent Size (Bytes):    $initial\n";
      print "Next Extent Size (Bytes):       $next\n";
    }
  }
  $bytes = ($rowsize * $rows) * 1.25;
  $meg = int($bytes / (1024*1024)) + 1;
  print "Calculated Space Reqd. (Bytes): $bytes\n";
  print "Calculated Space Reqd. (MB):    $meg\n";
  exit;
}

# --------------------------------------------------------------------
# -ar
# --------------------------------------------------------------------
sub ar_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ar" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &msg("Collecting 'before' Sample");
  $sqlstmt = 'select \'X\', value, name from v\$sysstat where name like \'%user%\';';

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $value = $words[1];
      $name1 = $words[2];
      $name2 = $words[3];
      if ($name1 eq "user" && $name2 eq "calls")   { $b4_calls   = $value; }
      if ($name1 eq "user" && $name2 eq "commits") { $b4_commits = $value; }
      $b4_time = time;
    }
  }

  &msg("Sleeping 10 seconds");
  sleep 10;
  &msg("Collecting 'after' Sample");

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $value = $words[1];
      $name1 = $words[2];
      $name2 = $words[3];
      if ($name1 eq "user" && $name2 eq "calls")   { $af_calls   = $value; }
      if ($name1 eq "user" && $name2 eq "commits") { $af_commits = $value; }
      $af_time = time;
    }
  }

  $df_commits = $af_commits - $b4_commits;
  $df_calls   = $af_calls   - $b4_calls  ;
  $df_time    = $af_time    - $b4_time   ;

  $rt_calls   = $df_calls / $df_time;
  $rt_commits = $df_commits / $df_time;

  print "$P | Time Interval: $df_time seconds\n\n";
  print "                       Commit Count     Call Count \n";
  print "                       ------------   ------------ \n";
  printf "Sample #1:             %12d   %12d \n",$b4_commits,$b4_calls;
  printf "Sample #2:             %12d   %12d \n",$af_commits,$af_calls;
  print "                       ============   ============ \n";
  printf "Difference:            %12d   %12d \n",$df_commits,$df_calls;
  printf "Rate (Per Second):     %12.1f   %12.1f \n",$rt_commits,$rt_calls;
  exit;
}

# --------------------------------------------------------------------
# -au
# --------------------------------------------------------------------
sub au_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-au" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', 
    name, value, \'()\', description from v\$parameter 
    where name like \'%aud%\' order by 2;';

  print "Parm                                     Value                               Desc.\n";
  print "---------------------------------------- ----------------------------------- ------------------------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    $name   = $words[1];
    $value  = $words[2];
    $verb2  = $words[3];

    $ndx    = index($_,"()");
    $desc   = substr($_,$ndx+3,40);

    if ($verb eq "X")
    {
      printf "%-40s %-35s %-20s\n",$name,$value,$desc;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -av
# --------------------------------------------------------------------
sub av_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-av" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($oraver_num < 8.1)
  {
      &msg("orastat -av Does not work on Oracle < 8.1");
      exit;
  }

  $days = $arg2;
  $days = 30 if $days eq ""; 

  if ($arg3 eq "H")
  {
    $timestring = "YYYY/MM/DD-HH24";
    $num = 60;
  }
  else
  {
    $arg3 = "D";
    $timestring = "YYYY/MM/DD";
    $num = 1440;
  }

  $days = $arg2;
  $days = 30 if $days eq ""; 

  $rows = 0;

  #---- 3.01 Determine "match string"
  $av_match = uc( "%archv0%" );   # default
  $sqlstmt = "
    select
      \'X\',
      destination
    from
      v\\\$archive_dest
    where
      destination is not null
    order by
      dest_id;";
   &oraSQL($sqlstmt);
   foreach (@oraSQLout)
   {
     ##print;
     @words   = split();
     $verb    = shift @words;
     if ($verb eq "X")
     {
       $dest     = shift @words;
       $av_match = uc( "%$dest%" );
       last;
     }
   }


  &msg("-av: Archivelog Volume - For the last $days days - Mode=$arg3");
  &msg("     Destination match string: $av_match" );

  # 2.96 and 3.00 changed line:    and name like '%archv01%'
  $sqlstmt = "
    select
      \'X\',
      to_char(completion_time,\'$timestring\'),
      sum(blocks * block_size),
      count(*)
    from
      v\\\$archived_log
    where
      completion_time > sysdate - $days
      and name like '$av_match'
    group by
      to_char(completion_time,\'$timestring\');";

  print "\n";
  print "              Archive    Archive    REDO Rate \n";
  print "Timestamp     Data (MB)  Log Count  (MB/Min)* \n";
  print "------------- ---------- ---------- ----------\n";

  $rows = 0;

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $ts      = shift @words;
      $size    = int((shift @words) / 1000000);
      $count   = shift @words;
      $mbmin   = int($size / $num);
  
      printf "%-13s %10s %10s %10s\n",$ts,$size,$count,$mbmin;
      $rows++;
      $total_size  += $size;
      $total_count += $count;
      $total_mbmin += $mbmin;
    }
  }

  if ( $rows == 0 )
  {
     &msg("$rows rows found!" );
     exit;
  }

  # Average line
  $ts = "(Average)";
  $avg_size  = int($total_size  / $rows);
  $avg_count = int($total_count / $rows);
  $avg_mbmin = int($total_mbmin / $rows);

  print "------------- ---------- ---------- ----------\n";
  printf "%-13s %10s %10s %10s\n",$ts,$avg_size,$avg_count,$avg_mbmin;

  # Total line
  $ts = "(Total)";
  $tot_size  = $total_size  ;
  $tot_count = $total_count ;
  $tot_mbmin = $total_mbmin ;
  printf "%-13s %10s %10s %10s\n",$ts,$tot_size,$tot_count,$tot_mbmin;

  print "\n";
  &msg("$rows Row(s) found - *: Average of a 24 hour period");
  exit;
}

# --------------------------------------------------------------------
# -az
# --------------------------------------------------------------------
sub az_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-az" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }


  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &msg("Collecting 'before' Sample");
  $sqlstmt = 'select 
    \'X\', 
    value, 
    name
  from 
    v\$sysstat 
  where
    name like \'%user%\' or
    name like \'%db block%\';'; 

  ##print "sqlstmt=$sqlstmt\n";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $value = $words[1];
      $name1 = $words[2];
      $name2 = $words[3];
      $name3 = $words[4];
      if ($name1 eq "user" && $name2 eq "calls")   { $b4_calls   = $value; }
      if ($name1 eq "user" && $name2 eq "commits") { $b4_commits = $value; }
      if ($name1 eq "db" && $name2 eq "block" && $name3 eq "gets"   ) { $b4_gets = $value; }
      if ($name1 eq "db" && $name2 eq "block" && $name3 eq "changes") { $b4_chgs = $value; }
      if ($name1 eq "table" && $name2 eq "scans" && $name3 eq "(long") { $b4_lscans = $value; }
      $b4_time = time;
    }
  }

  #Get latch detail samples
  chomp( @stats = `$P.pl -la | findstr /B /I /C:"library cache"`);
  $b4_lw_lc = pop @stats;
  chomp( @stats = `$P.pl -la | findstr /B /I /C:"cache buffers chains"`);
  $b4_lw_cb = pop @stats;

  &msg("Sleeping 10 seconds");
  sleep 10;

  &msg("Collecting 'after' Sample");

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $value = $words[1];
      $name1 = $words[2];
      $name2 = $words[3];
      $name3 = $words[4];
      if ($name1 eq "user" && $name2 eq "calls")   { $af_calls   = $value; }
      if ($name1 eq "user" && $name2 eq "commits") { $af_commits = $value; }
      if ($name1 eq "db" && $name2 eq "block" && $name3 eq "gets"   ) { $af_gets = $value; }
      if ($name1 eq "db" && $name2 eq "block" && $name3 eq "changes") { $af_chgs = $value; }
      if ($name1 eq "table" && $name2 eq "scans" && $name3 eq "(long") { $af_lscans = $value; }
      $af_time = time;
    }
  }

  #Get latch detail samples
  chomp( @stats = `$P.pl -la | findstr /B /I /C:"library cache"`);
  $af_lw_lc = pop @stats;
  chomp( @stats = `$P.pl -la | findstr /B /I /C:"cache buffers chains"`);
  $af_lw_cb = pop @stats;

  $df_commits = $af_commits - $b4_commits;
  $df_calls   = $af_calls   - $b4_calls  ;
  $df_time    = $af_time    - $b4_time   ;

  $df_gets    = $af_gets    - $b4_gets   ;
  $df_chgs    = $af_chgs    - $b4_chgs   ;

  $df_lscans = $af_lscans   - $b4_lscans   ;

  $rt_calls   = $df_calls / $df_time;
  $rt_commits = $df_commits / $df_time;

  $rt_gets    = int($df_gets / $df_time);
  $rt_chgs    = int($df_chgs / $df_time);

  $rt_lw_lc   = int(($af_lw_lc - $b4_lw_lc) / $df_time);
  $rt_lw_cb   = int(($af_lw_cb - $b4_lw_cb) / $df_time);

  $rt_lscans  = int($df_lscans / $df_time);

  print "rt_gets=$rt_gets b4_gets=$b4_gets af_gets=$af_gets rt_lw_lc=$rt_lw_lc rt_lw_cb=$rt_lw_cb\n";

  print "$P | Time Interval: $df_time seconds\n\n";
  print "                       Commit Count     Call Count \n";
  print "                       ------------   ------------ \n";
  printf "Sample #1:             %12d   %12d \n",$b4_commits,$b4_calls;
  printf "Sample #2:             %12d   %12d \n",$af_commits,$af_calls;
  print "                       ============   ============ \n";
  printf "Difference:            %12d   %12d \n",$df_commits,$df_calls;
  printf "Rate (Per Second):     %12.1f   %12.1f \n",$rt_commits,$rt_calls;

  &msg("Collecting CPU busy sample");
  ( $numcpus, $cpubusy, $cpuidle ) = &cpustats();
  &msg("cpuidle=$cpuidle cpubusy=$cpubusy");

  print "\n\nUNDER CONSTRUCTION!!!\n\n"; exit;

  &msg("Collecting Session Count sample");
  @z = `$P.pl -se | tail -10 |  grep 'Count of' | grep ACTIVE | awk '{ print \$NF }'`;
  chop($ses_act   = $z[0]);
  $ses_act--;   # Subtract the orastat session
  chop($ses_inact = $z[1]);
  $ses_tot = $ses_act + $ses_inact;
  &msg("ses_act=$ses_act ses_tot=$ses_tot ses_inact=$ses_inact");
  
  &msg("Collecting Latch Wait sample");
  chop($ses_lw = `$P.pl -sw | grep 'latch' | wc -l`);
  &msg("ses_lw=$ses_lw");
  
  $usersqlfile = "$oh\admin\AZ.sql";
  $sql_time = "?";
  &msg("Looking for $usersqlfile");
  if (-e $usersqlfile)
  {
    &msg("Collecting user sql time (from $usersqlfile)");
    $tmp = "D:\temp\\${P}_$$.txt";
    system("copy $usersqlfile $tmp");
    system("echo 'exit' >> $tmp");
    chop($sql_time = `timex sqlplus internal \@$tmp 2>&1 | grep ^real | awk '{ print \$2 }'`);
    &msg("sql_time=$sql_time");
    system("rm $tmp");
  }

  # Most recent IO wait
  chop($mr_io_wt = `$P.pl -po | grep '^Average' | awk '{ print \$2 }'`);

  # Hit ratio
  chop($hitr_tot = `$P.pl -pd | grep '^DB_BLOCK' | grep DEFAULT | awk '{ print \$9 }'`);
  chop($hitr_now = `$P.pl -pb | grep '^Current' | awk '{ print \$NF }'`);

  # Sort disk/mem ratio
  chop($sort_disk = `$P.pl -st | grep 'sorts (disk)' | awk '{ print \$4 }'`);
  chop($sort_mem  = `$P.pl -st | grep 'sorts (memory)' | awk '{ print \$4 }'`);
  &msg("sort_disk=$sort_disk sort_mem=$sort_mem");
  $sort_ratio = 100 * ($sort_disk/$sort_mem);
  $sort_ratio = sprintf("%8.5f",$sort_ratio);
  &msg("sort_ratio=$sort_ratio");

  # Sample biggest wait
  chop($_ = `$P.pl -wx | tail -1`);
  @words = split();
  $wait_most_event = $words[0];
  $wait_most_count = $words[1];
  $wait_most_event = "?" if  $wait_most_event =~ /^-/;
  $wait_most_count = "?" if  $wait_most_count =~ /^-/;

  # Get archiver count
  chop($archiver_count = `$P.pl -ap | grep BUSY | wc -l`);

  # Get arc fs full
  chop($arc_full = `$P.pl -l | grep 'Archive Dest FS Full' | awk '{ print \$NF }' | sed s/%//g`);

  chop($date = `date +%Y/%m/%d-%H:%M:%S`);
  &msg("Logging to ${oh}\\adm\\ora_activity_rate_$os.txt");

  $out = sprintf "$date [1] CommitRt: %5.0f DBCallRt: %5.0f CPUBusy%: %5s SQLETime: %5s ArcsBusy: %5s",$rt_commits,$rt_calls,$cpubusy,$sql_time,$archiver_count;
  system("echo '$out' >> ${oh}\\adm\\ora_activity_rate_$os.txt");

  $out = sprintf "$date [2] BufGetRt: %5s BufChgRt: %5s ActivSes: %5s TotalSes: %5s LatWtSes: %5s",$rt_gets,$rt_chgs,$ses_act,$ses_tot,$ses_lw;
  system("echo '$out' >> ${oh}\\adm\\ora_activity_rate_$os.txt");

  $out = sprintf "$date [3] LW_DC_Rt: %5s LW_CB_Rt: %5s MaxWait: $wait_most_event/$wait_most_count",$rt_lw_lc,$rt_lw_cb;
  system("echo '$out' >> ${oh}\\adm\\ora_activity_rate_$os.txt");

  $out = sprintf "$date [4] MR_IO_Wt: %5s HitRaTot: %5s HitRaNow: %5s ArcFSFul: %5s SrtDskRt: %5s LngScnRt: %5s",$mr_io_wt,$hitr_tot,$hitr_now,$arc_full,$sort_ratio, $rt_lscans;
  system("echo '$out' >> ${oh}\\adm\\ora_activity_rate_$os.txt");

  system("echo '.' >> ${oh}\\adm\\ora_activity_rate_$os.txt");

  exit;
} 

sub cpustats
{
    eval { "use Win32::PerfLib"; };
    $processor = 238;
    $proctime = 6;

    $perflib = new Win32::PerfLib();
    $proc_ref0 = {};
    $proc_ref1 = {};
    $perflib->GetObjectList($processor, $proc_ref0);
    sleep 1;
    $perflib->GetObjectList($processor, $proc_ref1);
    $perflib->Close();
    $instance_ref0 = $proc_ref0->{Objects}->{$processor}->{Instances};
    $instance_ref1 = $proc_ref1->{Objects}->{$processor}->{Instances};
    foreach $p (keys %{$instance_ref0})
    {
        $counter_ref0 = $instance_ref0->{$p}->{Counters};
        $counter_ref1 = $instance_ref1->{$p}->{Counters};
        foreach $i (keys %{$counter_ref0})
        {
            next if $instance_ref0->{$p}->{Name} eq "_Total";

            if($counter_ref0->{$i}->{CounterNameTitleIndex} == $proctime)
            {
                $Numerator0 = $counter_ref0->{$i}->{Counter};
                $Denominator0 = $proc_ref0->{PerfTime100nSec};
                $Numerator1 = $counter_ref1->{$i}->{Counter};
                $Denominator1 = $proc_ref1->{PerfTime100nSec};
                $proc_time{$p} = int( ( 1 - ( ( $Numerator1 - $Numerator0 ) /
                                              ( $Denominator1 - $Denominator0 ) ) ) * 100 );
                push(@output, ($proc_time{$p}));
            }
        }
    }

    $summ = 0;
    foreach ( @output ) { $summ += $_; }
    $numcpus = scalar( @output );
    $avgcpu  = $summ / ( scalar( @output ) );
    $avgcpu  = ( $avgcpu >= 0.0 ) ? $avgcpu : 0;
    $avgcpu = sprintf( "%4.02f", $avgcpu );
    $idlecpu = 100.0 - $avgcpu;
    $idlecpu = sprintf( "%4.02f", $idlecpu );
    return( $numcpus, $avgcpu, $idlecpu );
}            

# --------------------------------------------------------------------
# -ba
# --------------------------------------------------------------------
sub ba_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ba" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "\n";
  print "SID      Level      Timestamp             DBSize  Elapsed  Age(Days)\n";
  print "-------- ---------- ------------------- -------- -------- ----------\n";

  $now = time;

  $sqlstmt = 'select \'Y\', SID, BCKLEVEL, TSNUM, TSTEXT, DBSIZE, ELAPSED
    from dbamon.backup_age 
    order by 5;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      $sid     = $words[1];
      $level   = $words[2];
      $tsnum   = $words[3];
      $tstext  = $words[4];
      $dbsize  = $words[5];
      $elapsed = $words[6];
      $age     = ($now - $tsnum) / (60*60*24); 
      printf("%-8s %-10s %-16s %8d %8d %10d \n",$sid,$level,$tstext,$dbsize,$elapsed,$age);
    }
  }
  print "======== ========== =================== ======== ======== ==========\n";
  exit;
}

# --------------------------------------------------------------------
# -bb
# --------------------------------------------------------------------
sub bb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "\n";

  &subEMail( 'border@dba.it.agilent.com', "monitor--$P $arg $host/$ENV{ORACLE_SID}", '' );
  $sqlstmt = '
    select
      \'Z\',
      handle,
      max(to_char(completion_time,\'YYYY/MM/DD-HH24:MI:SS\')),
      max(elapsed_seconds),
      count(distinct media)
    from
      v\$backup_piece
    group by
      handle
    order by
      3 ;
  ';

  print "\n";
  print "Completion Timestamp Piece Handle              Elapsed-Min.   Media Cnt.\n";
  print "-------------------- ------------------------- ------------   ----------\n";
  $bkcount = 0;

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];

    if ($verb eq "Z")
    {
      $handle   = $words[1];
      $comptime = $words[2];
      $mins     = int($words[3] / 60);
      $mediacnt = $words[4];

      printf "%-20s %-25s %12s %12s\n",$comptime, $handle, $mins, $mediacnt;

      $date = substr($comptime,0,10);
      $datetotal{$date} += $mins;

      $bkcount++;
    }
  }
  print "\n";
  &msg("Found $bkcount distinct backup pieces");

  print "\n";
  print "Date       Total Bkup Mins.\n";
  print "---------- ----------------\n";

  foreach (sort keys %datetotal)
  {
    $z = int($datetotal{$_} / 50);
    $banner = "*" x $z;
    printf "%-10s %16s %-50s\n",$_, $datetotal{$_}, $banner;
  }

  exit;
}

# --------------------------------------------------------------------
# -bc
# --------------------------------------------------------------------
sub bc_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bc" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ( $oraver < 8.1 )
  {
      &msg("-bc: Buffer Contents (only works in Oracle 8i+)");
      exit;
  }
  elsif( int( $oraver ) >= 9 )
  {
      &bcRun9();
  }
  else
  {
      $gtot = 0;

      $bp = "DEFAULT"; &bcRun();
      $bp = "KEEP";    &bcRun();
      $bp = "RECYCLE"; &bcRun();

      print "\n";
      &msg("Total Blocks (All Buffers): $gtot");
      &bcRun();
  }

  sub bcRun   
  {
    print "\n";
    &msg("Buffer Pool: $bp");
    $tot = 0;
    $sqlstmt = "
      create or replace view v\\\$buffer_pool_lo_hi_bnum as 
      select name , 
      min(START_BUF#) lo_bnum, 
      max(END_BUF#) hi_bnum, 
      max(END_BUF#)-min(START_BUF#) buffers 
      from v\\\$buffer_pool v, x\\\$kcbwds s 
      where (LO_SETID=SET_ID or HI_SETID=SET_ID) and buffers > 0 
      group by name;

      select \'X\', d.owner, o.name, COUNT(buf#) BLOCKS 
      from 
	obj\\\$ o, 
	x\\\$bh x,
	dba_objects d
      where o.dataobj# = x.obj 
      and x.state != 0 
      and o.owner# !=0 
      and buf# >= (select lo_bnum from v\\\$buffer_pool_lo_hi_bnum where 
      name = \'$bp\' and buffers > 0) 
      and buf# <= (select hi_bnum from v\\\$buffer_pool_lo_hi_bnum where 
      name = \'$bp\' and buffers > 0) 
      and o.obj# = d.object_id
      group by d.owner, o.name
      order by 4; ";

    print "Buffer   Object                                         Blocks\n";
    print "-------- ---------------------------------------- ------------\n";
    &oraSQL($sqlstmt);
    foreach (@oraSQLout)
    {
      @words  = split();
      $verb    = $words[0];
      $owner   = $words[1];
      $table   = $words[2];
      $blocks  = $words[3];

      $tbl = "$owner.$table";

      if ($verb eq "X")
      {
	printf "%-8s %-40s %12s\n",$bp,$tbl,$blocks;
	$tot += $blocks;
	$gtot += $blocks;
      }
    }
    print "======== ======================================== ============\n";
    printf "%-8s %-40s %12s\n","$bp","Total:",$tot;
  } # END OF bcRUN SUBROUTINE

  sub bcRun9
  {
    $tot = 0;

    #---- New way
    $bc_query = "select 
      decode(pd.bp_id,
               1,'KEEP',
               2,'RECYCLE',
               3,'DEFAULT',
               4,'2K SUBCACHE',
               5,'4K SUBCACHE',
               6,'8K SUBCACHE',
               7,'16K SUBCACHE', 
               8,'32K SUBCACHE',
      'UNKNOWN') subcache, 
      bh.owner#,
      bh.object_name,
      bh.blocks 
    from 
      x\\\$kcbwds ds, 
      x\\\$kcbwbpd pd, 
      (select /*+ use_hash(x) */ 
         set_ds,
         o.name object_name,
         o.owner#, 
         count(*) BLOCKS 
       from 
         obj\\\$ o, 
         x\\\$bh x 
       where 
         o.dataobj# = x.obj 
         and x.state !=0 and o.owner# !=0 
       group by 
         set_ds,o.name, o.owner#) bh 
    where 
      ds.set_id >= pd.bp_lo_sid 
      and ds.set_id <= pd.bp_hi_sid 
      and pd.bp_size != 0 
      and ds.addr=bh.set_ds";

    $sqlstmt = "
    WHENEVER SQLERROR CONTINUE;
    DROP TABLE orastat_bc;
    WHENEVER SQLERROR EXIT 1;
    create global temporary table orastat_bc as $bc_query;
    insert into orastat_bc $bc_query;
    --
    select count(*) from orastat_bc;
    select * from orastat_bc;
    select
      'X',
      bc.subcache,
      us.username,
      bc.object_name,
      sum(bc.blocks)
    from
      orastat_bc bc,
      dba_users us
    where
      bc.owner# = us.user_id
    group by 
      bc.subcache,
      us.username,
      bc.object_name
    order by
      2, 5 desc;  
    ";

    print "Pool     Object                                         Blocks\n";
    print "-------- ---------------------------------------- ------------\n";
    &oraSQL($sqlstmt);
    foreach (@oraSQLout)
    {
      #print;
      #print if /^ORA/;
      @words  = split();
      $verb    = $words[0];
      $pool    = $words[1];
      $owner   = $words[2];
      $table   = $words[3];
      $blocks  = $words[4];

      $tbl = "$owner.$table";

      if ($verb eq "X")
      {
	printf "%-8s %-40s %12s\n",$pool,,$tbl,$blocks;
	$tot += $blocks;
	$gtot += $blocks;
      }
    }
    print "======== ======================================== ============\n";
    printf "%-8s %-40s %12s\n","Total:","",$tot;
  } # END OF bcRUN9 SUBROUTINE

  exit;
}

# --------------------------------------------------------------------
# -bd
# --------------------------------------------------------------------
sub bd_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bd" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    alter session set sort_area_size          = 100000000;
    alter session set sort_area_retained_size = 100000000;
    select
      \'X\',
      ts.name,
      df.name,
      translate(ba.status,\' \',\'_\')
    from
      v\$backup ba,
      v\$datafile df,
      v\$tablespace ts
    where
      ba.file# = df.file#
      and df.ts# = ts.ts#
    order by
      2, 3;';

  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "BackupStatus   Tablespace           Datafile \n";
  print "-------------- -------------------- ------------------------------------------------\n";

  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $ts = shift @words;
      $df = shift @words;
      $st = shift @words;
  
      printf "%-14s %-20s %-40s\n",$st,$ts,$df;
      $total++;
    }
  }
  print "\n";
  &msg("Total of $total Datafiles found in v\$backup ");
}

# --------------------------------------------------------------------
# -bi 
# --------------------------------------------------------------------
sub bi_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bi" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
     select \'X\',
            p.handle,
            p.media,
            to_char(p.completion_time,\'YYYY/MM/DD-HH24:MI:SS\'),
            decode(s.backup_type,\'L\',\'ARC\',\'I\',\'LVL1\',\'D\',\'LVL0\'),
            p.set_stamp 
            from
	        v\$backup_set s, 
	        v\$backup_piece p
	    where p.set_count=s.set_count
	    and p.deleted != \'YES\'
	    order by s.completion_time desc;';

     $cnt=-1; 
  
     &oraSQL($sqlstmt);
     foreach (@oraSQLout)
     {
        @words = split;
        $verb = $words[0];

        if ($verb eq "X")
        {
           $bkp_piece    = $words[1];
           $media_name   = $words[2];
           $end_date     = $words[3];
           $bkp_type     = $words[4];
            
           $cnt+=1;
           $b_pieces[$cnt] = $bkp_piece; 
           $b_info{$bkp_piece}{"media_name"} = $media_name; 
           $b_info{$bkp_piece}{"end_date"} = $end_date; 
           $b_info{$bkp_piece}{"bkp_type"} = $bkp_type; 
        }
     }

     if (uc($arg2) eq "SINCE")
     {
	$retry_cnt = 0;
	do 
	{
           $invalid_date = "false";
	   print ("\n");
           print ("Enter PIT (YYYY/MM/DD-HH24:MI:SS)\n"); 
           print ("          (ex: 2001/08/12-18:05:00): "); 
           $pit = <STDIN>;
           chop($pit);
           if (substr($pit,4,1) ne "/" ||
               substr($pit,7,1) ne "/" ||
               substr($pit,10,1) ne "-" ||
               substr($pit,13,1) ne ":" ||
               substr($pit,16,1) ne ":") 
           {
              print "\nERROR: Invalid date format.\n";
              $invalid_date = "true";
	      $retry_cnt += 1;
           }
	   if ($retry_cnt == 3)
	   {
              print "       Unable to continue after $retry_cnt attempts.\n\n";
	      exit 1;
	   }
         } until ($invalid_date ne "true");
     }
     elsif (uc($arg2) eq "BETWEEN")
     {
	$retry_cnt = 0;
	do
	{
	   $invalid_date = "false";
	   print ("\n");
           print ("Enter end time (YYYY/MM/DD-HH24:MI:SS)\n"); 
           print ("               (ex: 2001/08/12-18:05:00): "); 
           $e_time = <STDIN>;
           chop($e_time);
           if (substr($e_time,4,1) ne "/" ||
               substr($e_time,7,1) ne "/" ||
               substr($e_time,10,1) ne "-" ||
               substr($e_time,13,1) ne ":" ||
               substr($e_time,16,1) ne ":") 
           {
              print "\nERROR: Invalid date format.\n";
              $invalid_date = "true";
	      $retry_cnt += 1;
           }
	   if ($retry_cnt == 3)
	   {
              print "       Unable to continue after $retry_cnt attempts.\n\n";
	      exit 1;
	   }
         } until ($invalid_date ne "true");

	$retry_cnt = 0;
	do
	{
	   $invalid_date = "false";
	   print ("\n");
           print ("Enter start time (YYYY/MM/DD-HH24:MI:SS)\n"); 
           print ("                 (ex: 2001/08/10-12:05:00): "); 
           $s_time = <STDIN>;
           chop($s_time);
           if (substr($s_time,4,1) ne "/" ||
               substr($s_time,7,1) ne "/" ||
               substr($s_time,10,1) ne "-" ||
               substr($s_time,13,1) ne ":" ||
               substr($s_time,16,1) ne ":") 
           {
              print "\nERROR: Invalid date format.\n";
              $invalid_date = "true";
	      $retry_cnt += 1;
           }
	   if ($retry_cnt == 3)
	   {
              print "       Unable to continue after $retry_cnt attempts.\n\n";
	      exit 1;
	   }
         } until ($invalid_date ne "true");

	 $retry_cnt = 0;
	 if ($stime gt $e_time)
	 {
	    $temp = $s_time;
	    $s_time = $e_time;
	    $e_time = $temp;
	 }
     }

     printf("\n");
     $end_date = "Date/Time";
     $bkp_type = "Type";
     $media_name = "Tape#";
     $bkp_piece = "Backup Piece";
     printf ("%-20s %-8s %-10s %-30s \n",$end_date,$bkp_type,$media_name,$bkp_piece); 
     for ($i=0;$i<=68;$i++)
     {
        printf ("-");
     }
     printf("\n");

     $cnt=-1;
     #$bt_found = "ARC";
     #$print_cnt = 0;
     #$pit_cnt = -1;
     foreach $bkp_piece (@b_pieces)
     {
        $cnt+=1;
        $media_name = $b_info{$bkp_piece}{"media_name"}; 
        $end_date   = $b_info{$bkp_piece}{"end_date"}; 
        $bkp_type   = $b_info{$bkp_piece}{"bkp_type"}; 

        if (uc($arg2) eq "ALL" || $arg2 eq "")
        {
           printf ("%-20s %-8s %-10s %-30s \n",$end_date,$bkp_type,$media_name,$bkp_piece);
        }
        elsif (uc($arg2) eq "SINCE")
        {
           if ($pit le $end_date &&
               $pit gt ${b_info{$bkp_piece + 1}{"end_date"}})
           { 
             printf ("%-20s %-8s %-10s %-30s \n",$end_date,$bkp_type,$media_name,$bkp_piece);
           }
        }
        elsif (uc($arg2) eq "BETWEEN")
        {
           if ($end_date le $e_time &&
               $end_date ge $s_time)
           { 
             printf ("%-20s %-8s %-10s %-30s \n",$end_date,$bkp_type,$media_name,$bkp_piece);
           }
        }
	else
	{
	   print "\nERROR: Invalid option.\n\n";
	   exit 1;
	}
     }
     print "\n"; 
     exit;
} 

# --------------------------------------------------------------------
# -bm
# --------------------------------------------------------------------
sub bm_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bm" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &msg("NOTE: Tablespaces displayed here have AT LEAST ONE datafile in Backup Mode, but");
  &msg("      not neccessarily all datafiles are in Backup Mode");

  $rows = 0;

  $sqlstmt = '
    alter session set sort_area_size          = 100000000;
    alter session set sort_area_retained_size = 100000000;
    select
      \'X\',
      ts.name,
      df.name
    from
      v\$backup ba,
      v\$datafile df,
      v\$tablespace ts
    where
      ba.status = \'ACTIVE\'
      and ba.file# = df.file#
      and df.ts# = ts.ts#
    order by
      2, 3;';

  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "    Tablespace           Datafile \n";
  print "    -------------------- ----------------------------------------\n";

  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $ts = shift @words;
      $df = shift @words;
  
      printf "TS: %-20s %40s \n",$ts,$df;
      $total++;
    }
  }
  print "\n";
  &msg("Total of $total Datafiles in Backup Mode ");
  exit;
}

# --------------------------------------------------------------------
# -bo
# --------------------------------------------------------------------
sub bo_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bo" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  #---- #1 get blocksize
  $sqlstmt = 'select \'X\', value from v\$parameter where name = \'db_block_size\';';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $db_block_size = shift @words;
    }
  }

  #---- #2 get blocksize
  $sqlstmt = 'select \'X\', count(*) from v\$bh;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $buffer_total_size = shift @words;
    }
  }
  ##print "bts=$buffer_total_size\n";

  #---- #1 group by status
  $sqlstmt = "
    SELECT 
      'X',
      status, 
      count(*) 
    FROM 
      v\\\$bh
    group by
      'X',
      status
    order by
      3 desc;
      ";

  print "\n";
  &msg("-bo: Block Status");
  print "\n";
  print "Block Status        Block Count       Size (mB)      % of Total\n";
  print "--------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_status   = shift @words;
      $bh_count    = shift @words;

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-15s %15s %15.2f %15.2f%%\n","Status=$bh_status",$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
    }
  }
  print "=============== =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-15s %15s %15.2f\n","Total:",$total_count, $total_mb;

  return if $arg2 eq "1";

  #---- #2 group by dirty
  $sqlstmt = "
    SELECT 
      'X',
      dirty, 
      count(*) 
    FROM 
      v\\\$bh
    group by
      'X',
      dirty
    order by
      3 desc;
      ";

  print "\n";
  &msg("-bo: Dirty Blocks");
  print "\n";
  print "Buffer Dirty        Block Count       Size (mB)      % of Total\n";
  print "--------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_status   = shift @words;
      $bh_count    = shift @words;

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-15s %15s %15.2f %15.2f%%\n","Dirty=$bh_status",$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
    }
  }
  print "=============== =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-15s %15s %15.2f\n","Total:",$total_count, $total_mb;

  #---- #3 group by temp 
  $sqlstmt = "
    SELECT 
      'X',
      temp, 
      count(*) 
    FROM 
      v\\\$bh
    group by
      'X',
      temp
    order by
      3 desc;
      ";

  print "\n";
  &msg("-bo: Temp Blocks");
  print "\n";
  print "Buffer TEMP         Block Count       Size (mB)      % of Total\n";
  print "--------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_status   = shift @words;
      $bh_count    = shift @words;

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-15s %15s %15.2f %15.2f%%\n","Temp=$bh_status",$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
    }
  }
  print "=============== =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-15s %15s %15.2f\n","Total:",$total_count, $total_mb;

  #---- #4 group by TS 
  $sqlstmt = "
    SELECT 
      'X',
      ts.name, 
      count(*) 
    FROM 
      v\\\$bh bh,
      v\\\$tablespace ts
    where
      bh.ts# = ts.ts# 
    group by
      'X',
      ts.name
    order by
      3 desc;
      ";

  print "\n";
  &msg("-bo: Blocks By Tablespace - Top 10");
  print "\n";
  print "Tablespace          Block Count       Size (mB)      % of Total\n";
  print "--------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;
  $rows = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_ts       = shift @words;
      $bh_count    = shift @words;

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-15s %15s %15.2f %15.2f%%\n",$bh_ts,$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
 
      last if $rows >= 10;
    }
  }
  print "=============== =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-15s %15s %15.2f\n","Total:",$total_count, $total_mb;


  #---- #5 group by DF 
  $sqlstmt = "
    SELECT 
      'X',
      df.name, 
      count(*) 
    FROM 
      v\\\$bh bh,
      v\\\$datafile df
    where
      bh.file# = df.file# 
    group by
      'X',
      df.name
    order by
      3 desc;
      ";

  print "\n";
  &msg("-bo: Blocks By Datafile - Top 10");
  print "\n";
  print "Datafile                                                    Block Count       Size (mB)      % of Total\n";
  print "------------------------------------------------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;
  $rows = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_df       = shift @words;
      $bh_count    = shift @words;

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-55s %15s %15.2f %15.2f%%\n",$bh_df,$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
 
      last if $rows >= 10;
    }
  }
  print "======================================================= =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-55s %15s %15.2f\n","Total:",$total_count, $total_mb;


  #---- #6 group by object 
  $sqlstmt = "
    SELECT 
      'X',
      ob.owner, 
      ob.object_name, 
      ob.object_type, 
      count(*) 
    FROM 
      v\\\$bh bh,
      dba_objects ob
    where
      bh.objd = ob.object_id
    group by
      'X',
      ob.owner, 
      ob.object_name, 
      ob.object_type 
    order by
      5 desc;
      ";

  print "\n";
  &msg("-bo: Blocks By Object - Top 25");
  print "\n";
  print "Object                                                      Block Count       Size (mB)      % of Total\n";
  print "------------------------------------------------------- --------------- --------------- ---------------\n";

  &oraSQL($sqlstmt);

  $total_count = 0;
  $rows = 0;

  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $bh_owner    = shift @words;
      $bh_name     = shift @words;
      $bh_type     = shift @words;
      $bh_count    = shift @words;

      $this_object = "$bh_owner.$bh_name ($bh_type)";

      $bh_mb  = ($db_block_size * $bh_count) / 1000000;
      $bh_pct = ($bh_count / $buffer_total_size) * 100 if $buffer_total_size > 0;

      printf "%-55s %15s %15.2f %15.2f%%\n",$this_object,$bh_count,$bh_mb,$bh_pct;
      
      $total_count += $bh_count;
 
      last if $rows >= 25;
    }
  }
  print "======================================================= =============== ===============\n";
  $total_mb = ($db_block_size * $total_count) / 1000000;
  printf "%-55s %15s %15.2f\n","Total:",$total_count, $total_mb;

  exit;
}

# --------------------------------------------------------------------
# bp
# --------------------------------------------------------------------
sub bp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
    to_char(logon_time, \'DD-MON-YY\'),
    to_char(logon_time, \'HH24:MM:SS\'),
    sid,serial#,process,status,module
    from v\$session
    where module like \'%backup%\';';

  print "Start Date Start Time  SID    Serial#  Process  Status     Module\n";
  print "---------- ---------- ------ --------- -------- ---------- --------------------------\n";

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    $date    = $words[1];
    $time    = $words[2];
    $sid     = $words[3];
    $serial  = $words[4];
    $process = $words[5];

    if ($process eq "ACTIVE" | $process eq "INACTIVE") 
    {
      $process = ".";
      $status = $words[5];
    }
    else {$status  = $words[6];}
    $module  = $words[7] . " " . $words[8] . " " . $words[9] . " " . $words[10];
    if ($verb eq "X")
    {
      printf "%9s %9s %8s %9s %8s %10s %s\n", $date, $time, $sid, $serial, $process, $status, $module;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -br 
# --------------------------------------------------------------------
sub br_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-br" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', max(recid) from v\$backup_set where incremental_level = \'0\';';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb = $words[0];
    if ($verb eq "X")
    {
      $l0_recid = $words[1];
    }
  }
  &msg("Most recent L0 recid=$l0_recid");

  $sqlstmt = "select \'X\', 
    to_char(completion_time,\'YYYY/MM/DD-HH24:MI:SS\')
    from v\\\$backup_set where recid = $l0_recid;";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb = $words[0];
    if ($verb eq "X")
    {
      $l0_ts = $words[1];
    }
  }
  &msg("Most recent L0 timestamp=$l0_ts");
  
  $sqlstmt = "select \'X\', 
    to_char(completion_time,\'YYYY/MM/DD-HH24:MI:SS\'),
    media
    from v\\\$backup_piece where recid >= $l0_recid;";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb = $words[0];
    if ($verb eq "X")
    {
      $ts = $words[1];
      $tp = $words[2];
      ##print "$ts $tp\n";
      if (index($tplist,$tp) == -1)
      { 
        $tplist .= "$tp,";
        $tphash{$tp} = 1;
      }
    }
  }
  &msg("List of backup media that have been written to since the most recent LVL0:");
  @sorted = sort keys %tphash;
  foreach (@sorted) 
  {
    print "Media: $_\n";
    $cnt++;
  }
  &msg("Found $cnt tape(s)");

  exit;
}

# --------------------------------------------------------------------
# -bs
# --------------------------------------------------------------------
sub bs_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-bs" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    alter session set sort_area_size          = 100000000;
    alter session set sort_area_retained_size = 100000000;
    SELECT 
      \'X\',
      df.name
    from
      v\$datafile df,
      v\$backup bk
    where
      bk.file# = df.file#
      and bk.status = \'ACTIVE\'; ';

  print "\n";
  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $dfname   = shift @words;
      $cmd = "ALTER DATABASE DATAFILE '$dfname' END BACKUP;";

      print "$cmd\n";
      $total++;
    }
  }
  print "\n";
  &msg("Created $total ALTER DATABASE statements");
  exit;
}

# --------------------------------------------------------------------
# -c
# --------------------------------------------------------------------
sub c_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-c" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $init_file = "$oh\\database\\init$os.ora";
  $i_file    = undef;
  print "Contents of init.ora file: $init_file\n\n";
  system( "type $init_file" );

  chomp( $line = `findstr /I /C:ifile $init_file` );
  if ( defined $line and length( $line ) > 0 )
  {
      @words = split( /\s+/, $line );
      print "#**** Contents of ifile: $ifile *****\n\n";
      $ifile = pop @words;
      system( "type $ifile" );
      print "#**** End of ifile: $ifile *****\n";
  }
  else
  {
      print "IFILE parameter Not specified within $init_file\n";
  }

  exit;
}

# --------------------------------------------------------------------
# -cb
# --------------------------------------------------------------------
sub cb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $now       = time();                                                # V02.34
  $cb_binary = 'D:\\temp\\orastat_cfbinary_' . "${os}" . '_' . "${$}.dbf";
  unlink "$cb_binary" if -e $cb_binary;                               # V02.34
  $bug_cbtrcdir = "$oh\\rdbms\\log";                                  # V02.34
                                                                      # V02.34
  sub cbSubLF # Return Latest File In Directory Or Undef              # V02.34
  {                                                                   # V02.34
    my( $dir, $suffix, $patt ) = @_;                                  # V02.34
                                                                      # V02.34
    chdir( $dir )                                                     # V02.34
    or                                                                # V02.34
    die "Can't chdir to $dir: $!";                                    # V02.34
                                                                      # V02.34
    opendir( DIR, $dir )                                              # V02.34
    or                                                                # V02.34
    die "Can't opendir $dir: $!";                                     # V02.34
    # Sort Into Descending Latest Modified Date Order                 # V02.34
    @sorted = sort { ( stat $b )[ 9 ] <=> ( stat $b )[ 9 ] } grep !/^\.\.?$/, readdir( DIR ); # V02.34
                                                                      # V02.34
    closedir( DIR)                                                    # V02.34
    or                                                                # V02.34
    die "Can't closedir $dir: $!";                                    # V02.34
                                                                      # V02.34
    $latest_file = undef;                                             # V02.34
    foreach ( @sorted )                                               # V02.34
    {                                                                 # V02.34
        next if -d $_ or $_ !~ /$suffix$/i;                           # V02.34
        if ( $_ =~ /$patt/ ) { $latest_file = "$dir\\$_"; last; }     # V02.34
    }                                                                 # V02.34
                                                                      # V02.34
    return $latest_file;                                              # V02.34
  } # End Of Internal Subroutine cbSubLF                              # V02.34
                                                                      # V02.34
  #                                                                   # V02.34
  ## Get SPID, USER_DUMP_DEST and then do the needful                 # V02.34
  #                                                                   # V02.34
  $sqlstmt = "
    SELECT
      'X', spid
    FROM
      v\\\$process a, v\\\$session b
    WHERE
      a.addr = b.paddr
      AND b.sid = (SELECT MIN(sid) FROM v\\\$mystat);
    SELECT 'Z', value 
    FROM
      v\\\$parameter
    WHERE name = \'user_dump_dest\';
    ALTER system checkpoint;
    ALTER database backup controlfile to trace;
    ALTER database backup controlfile to \'$cb_binary\';
  ";

  &oraSQLold( $sqlstmt, $os_conn );                                   # V02.34
  &subEMerr( 'larry_rush@agilent.com',                                # V02.38
             "ERROR--$P $arg $host/$ENV{ORACLE_SID} at line# " .      # V02.38
             __LINE__,                                                # V02.38
             join( "\n", @oraSQLout ) )                               # V02.38
      if ( $oraCMDrc != 0 );                                          # V02.38
                                                                      # V02.34
  $spid = undef;                                                      # V02.34
  $udd  = undef;                                                      # V02.34
  foreach ( @oraSQLout )                                              # V02.34
  {                                                                   # V02.34
      if    ( $_ =~ /^X/ ) { ( $verb, $spid ) = split(); }            # V02.34
      elsif ( $_ =~ /^Z/ ) { ( $verb, $udd )  = split(); };           # V02.34
  }                                                                   # V02.34
                                                                      # V02.34
  &subEMerr( 'larry_rush@agilent.com',                                # V02.38
             "ERROR--$P $arg $host/$ENV{ORACLE_SID} at line# " .      # V02.38
             __LINE__,                                                # V02.38
             join( "\n", @oraSQLout ) )                               # V02.38
      if not defined $spid or not defined $udd;                       # V02.38
                                                                      # V02.34
  &subEMerr( 'larry_rush@agilent.com',                                # V02.38
             "ERROR--$P $arg $host/$ENV{ORACLE_SID} at line# " .      # V02.38
             __LINE__,                                                # V02.38
             "USER_DUMP_DEST: $udd Does Not Exist!" )                 # V02.38
      if ! -d $udd;                                                   # V02.38
                                                                      # V02.34
  if ( -e $cb_binary )                                                # V02.34
  {                                                                   # V02.34
    print "Controlfile_Binary_Backup was successful to: $cb_binary\n";# V02.34
    @words = split( /\\/, $cb_binary );                               # V02.38
    $bf    = pop @words;                                              # V02.38
    chomp( @list = `dir /N /O:-D $cb_binary` );                       # V02.38
    map{ ( $_ =~ /$bf$/i ) ? print "$_\n" : undef } @list;            # V02.38
  }                                                                   # V02.34
  else                                                                # V02.34
  {                                                                   # V02.34
    &msg( "ERROR: Could not find the binary backup file: " .          # V02.34
          "$cb_binary" );                                             # V02.34
  }                                                                   # V02.34
                                                                      # V02.34
  $latest_trcfile = &cbSubLF( $udd, ".TRC", $spid );                  # V02.34
  $latest_trcfile = &cbSubLF( $bug_cbtrcdir, ".TRC", $spid )          # V02.34
      if ( not defined $latest_trcfile );                             # V02.34
                                                                      # V02.34
  if ( defined $latest_trcfile &&                                     # V02.34
       ( stat $latest_trcfile )[ 9 ] > $now )                         # V02.34
  {                                                                   # V02.34
    print "Controlfile_Trace_Backup was successful to: " .            # V02.34
          "$latest_trcfile\n";                                        # V02.34
    @words = split( /\\/, $latest_trcfile );                          # V02.38
    $tf    = pop @words;                                              # V02.38
    chomp( @list = `dir /N /O:-D $latest_trcfile` );                  # V02.38
    map{ ( $_ =~ /\.trc$/i ) ? print "$_\n" : undef } @list;          # V02.38
  }                                                                   # V02.34
  else                                                                # V02.34
  {                                                                   # V02.34
    &msg( "ERROR: Could not find the trace backup file" );            # V02.34
  }                                                                   # V02.34
                                                                      # V02.34
  &msg("Done");
  exit;
}

# --------------------------------------------------------------------
# -ce
# --------------------------------------------------------------------
sub ce_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ce" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $editor_cmd = ( defined $ENV{"EDITOR"} ) ? $ENV{"EDITOR"} : $default_editor;
  system("$editor_cmd $initfile");
  exit;
}

# --------------------------------------------------------------------
# -cf
# --------------------------------------------------------------------
sub cf_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cf" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
       exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', name, status from v\$controlfile;';

  print "Name                                               Status\n";
  print "-------------------------------------------------- --------------------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb = $words[0];
    $name = $words[1];
    $stat = $words[2];
    if ($verb eq "X")
    {
      printf "%-50s %-20s\n",$name,$status;
    }
  }
  exit;
}

#------------------------------------------------------------------------------
# -cfbl
#------------------------------------------------------------------------------
sub cfbl_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cfbl" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
       exit 1;
  }
  
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
 
  $sqlstmt = "select 'X', VALUE from v\$parameter where name='user_dump_dest';";

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words = split();
    next if $words[ 0 ] ne "X";

    $cfdir = $words[ 1 ];
  }

  &msg( "Searching CFDIR $cfdir" );
  opendir(CFDIR,"$cfdir");
  chomp( @filenames = sort readdir(CFDIR) );
  closedir(CFDIR);
  print "\n";
  print "Control File To Trace File Name                        CF Backup  Create Time         Age(Secs) FileSize\n";
  print "------------------------------------------------------ ------------------------------ --------- ---------\n";
  $numlogs = 0;
  $totsize = 0;
  $age     = -1;

  foreach $file (@filenames)
  {
    next if $file =~ /^\./ or $file !~ /.trc$/i;

    $fullfilename = "$cfdir" . '\\' . "$file";
    if (system("findstr /I /C:$os $fullfilename > nul") == 0)
    {
        $cfmtime = ( stat "$fullfilename" ) [9];
        $cfsize  = ( stat "$fullfilename" ) [7];
        $cfctime = ( stat "$fullfilename" ) [10];
        $cfctime_fmt = &timeToText("$cfctime");

        $flag = "*Hit*" if $cfctime > $arg2 && $arg2;

        $age = time - $cfmtime;
        $oldest = $age if $age > $oldest;
        printf "%-54s %10s %19s %9s %9s %-5s\n","$fullfilename",$cfctime,$cfctime_fmt,$age,$cfsize,$flag;
        $numlogs++;
     }
  }
  print "\n";
  &msg("Found $numlogs backups");
  exit;
}

# --------------------------------------------------------------------
# -ck
# --------------------------------------------------------------------
sub ck_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ck" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', 
    min(sysdate - CHECKPOINT_TIME) from v\$datafile_header;';

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb = $words[0];
    if ($verb eq "X")
    {
      $minago = $words[1] * 24 * 60;
      $minago = '.' if $minago eq "";
      printf "\nLastCheckpoint: %10.2f (Minutes Ago)\n",
        $minago;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -cm
# --------------------------------------------------------------------
sub cm_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cm" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "Parm                                     IsDefault IsSesMod  IsSysMod  IsMod     IsAdj     Value\n";
  print "---------------------------------------- --------- --------- --------- --------- --------- ------------------------------\n";

  $sqlstmt = 'select
    \'X\', 
    name, 
    translate(value,\' \',\'_\'),
    isdefault,
    isses_modifiable,
    issys_modifiable,
    ismodified,
    isadjusted
  from 
    v\$parameter 
  order by 
    2;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    @words  = split();
    $verb   = $words[0];
    $name   = $words[1];
    $value  = $words[2];
    $isdef  = $words[3];
    $isses  = $words[4];
    $issys  = $words[5];
    $ismod  = $words[6];
    $isadj  = $words[7];

    if ($verb eq "X")
    {
      printf "%-40s %-9s %-9s %-9s %-9s %-9s %-30s\n",
        $name,
        $isdef,
        $isses,
        $issys,
        $ismod,
        $isadj,
        $value;
    }
  }
}

# --------------------------------------------------------------------
# -cp
# --------------------------------------------------------------------
sub cp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "Parm                                     Value                               Desc.\n";
  print "---------------------------------------- ----------------------------------- ------------------------------\n";

  $sqlstmt = 'select \'X\', name, value, \'()\', description from v\$parameter order by 2;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    $name   = $words[1];
    $value  = $words[2];
    $verb2  = $words[3];

    $ndx    = index($_,"()");
    $desc   = substr($_,$ndx+3,40);

    if ($verb eq "X")
    {
      printf "%-40s %-35s %-20s\n",$name,$value,$desc;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -ct
# --------------------------------------------------------------------
sub ct_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ct" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\',
	      a.tablespace_name as tablespace_name,
	      d.contents,
	      a.sum_bytes as bytes_avail,
	      a.max_id,
	      nvl(b.sum_bytes,0) as bytes_used,
	      nvl(c.sum_bytes,0) as bytes_free,
	      round(nvl(b.sum_bytes,0.00) / a.sum_bytes * 100.0, 2) as pct_used,
	      d.status
	 from
	      ( select
		  tablespace_name,
		  sum(bytes) sum_bytes,
		  count(1) max_id
		from
		  dba_data_files
		group by
		  tablespace_name
	      ) a,
	      ( select
		  tablespace_name,
		  sum(bytes) sum_bytes
		from
		  dba_extents
		group by
		  tablespace_name
	      ) b,
	      ( select
		  tablespace_name,
		  sum(bytes) sum_bytes
		from
		  dba_free_space
		group by
		  tablespace_name
	      ) c,
	      dba_tablespaces d
	where a.tablespace_name = b.tablespace_name (+)
	  and a.tablespace_name = c.tablespace_name (+)
	  and a.tablespace_name = d.tablespace_name (+)
      and d.contents = \'TEMPORARY\';';

  print "Tablespace           Type Status   NumDF  Size(MB)  Free(MB)  Used(MB)   Pct\n";
  print "-------------------- ---- -------- ----- --------- --------- --------- -----\n";

  $ts_count = -1;
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $ts_count+=1;
      $found = 1;
      $ts = $words[1];
      $type = substr($words[2],0,4);
      $size = $words[3]/1048576;
      $max = $words[4];
      $used = $words[5]/1048576;
      $free = $words[6]/1048576;
      $pct = $words[7];
      $status = $words[8];
      $temp_ts[$ts_count] = $ts;
      printf("%-20s %-4s %-8s %5d %9d %9d %9d %5.1f\n",$ts,$type,$status,$max,$size,$free,$used,$pct);
      $tot_size += $size;
      $tot_free += $free;
      $tot_used += $used;
      $tot_df   += $max;
    }
  }

  if (!$found)
  {
    &msg("Something is wrong - query returned no rows");
    exit 1;
  }
  
  ####   !echo 'Issuing coalesce statements;';
  foreach $temp_tablespace (@temp_ts)
  {
      print "\nCoalescing $temp_tablespace tablespace...\n";
      open( SO,">$sqlin" );
      print SO ( "$os_conn
                 alter tablespace $temp_tablespace default storage (pctincrease 1);
                 alter tablespace $temp_tablespace coalesce;
                 alter tablespace $temp_tablespace coalesce;
                 alter tablespace $temp_tablespace coalesce;
                 alter tablespace $temp_tablespace coalesce;
                 alter tablespace $temp_tablespace default storage (pctincrease 0);
                 exit;" );
      close( SO );
      system("$os_scmd \@$sqlin 2>&1" );
  }

  print "\n";
  print "Tablespace           Type Status   NumDF  Size(MB)  Free(MB)  Used(MB)   Pct\n";
  print "-------------------- ---- -------- ----- --------- --------- --------- -----\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $found = 1;
      $ts = $words[1];
      $type = substr($words[2],0,4);
      $size = $words[3]/1048576;
      $max = $words[4];
      $used = $words[5]/1048576;
      $free = $words[6]/1048576;
      $pct = $words[7];
      $status = $words[8];
      printf("%-20s %-4s %-8s %5d %9d %9d %9d %5.1f\n",$ts,$type,$status,$max,$size,$free,$used,$pct);
      $tot_size += $size;
      $tot_free += $free;
      $tot_used += $used;
      $tot_df   += $max;
    }

    if (!$found)
    {
      print "$P | Something is wrong - query returned no rows\n";
      exit 1;
    }
  }
  exit; 
}

# --------------------------------------------------------------------
# -cv
# --------------------------------------------------------------------
sub cv_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-cv" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select \'X\','alter ' || object_type || ' ' || owner ||
              '.' || object_name || ' compile' 
              from dba_objects where status='INVALID';";

  &oraSQL($sqlstmt);
  #$tbl_cnt=-1;
  print "\n";
  foreach (@oraSQLout)
  {
    if ($_ =~ "X")
    {
      s/X//;
      s/\n//;

      if (/PACKAGE BODY/)
      {
        s/PACKAGE BODY/PACKAGE/;
        s/;$/ body;/;
      }
      print "$_;\n";
    }
  }
  print "\n";
  exit;
}

# ------------------------------------------------------------------------------
# -dc
# ------------------------------------------------------------------------------
sub dc_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-dc" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select \'X\', tablespace_name, file_name, bytes, file_id
              from dba_data_files
              order by 1, 2;";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $df = $words[2];
      $newdf = $df;
      $newdf =~ s/$os/NEWDBNAME/g;

      @words = split(/\\/,$df);
      $lastnode = $words[$#words];
      ##print "1=@words[1] #=$#words\n";
      ##print "lastnode=$lastnode\n";
      $newdir = $df;
      $newdir =~ s/$lastnode//g;
      $newdir =~ s/$os/NEWDBNAME/g;
      chop($newdir);

      printf("md  %s\n", $newdir);
      printf("copy %s      %s\n",$df, $newdf);
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -dd
# --------------------------------------------------------------------
sub dd_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-dd" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
  
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - Object=$arg3.$arg4 Type=$arg2");

  if ( not defined $arg2 and not defined $arg3 and not defined $arg4 )
  {
      &msg( "\nReguired Args Not Specified!" );
      exit;
  }

  if ($oraver < 9)
  {
    &msg("orastat -dd Will only work with Oracle 9i+");
    exit;
  }
  
  $arg2_uc = uc $arg2;
  $arg3_uc = uc $arg3;
  $arg4_uc = uc $arg4;

  $rows = 0;
  print "\n";

  $sqlstmt = "
    SELECT 
      \'X\',
      dbms_metadata.get_ddl ('$arg2_uc','$arg4_uc','$arg3_uc')
    FROM 
      dual;";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if $printflag && /^ /;
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $printflag = 1;
      $rows++;
    }
  }
  print "\n";
  &msg("Object $arg3.$arg4 type=$arg2 Not Found") if !$rows;
  exit;
}

# ------------------------------------------------------------------------------
# -de
# ------------------------------------------------------------------------------
sub de_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-de" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
  select \'W\', file_name, tablespace_name from dba_temp_files 
  union all
  select \'X\', file_name, tablespace_name from dba_data_files 
  union all
  select \'Y\', member, \' \' from v\$logfile
  union all
  select \'Z\', name, \' \' from v\$controlfile
  order by 1, 2;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];
    $ts      = $words[2];
    if ($verb eq "W")
    {
      $f = $words[1];
      print "CMD $f  # Temp File (TS=$ts)\n";
    }
    if ($verb eq "X")
    {
      $f = $words[1];
      print "CMD $f  # DF (TS=$ts)\n";
    }
    if ($verb eq "Y")
    {
      $f = $words[1];
      print "CMD $f  # Online Redo Log\n";
    }
    if ($verb eq "Z")
    {
      $f = $words[1];
      print "CMD $f  # Control File\n";
    }
  }
  exit;
}

# ------------------------------------------------------------------------------
# -df 
# ------------------------------------------------------------------------------
sub df_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-df" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";
  printf("%-16s %-4s %-55s %-9s %-9s %-9s\n",
         "Tablespace", "DF", "Datafile", "Datafile", "Datafile", "FS Free" );
  printf("%-16s %-4s %-55s %-9s %-9s %-9s\n",
         "Name      ", "Num", "Path", "Status", "Size(MB)", "Space(MB)" );
  printf("%-16s %-4s %-55s %-9s %-9s %-9s\n",
         "-" x 16, "-" x 4, "-" x 52, "-" x 9, "-" x 9, "-" x 9 );

  $sqlstmt = "select \'X\',
  tablespace_name,
  file_name,
  nvl(bytes,0),
  file_id,
  nvl(status,\'?\')
from dba_data_files order by 2, 5;";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $dfcount++;
      $status = "";
      $ts = $words[1];
      $df = $words[2];
      $size = $words[3]/1048576;
      $num = $words[4];
      $status = $words[5];

    ##print "num=$num \n";

      if (-c $df)
      {
        $fs = "(Raw)";
      }
      else
      {
        if ($arg2 eq "FSCHECK")
        {
          $df_drive = substr($df,0,2);
          @words = `dir /-C $df_drive`;
          @words = split (' ',${words[-1]});
          $fs = $words[-3];
          $fs = sprintf("%9d",$fs / 1024 / 1024);
        }
        else
        {
          $fs = ".";
        }
      }

      if ($ts eq $lastts)
      {
        #$tsout = '""';
        $tsout = $ts;
      }
      else
      {
        $tsout = $ts;
      }

      printf("%-16s %4d %-55s %-8s %9d %9s\n",$tsout,$num,$df,$status,$size,$fs);
      $lastts = $ts;
    }
  }
  print "\n";
  &msg("Found $dfcount datafile(s)");
  exit;
}

# ------------------------------------------------------------------------------
# -df2
# ------------------------------------------------------------------------------
sub df2_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-df2" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";
  print "Tablespace       DF   Datafile                                                Datafile  Datafile  FS Free  \n";
  print "Name             Num  Path                                                    Status    Size(MB)  Space(MB)\n";
  print "---------------- ---- ------------------------------------------------------- --------- --------- ---------\n";

  $sqlstmt = 'select \'X\',
  tablespace_name,
  file_name,
  nvl(bytes,0),
  file_id,
  nvl(status,\'?\')
  from dba_data_files order by 4 desc;';

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $dfcount++;
      $status = "";
      $ts     = $words[1];
      $df     = $words[2];
      $size   = $words[3]/1048576;
      $num    = $words[4];
      $status = $words[5];

      ##print "num=$num \n";

      if (-c $df)
      {
        $fs = "(Raw)";
      }
      else
      {
        if ($arg2 eq "FSCHECK")
        {
          $df_drive = substr($df,0,2);
          @words = `dir /-C $df_drive`;
          @words = split (' ',${words[-1]});
          $fs = $words[-3];
          $fs = sprintf("%9d",$fs / 1024 / 1024);
        }
        else
        {
          $fs = ".";
        }
      }

      if ($ts eq $lastts)
      {
        #$tsout = '""';
        $tsout = $ts;
      }
      else
      {
        $tsout = $ts;
      }
 
      printf("%-16s %4d %-55s %-8s %9d %9s\n",$tsout,$num,$df,$status,$size,$fs);
      $lastts = $ts;
    }
  }
  print "\n";
  &msg("Found $dfcount datafile(s)");
  exit;
}

# ------------------------------------------------------------------------------
# -df3
# ------------------------------------------------------------------------------
sub df3_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-df3" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = ' select \'Z\', 
      name 
    from
      v\$datafile 
    order by 2;';

  print "\nDatafile\n";
  print "-------------------------------------------------\n";
  
  $dfcount = 0; 

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    $df      = $words[1];

    if ($verb eq "Z")
    {
      printf "%-50s\n",$df;
      $dfcount++; 
    }
  }
  print "\n";
  &msg("Found $dfcount datafiles"); 
  exit;
}

# --------------------------------------------------------------------
# -dh
# --------------------------------------------------------------------
sub dh_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-dh" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $rows = 0;

  &msg("(This command will run for awhile on databases with many datafiles)");
  $sqlstmt = '
    SELECT 
      \'X\',
      a.tablespace_name,
      a.file_name,
      a.bytes,
      (b.maximum+c.blocks-1)*d.db_block_size highwater
    from
      dba_data_files a,
      (select 
        file_id,
        max(block_id) maximum          
       from dba_extents          
       group by file_id) b,
      dba_extents c,
      (select 
        value db_block_size          
        from v\$parameter          
        where name=\'db_block_size\') d
    where
      a.file_id  = b.file_id
      and c.file_id  = b.file_id
      and c.block_id = b.maximum
    order by
      a.tablespace_name,a.file_name;
  ';


  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "Tablespace       Datafile                                                             Size     Used HWM      Savings Resize\n"; 
  print "---------------- ------------------------------------------------------------ ------------ ------------ ------------ ------\n";

  $spaceflag = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $ts = shift @words;
      $dfname   = shift @words;
      $bytes    = shift @words;
      $hwm      = shift @words;

      $safe = int(($hwm + 1000000)/1000000) . "M";

      $savings = $bytes - $hwm;
      $t_savings += $savings;
  
      $resize_cmds .= "ALTER DATABASE DATAFILE '$dfname' RESIZE $safe ;\n";

      printf "%-16s %-60s %12s %12s %12s %6s\n",
	$ts ,
	$dfname  ,
	$bytes ,
	$hwm ,
	$savings ,
        $safe
        ;

      $rows++;
      $total += $count;
    }
  }
  print "================ ============================================================ ============ ============ ============ ======\n";
  printf "%-16s %-60s %12s %12s %12s %6s\n",
    'Total:' ,
    ''  ,
    '' ,
    '' ,
    $t_savings ,
    ''
    ;

  print "\n";

  &msg("Resize Commands:");
  print "$resize_cmds\n";

  exit;
}

# --------------------------------------------------------------------
# -dn
# --------------------------------------------------------------------
sub dn_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-dn" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &subEMail( 'border@dba.it.agilent.com', "monitor--$P $arg $host/$ENV{ORACLE_SID}", '' );
  $rows = 0;
  $resize_cmd_count = 0;

  &msg("-dn: Datafile 'USED' High-Water-Mark");
  &msg("(This command will run for awhile on databases with many datafiles)");
  $sqlstmt = '
    select \'W\',
      value
    from 
    v\$parameter
    where
      name = \'db_block_size\';
    WHENEVER SQLERROR CONTINUE;
    drop table DBAMON.ORASTAT_DH_1;
    WHENEVER SQLERROR EXIT 1;
    create global temporary table DBAMON.ORASTAT_DH_1
      as     
        select 
          file_id,
          max(block_id + blocks) highwater_blocks
        from
          dba_extents
        group by
          file_id;
    insert into DBAMON.ORASTAT_DH_1
        select 
          file_id,
          max(block_id + blocks) highwater_blocks
        from
          dba_extents
        group by
          file_id;
    -- select * from DBAMON.ORASTAT_DH_1;
    SELECT 
      \'X\',
      a.tablespace_name,
      a.file_name,
      a.bytes,
      b.highwater_blocks
    from
      dba_data_files a,
      DBAMON.ORASTAT_DH_1 b
    where
      a.file_id  = b.file_id
    order by
      a.tablespace_name,
      a.file_name;
    WHENEVER SQLERROR CONTINUE;
    drop table DBAMON.ORASTAT_DH_1;
    WHENEVER SQLERROR EXIT 1;
  ';

  ##print "sqlstmt=$sqlstmt\n";

  print "\n";
  print "Tablespace       Datafile                                                             Size     Used HWM      Savings Resize\n"; 
  print "---------------- ------------------------------------------------------------ ------------ ------------ ------------ ------\n";

  $spaceflag = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print; 
    print if /^ORA/ && !/^ORA-00942/;
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "W")
    {
      $block_size = shift @words;
    }

    if ($verb eq "X")
    {
      $ts = shift @words;
      $dfname     = shift @words;
      $bytes      = shift @words;
      $hwm_blocks = shift @words;

      $hwm        = $hwm_blocks * $block_size;
      $hwm_mb     = int($hwm / 1000000); 

      $bytes_mb   = int($bytes / 1000000); 

      $safe = int(($hwm + 1000000)/1000000) . "M";

      $savings = $bytes - $hwm;
  
      $savings_mb = int($savings / 1000000); 
      $t_savings += $savings;

      if ($savings > 10000000)
      {
        $resize_cmds .= "ALTER DATABASE DATAFILE '$dfname' RESIZE $safe  -- Datafile_Size=${bytes_mb}M Data HWM=${hwm_mb}M Savings=${savings_mb}M ; \n";
        $resize_cmd_count++;
        $t_savings_cmd += $savings;
      }

      printf "%-16s %-60s %12s %12s %12s %6s\n",
	$ts ,
	$dfname  ,
	$bytes ,
	$hwm ,
	$savings ,
        $safe
        ;

      $rows++;
      $total += $count;
    }
  }
  print "================ ============================================================ ============ ============ ============ ======\n";
  printf "%-16s %-60s %12s %12s %12s %6s\n",
    'Total:' ,
    ''  ,
    '' ,
    '' ,
    $t_savings ,
    ''
    ;

  print "\n";

  &msg("Now, you will see syntactically correct ALTER DATAFILE commands for all");
  &msg("  datafiles which can be resized to save at least 10M per datafile.\n");

  &msg("Resize Commands:");
  print "$resize_cmds\n";

  $t_savings_mb = int($t_savings_cmd / 1000000);
  $t_savings_gb = int($t_savings_cmd / 1000000000);

  &msg("If all $resize_cmd_count ALTER DATAFILE RESIZE commands are run:");
  &msg("  Total Savings (mB): $t_savings_mb");
  &msg("  Total Savings (gB): $t_savings_gb");

  exit;
}

# --------------------------------------------------------------------
# -dp
# --------------------------------------------------------------------
sub dp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-dp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'Z\', 
      property_name, 
      substr( property_value, 1, 20 )
    from
      database_properties
    order by 2;';

  print "\n";
  print "Name                           Value\n";
  print "------------------------------ ------------------\n";
  
  $dfcount = 0; 

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];

    if ($verb eq "Z")
    {
      $name   = $words[1];
      $val    = $words[2];
      printf "%-30s %-20s\n",$name, $val;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -du
# --------------------------------------------------------------------
sub du_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-du" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  
  $rows = 0;

  $sqlstmt = "
    SELECT 
      \'X\',
      df.name,
      df.unrecoverable_change#,
      to_char(unrecoverable_time,\'YYYY/MM/DD-HH24:MI:SS\') 
    FROM 
      v\\\$datafile df
    where 
      df.unrecoverable_change# > 0
    order by
      2;";

  print "\n";
  print "Name                                               Unrec. Change#  Unrec. Date\n";
  print "-------------------------------------------------- --------------- -------------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $objid = "";

      $dfname   = shift @words;
      $chgnum   = shift @words;
      $chgdat   = shift @words;

      printf "%-50s %-15s %-19s\n",$dfname,$chgnum,$chgdat;
      $rows++;
   
      $maxdat = $chgdat if $chgdat gt $maxdat;
    }
  }
  print "\n";
  &msg("$rows Datafile(s) with unrecoverable changes found");
  &msg("Most recent UNRECOVERABLE_TIME value found: $maxdat");
  exit;
}

# --------------------------------------------------------------------
# -ec
# --------------------------------------------------------------------
sub ec_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ec" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $editor_cmd = ( defined $ENV{"EDITOR"} ) ? $ENV{"EDITOR"} : $default_editor;

  $i_file    = undef;

  chomp( $line = `findstr /I /C:ifile $initfile` );
  if ( defined $line and length( $line ) > 0 )
  {
      @words = split( /\s+/, $line );
      $ifile = pop @words;
      $ifile =~ s/'//g;
      print "IFILE: \"$ifile\"\n";
      system ( "$editor_cmd $ifile" );
  }
  else
  {
      print "IFILE parameter Not specified within: $init_file\n";
  }

  exit;
}

# ------------------------------------------------------------------------------
# -er
# ------------------------------------------------------------------------------
sub er_task
{

  my( $package, $file, $line ) = caller;
  if ( $arg ne "-er" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $count = 0;

  $sqlstmt = 'select \'X\', 
    OWNER, NAME, TYPE, SEQUENCE, LINE, POSITION, trim(trailing \' \' from TEXT) from dba_errors ;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $text = substr($_,index($_,$words[7]));
      print "Owner=$words[1] Name=$words[2] Type=$words[3] Seq=$words[4] Line=$words[5] Position=$words[6] Text=$text";
      $count++;
    }
  }

  if ($count < 1) { &msg("No rows selected"); }
  exit;
}

# --------------------------------------------------------------------
# -ev
# --------------------------------------------------------------------
sub ev_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ev" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select 
      \'X\', 
      translate(event,\' \',\'_\'),
      total_waits
    from v\$system_event order by 2;';

  #&msg("sqlstmt=$sqlstmt");
  print "Event                                         Count\n";
  print "---------------------------------------- ----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    if ($verb eq "X")
    {
      $event = $words[1];
      $count = $words[2];
      printf "%-40s %10d \n",$event,$count;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -ex
# --------------------------------------------------------------------
sub ex_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ex" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  #$sqlstmt = '
  #  select \'X\',  
  #  to_char(timestamp,\'YYYY/MM/DD-HH24:MI:SS\'),
  #  lpad(\' \',2*level) || operation || \'\' || options || \' \' || object_name 
  #  from plan_table
  #  connect by prior id=parent_id        
  #  start with id=1
  #  order by 2;';
  $linesize = ( $oraver_num < 9 ) ? "set charwidth 16" : "set linesize 9999";
  $sqlstmt = "
    $linesize;
    select \'X\',  
    to_char(timestamp,\'YYYY/MM/DD-HH24:MI:SS\'),
    replace(operation,   ' ','.'),
    replace(options,     ' ','.'),
    replace(object_owner,' ','.'),
    replace(object_name, ' ','.'),
    replace(object_type, ' ','.'),
    nvl(to_char(id),'.'), 
    nvl(to_char(parent_id),'.')
    from plan_table
    order by 2, id;";

  #&msg("sqlstmt=$sqlstmt");
  print "plan_table Explain Data\n";
  print "====================================\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $ts   = $words[1];
      $op   = $words[2];
      $opt  = $words[3];
      $obj  = $words[4] . '.' . $words[5];
      $typ  = $words[6];
      #print "ts=$ts op=$op opt=$opt \n";
      printf "%-s %-16s %-16s %-16s id=$words[7] parent_id=$words[8] \n", $ts, $op, $opt, $obj, $typ ;
    }
  }
  exit;
}

# ------------------------------------------------------------------------------
# -fm
# ------------------------------------------------------------------------------
sub fm_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-fm" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  
  $file = $arg2;

  if ( $file eq "" )
  {
      print "$P | Usage -fm FILENAME\n";
      exit 1;
  }

  &msg("-fm: File status $file");  

  if (! -e "$file")
  {
    &msg("File $file does not exist"); 
    exit;
  }

  $fileage_secs = time - ( stat "$file" ) [9];
  $fileage_hours = sprintf("%.2f",($fileage_secs / 3600));
  &msg("Hours_Since_Update: $fileage_hours"); 
  exit;
}

# ------------------------------------------------------------------------------
# -fs
# ------------------------------------------------------------------------------
sub fs_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-fs" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
    df.tablespace_name,
    df.file_name,
    df.file_id,
    nvl(df.bytes,0),
    nvl(sum(fs.bytes),0)
    from dba_data_files df, dba_free_space fs
    where df.file_id = fs.file_id
    group by \'X\', df.tablespace_name, df.file_name, df.file_id, df.bytes
    order by 2,4;';

  print "Tablespace       DF   Datafile                                                Datafile  Datafile  Datafile  \n";
  print "Name             Num  Path                                                    Alloc(Kb) Used(Kb)  Free(Kb)  \n";
  print "---------------- ---- ------------------------------------------------------- --------- --------- --------- \n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $dfcount++;
      $ts = $words[1];
      $df = $words[2];
      $num = $words[3];
      $size = $words[4]/1024;
      $free = $words[5]/1024;

      printf("%-16s %4d %-55s %9d %9d %9d\n",$ts,$num,$df,$size,$size-$free,$free);
    }
  }
  print "\n";
  &msg("Found $dfcount datafile(s) with free space.");
  exit;
}

# ------------------------------------------------------------------------------
# -ft
# ------------------------------------------------------------------------------
sub ft_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ft" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'Y\',
      name, value
      from v\$parameter
      where
      name in (\'db_block_size\');
    select \'X\',
      usn, 
      slt,
      seq,
      state,
      undoblocksdone,
      undoblockstotal,
      cputime
      from v\$fast_start_transactions
      order by 2,3,4;';

  print "\n";
  print "Undo Seg/Slt/Inc Rollback State   CPU Time   Blks Done  Blks Total mB Done    mb Total   Pct.\n";
  print "---------------- ---------------- ---------- ---------- ---------- ---------- ---------- ----\n";

  $transcount = 0;

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      $blocksize  = $words[1];
    }
    if ($verb eq "X")
    {
      $transcount++;

      $usn    = $words[1];
      $slt    = $words[2];
      $seq    = $words[3];
      $state  = $words[4];
      $blks_d = $words[5];
      $blks_t = $words[6];
      $cputim = $words[7];

      $undo = "$usn/$slt/$seq";
      $pctdone = ($blks_d / $blks_t) * 100;
      $pctdone = sprintf("%.2f",$pctdone);

      $undo_t_mb = int(($blocksize * $blks_t) / 1000000);
      $undo_d_mb = int(($blocksize * $blks_d) / 1000000);

      printf("%-16s %-16s %10s %10s %10s %10s %10s %-4s\n",    
        $undo,
        $state,
        $cputim,
        $blks_d,
        $blks_t,
        $undo_d_mb,
        $undo_t_mb,
        $pctdone);
    }
  }
  print "\n";
  &msg("Found $transcount Transactions being rolled back");
  exit;
}

# --------------------------------------------------------------------
# -hd
# --------------------------------------------------------------------
sub hd_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-hd" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $rows = 0;

  if ($arg2 ne "") 
  {
    $touches_threshold = $arg2; 
  }
  else
  {
    $touches_threshold = 1000; 
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - Having Block Touches > $touches_threshold");

  $sqlstmt = "
    select \'Y\',
      name, value
      from v\\\$parameter
      where
      name in (\'db_block_size\');
    SELECT 
      \'X\',
      bh.obj,
      ob.owner,
      ob.object_name,
      ob.object_type,
      df.name,
      COUNT(2) buffers,
      sum(bh.tch) sum
    FROM 
      x\\\$bh bh,
      dba_objects ob,
      v\\\$datafile df
    where 
      bh.obj = ob.object_id
      and bh.file# = df.file#
    GROUP BY 
      bh.obj,
      ob.owner,
      ob.object_name,
      ob.object_type,
      df.name
    HAVING
      sum(tch) > $touches_threshold 
    order by
      8 desc;";

  print "\n";
  print "Name                                               Type       Datafile                                           Buffers    Size(MB)   Touches\n";
  print "-------------------------------------------------- ---------- -------------------------------------------------- ---------- ---------- ----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "Y")
    {
      $blocksize = $words[1];
      ##print "blocksize=$blocksize\n";
      next;
    }

    if ($verb eq "X")
    {
      $objid = "";

      $objid   = shift @words;
      $owner   = shift @words;
      $obj     = shift @words;
      $type    = shift @words;
      $dfname = shift @words;
      $buffers = shift @words;
      $touches = shift @words;

      $objname = "$owner.$obj";
      $objsize = int(($buffers * $blocksize) / 1000000);
  
      printf "%-50s %-10s %-50s %10s %10s %10s\n",$objname,$type,$dfname,$buffers,$objsize,$touches;
      $rows++;
    }
  }
  print "\n";
  &msg("$rows Row(s) found");
  exit;
}

# --------------------------------------------------------------------
# -ht
# --------------------------------------------------------------------
sub ht_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ht" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $rows = 0;

  if ($arg2 ne "") 
  {
    $touches_threshold = $arg2; 
  }
  else
  {
    $touches_threshold = 1000; 
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - Having Block Touches > $touches_threshold");

  $sqlstmt = "
    select \'Y\',
      name, value
      from v\\\$parameter
      where
      name in (\'db_block_size\');
    SELECT 
      \'X\',
      bh.obj,
      ob.owner,
      ob.object_name,
      ob.object_type,
      COUNT(2) buffers,
      sum(bh.tch) sum 
    FROM 
      x\\\$bh bh,
      dba_objects ob
    where 
      bh.obj = ob.object_id
    GROUP BY 
      bh.obj,
      ob.owner,
      ob.object_name,
      ob.object_type
    HAVING
      sum(tch) > $touches_threshold 
    order by
      7 desc;";

  print "\n";
  print "Name                                     Type          Buffers   Size(MB)    Touches\n";
  print "---------------------------------------- ---------- ---------- ---------- ----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "Y")
    {
      $blocksize = $words[1];
      ##print "blocksize=$blocksize\n";
      next;
    }

    if ($verb eq "X")
    {
      $objid = "";

      $objid   = shift @words;
      $owner   = shift @words;
      $obj     = shift @words;
      $type    = shift @words;
      $buffers = shift @words;
      $touches = shift @words;

      $objname = "$owner.$obj";
      $objsize = int(($buffers * $blocksize) / 1000000);
  
      printf "%-40s %-10s %10s %10s %10s\n",$objname,$type,$buffers,$objsize,$touches;
      $rows++;
    }
  }
  print "\n";
  &msg("$rows Row(s) found");
  exit;
}

# --------------------------------------------------------------------
# -in
# --------------------------------------------------------------------
sub in_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-in" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $key = '{INDEX}' if $arg2 eq '-key'; 
  $sqlstmt = 'select \'Y\',
    name, value
    from v\$parameter
    where
    name in (\'db_block_size\',\'db_block_buffers\');
  select 
    \'X\', 
    sg.owner, 
    sg.segment_name, 
    sg.tablespace_name, 
    sg.blocks, 
    sg.extents, 
    sg.max_extents,
    nx.status,
    nx.table_name
  from 
    dba_segments sg,
    dba_indexes nx
  where 
    segment_type = \'INDEX\'
    and sg.segment_name = nx.index_name
    and sg.owner  = nx.owner
  order by 2,3;';

  print "\n";
  print "Owner.Index                               TableSpace   Blocks     MB Xtents  MaxXtnt Status\n";
  print "----------------------------------------- ------------ ------ ------ ------ -------- ----------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        $bs = $words[2];
        ##print "bs=$bs\n";
      }
    }
    if ($verb eq "X")
    {
      $owner = $words[1];
      $table = $words[2];
      $ts    = $words[3];
      $blks  = $words[4];
      $exts  = $words[5];
      $maxx  = $words[6];
      $stat  = $words[7];
      $tbl   = $words[8];
      $mb   = ($blks * $bs) / (1024 * 1024);

      if ($maxx == 2147483645)
      {
        $maxx = "(Unlim)";
      }

      $t = $owner . "." . $table;
      if (substr($owner,0,3) ne "SYS")
      {
        printf("%-41s %-12s %6d %6d %6d %8s %-10s $key\n",$t,$ts,$blks,$mb,$exts,$maxx,$stat);
        $found++;
        $tot_mb += $mb;
        $tot_xt += $exts;
      }
    }
  }
  print  "========================================= ============ ====== ====== ====== ======== ==========\n";
  printf "(%6s Total Indexes)                                        %6d %6d\n\n",$found,$tot_mb,$tot_xt;
  exit;
}

# --------------------------------------------------------------------
# -iv
# --------------------------------------------------------------------
sub iv_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-iv" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', 
    owner || \'.\' || object_name, status, object_type from dba_objects 
    where status != \'VALID\' order by 1, 2;';

  print "Object Name                                        Object Type    Status\n";
  print "-------------------------------------------------- -------------- ------\n";
  
  $count = 0;

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    #print;
    @words = split();
    $verb   = $words[0];
    $object = $words[1];
    $status = $words[2];
    $type   = $words[3] . ' ' . $words[4];
    if ($verb eq "X")
    {
      printf "%-50s %-14s %-16s\n",$object,$type,$status;
      ++$count;
    }
  }
  print "\n";
  &msg("Found $count Invalid Objects");
  exit;
}

# --------------------------------------------------------------------
# -l
# --------------------------------------------------------------------
sub l_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-l" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  &oraSQL("
set time off
host echo 'Running archive log list'
archive log list;
  ");

  print "\n";
  foreach (@oraSQLout)
  {
    s/SQL> //g;
    s/SVRMGR> //g;
    print if $_ !~ /^$/;
    @words = split;
    $verb    = $words[0];
    if (/Database log mode/) 
    { 
      $arclogmode = $words[3] . $words[4];
    }
    if (substr($_,0,18) eq "Automatic archival") 
    { 
      $arcauto    = $words[2];
    }
    if (substr($_,0,19) eq "Archive destination") 
    { 
      $arcdirraw  = $words[2];
      $arcdirraw  =~ s/^\?/$oh/;
      $arcdir     = $arcdirraw;
    }
  }

  $arcfs  = substr( $arcdir, 0, 2 );
  $letter = substr( $arcdir, 0, 1 );

  chomp( $srvinfo = `srvinfo -ns | findstr /I /C:$letter\$` );
  $srvinfo =~ s/^\s+//;
  ( $drive_letter, $file_sys, $total, $free, $used ) = split( /\s+/, $srvinfo, 5 );
  $pct_used = ( $used / $total ) * 100;

  &msg("Archive Dest.:        $arcdirraw");
  &msg("Archive Log Mode:     $arclogmode");
  &msg("Auto Archive:         $arcauto");
  &msg("Archive Dest Dir.:    $arcdir");
  &msg("Archive Dest FS:      $arcfs");
  &msg( sprintf( "%s %5.2f%", "Archive Dest FS Full:", $pct_used ) );

  exit;
}

# --------------------------------------------------------------------
# -la
# --------------------------------------------------------------------
sub la_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-la" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - Where GETS>1000000 And MISSES>100");

  $sqlstmt = '
    select \'X\', GETS,
        MISSES,
        round(((GETS-MISSES)*100) / GETS , 2),
        IMMEDIATE_GETS,
        IMMEDIATE_MISSES,
        sleeps,
        spin_gets,
        addr,
        latch#,
        name
    from    v\$latch
    where MISSES > 0 and gets > 0
    --where   GETS > 1000000
    --and     MISSES > 100
    order   by ((GETS-MISSES) / GETS) desc;';

  print "Latch Name                                                           Gets         Misses    Immed. Gets  Immed. Misses         Sleeps      Spin Gets Get Hit%\n";
  print "---------------------------------------------------------- -------------- -------------- -------------- -------------- -------------- -------------- --------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $gets    = shift @words;
      $misses  = shift @words;
      $hitrat  = shift @words;
      $immgets = shift @words;
      $immmiss = shift @words;
      $sleeps  = shift @words;
      $spingets= shift @words;
      $latchnum= shift @words;
      $addr    = shift @words;
      $name    = "";
      foreach (@words)
      {
	$name .= "$_ ";
      }

      $nameaddr = "$name / $latchnum / $addr";
      printf "%-58s %14s %14s %14s %14s %14s %14s %7.2f%%\n",$nameaddr,$gets,$misses,$immgets,$immmiss,$sleeps,$spingets,$hitrat;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -lf -rl
# --------------------------------------------------------------------
sub lf_or_rl_task
{
  my( $package, $file, $line ) = caller;
  unless ( $arg eq "-lf" or $arg eq "-rl" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = ' select \'X\', group#, member, status from v\$logfile order by 1,2;';

  print "Group Name                                               Status\n";
  print "----- -------------------------------------------------- --------------------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb   = $words[0];
    $group  = $words[1];
    $member = $words[2];
    $status = $words[3];
    $fstchg = $words[4];
    if ($verb eq "X")
    {
      printf "%5d %-50s %-20s %-10s\n",$group,$member,$status,$fstchg;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -lg
# --------------------------------------------------------------------
sub lg_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lg" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'Z\', 
      sid, 
      serial#, 
      context, 
      sofar, 
      totalwork, 
      last_update_time - start_time, 
      --round(sofar/totalwork*100,2) "PctComplete",
      to_char(start_time,\'YYYY/MM/DD-HH24:MI:SS\') ,
      to_char(last_update_time,\'YYYY/MM/DD-HH24:MI:SS\') ,
      opname
    from
      v\$session_longops 
      where 
        time_remaining > 0 
    order by 8, 4;';

  print "SID-Ser#  Start               Last Updated        El. Mins Work Sofar Work Total Pct. Done  Desc.\n";
  print "--------- ------------------- ------------------- -------- ---------- ---------- ---------- ------------------------\n";
  
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    $sid     = $words[1];
    $ser     = $words[2];
    #$context = $words[3];
    $sofar   = $words[4];
    $totalwk = $words[5];
    $etime   = sprintf("%5.1f",$words[6] * (24*60));
    $stime   = $words[7];
    $utime   = $words[8];
    $opname  = $words[9] . $words[10]  . $words[11]  . $words[12];

    if ($totalwk > 0) { $pctcomp = sprintf("%7.3f",($sofar/$totalwk)*100); }
    else              { $pctcomp = "*"; }

    $totalsize = $blocksize * $sofar;
    $gbrate = "?";

    $sidser = "$sid-$ser";

    if ($verb eq "Z")
    {
      printf "%-9s %-19s %-19s %8s %10s %10s %10s %-10s \n",$sidser,$stime,$utime,$etime, $sofar,$totalwk,$pctcomp,$opname;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -lh
# --------------------------------------------------------------------
sub lh_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lh" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  if ($oraver_num < 8.1)
  {
    print "This feature does not work with $oraver\n";
    exit;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " (catblock Must Have Been Run)");

  $sqlstmt = '
  alter session set optimizer_goal=RULE;
  select \'X\',
    waiting_session,
    holding_session,
    lock_type,
    mode_held,
    mode_requested,
    lock_id1,
    lock_id2
  from 
    dba_waiters
    order by 1;';

  print "\n";
  print " Waiting  Holding Lock         Mode         Mode         Lock\n";
  print " Session  Session Type         Holding      Reqd.        ID1 ID2\n";
  print "-------- -------- ------------ ------------ ------------ -------------------------\n";

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $ses_wt   = $words[1];
      $ses_hld  = $words[2];
      $lk_type  = $words[3];
      $mode_hld = $words[4];
      $mode_req = $words[5];
      $lock_id  = "$words[6] $words[7] $words[8] $words[9] $words[10]";
      $age = sprintf("%5.1f",$age);
      printf "%-8s %-8s %-12s %-12s %-12s %s %s \n", $ses_wt, $ses_hld, $lk_type, $mode_hld, $mode_req, $lock_id;
      $counter++;
    }
  }
  print "\n";
  if ($counter < 1)
  {
    &msg("No Lock Waiters Found"); 
  }
  else
  {
    &msg("$counter Lock Waiters Found"); 
  }
  exit;
}

# --------------------------------------------------------------------
# -li
# --------------------------------------------------------------------
sub li_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-li" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', 
    resource_name, 
    INITIAL_ALLOCATION , 
    CURRENT_UTILIZATION ,
    MAX_UTILIZATION   ,
    LIMIT_VALUE
    from v\$resource_limit order by 2;';

  print "Resource                          Initial    Current    Maximum      Limit\n";
  print "Name                                Value      Value      Value      Value\n";
  print "------------------------------ ---------- ---------- ---------- ----------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    $name   = $words[1];
    $init   = $words[2];
    $curr   = $words[3];
    $max    = $words[4];
    $limit  = $words[5];

    if ($verb eq "X")
    {
      printf "%-30s %10s %10s %10s %10s\n",$name,$init,$curr,$max,$limit;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -lk
# --------------------------------------------------------------------
sub lk_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lk" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    alter session set optimizer_goal=RULE;
    select \'X\', 
        rpad(osuser, 9)||lpad(p.spid, 5) osuser,
	rpad(s.username,8)||lpad(s.sid, 4)||lpad(s.serial#, 5) username,
	decode(l.type, 
		\'MR\', \'Media_Reco\', 
		\'RT\', \'Redo_Thred\',
		\'UN\', \'User_Name\',
		\'TX\', \'Trans\',
		\'TM\', \'DML\',
		\'UL\', \'PL/SQL_Usr\',
		\'DX\', \'Dist._Tran\',
		\'CF\', \'Cntrl_File\',
		\'IS\', \'Inst_State\',
		\'FS\', \'File_Set\',
		\'IR\', \'Inst_Reco\',
		\'ST\', \'Disk_Space\',
		\'TS\', \'Temp_Seg\',
		\'IV\', \'Cache_Inv\',
		\'LS\', \'Log_Switch\',
		\'RW\', \'Row_Wait\',
		\'SQ\', \'Seq_Number\',
		\'TE\', \'Extend_Tbl\',
		\'TT\', \'Temp_Table\',
		l.type) locktype,
	\'.\' object_name,
	decode(lmode,1,Null,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',\' \') held,
	decode(request,1,Null,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',\' \') request
	from v\$lock l, v\$session s, v\$process p
	where s.sid = l.sid and
		s.username <> \' \' and
		s.paddr = p.addr and
		l.type <> \'TM\' and
		(l.type <> \'TX\' or l.type = \'TX\' and l.lmode <> 6)
union
select \'X\', rpad(osuser, 9)||lpad(p.spid, 5) osuser,
	rpad(s.username,8)||lpad(s.sid, 4)||lpad(s.serial#, 5) username,
	decode(l.type, 
		\'MR\', \'Media_Reco\', 
		\'RT\', \'Redo_Thred\',
		\'UN\', \'User_Name\',
		\'TX\', \'Trans\',
		\'TM\', \'DML\',
		\'UL\', \'PL/SQL_Usr\',
		\'DX\', \'Dist._Tran\',
		\'CF\', \'Cntrl_File\',
		\'IS\', \'Inst_State\',
		\'FS\', \'File_Set\',
		\'IR\', \'Inst_Reco\',
		\'ST\', \'Disk_Space\',
		\'TS\', \'Temp_Seg\',
		\'IV\', \'Cache_Inv\',
		\'LS\', \'Log_Switch\',
		\'RW\', \'Row_Wait\',
		\'SQ\', \'Seq_Number\',
		\'TE\', \'Extend_Tbl\',
		\'TT\', \'Temp_Table\',
		l.type) locktype,
	object_name,
	decode(lmode,1,NULL,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',NULL) held,
	decode(request,1,NULL,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',NULL) request
	from v\$lock l, v\$session s, v\$process p, sys.dba_objects o
	where s.sid = l.sid and
		o.object_id = l.id1 and
		l.type = \'TM\' and
		s.username <> \' \' and
		s.paddr = p.addr
union
select \'X\', rpad(osuser, 9)||lpad(p.spid, 5) osuser,
	rpad(s.username,8)||lpad(s.sid, 4)||lpad(s.serial#, 5) username,
	decode(l.type, 
		\'MR\', \'Media_Reco\', 
		\'RT\', \'Redo_Thred\',
		\'UN\', \'User_Name\',
		\'TX\', \'Trans\',
		\'TM\', \'DML\',
		\'UL\', \'PL/SQL_Usr\',
		\'DX\', \'Dist._Tran\',
		\'CF\', \'Cntrl_File\',
		\'IS\', \'Inst_State\',
		\'FS\', \'File_Set\',
		\'IR\', \'Inst_Reco\',
		\'ST\', \'Disk_Space\',
		\'TS\', \'Temp_Seg\',
		\'IV\', \'Cache_Inv\',
		\'LS\', \'Log_Switch\',
		\'RW\', \'Row_Wait\',
		\'SQ\', \'Seq_Number\',
		\'TE\', \'Extend_Tbl\',
		\'TT\', \'Temp_Table\',
		l.type) locktype,
	\'(Rollback=\'||rtrim(r.name)||\')\' object_name,
	decode(lmode,1,NULL,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',NULL) held,
	decode(request,1,NULL,2,\'Row_Share\',3,\'Row_Excl\',4,\'Share\',
		5,\'Sh_Row_Ex\',6,\'Exclusive\',NULL) request
	from v\$lock l, v\$session s, v\$process p, v\$rollname r
	where s.sid = l.sid and
		l.type = \'TX\' and
		l.lmode = 6 and
		trunc(l.id1/65536) = r.usn and
		s.username <> \' \' and
		s.paddr = p.addr
	order by 5, 6;';

  $locks = 0;

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    $osuser   = $words[1];
    $pid      = $words[2];
    $dbuser   = $words[3];
    $dbid     = $words[4];
    $dbser    = $words[5];
    $type     = $words[6];
    $obj      = $words[7];
    $held     = $words[8];
    $reqd     = $words[9];

    if ($verb eq "X")
    {
      if ($locks == 0)
      {
	print "OS User  PID   DB User  ID   Ser# Type       Object Name     Lock Held Lock Reqd\n";
	print "-------- ----- -------- ---- ---- ---------- --------------- --------- ---------\n";
      }
      printf "%-8s %-5s %-8s %4s %4s %-10s %-15s %-8s %-8s\n",
              $osuser,$pid,$dbuser,$dbid,$dbser,$type,$obj,$held,$reqd;
      $locks++;
    }
  }
  print "\n";
  &msg("$locks Locks Found");
  exit;
}

# --------------------------------------------------------------------
# -ln
# --------------------------------------------------------------------
sub ln_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ln" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $rows = 0;

  $sqlstmt = '
    SELECT 
      \'X\',
      nvl(owner,\'(Null)\'), 
      db_link,
      created,
      nvl(username,\'(Null)\'),
      host
    FROM 
      dba_db_links
    order by
      2, 3 ;';

  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "Owner        DB Link Name         Created      User Name    Host \n";
  print "------------ -------------------- ------------ ------------ ------------------------------------------\n";

  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $owner    = shift @words;
      $dblink   = shift @words;
      $created  = shift @words;
      $username = shift @words;
      $host     = shift @words;

      printf "%-12s %-20s %-12s %-12s %s\n",$owner,$dblink,$created,$username,$host;
      $rows++;
    }
  }
  print "\n";
  &msg("Found $rows DB Links");
  exit;
}

# --------------------------------------------------------------------
# -lo
# --------------------------------------------------------------------
sub lo_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lo" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    alter session set optimizer_goal=RULE;
    select \'X\',
            nvl(S.USERNAME,\'Internal\') username, 
	    nvl(S.TERMINAL,\'None\') terminal, 
	    L.SID||\',\'||S.SERIAL# Kill, 
	    U1.NAME||\'.\'||substr(T1.NAME,1,20) tab, 
	    decode(L.LMODE,1,\'No_Lock\', 
		    2,\'Row_Share\', 
		    3,\'Row_Exclusive\', 
		    4,\'Share\', 
		    5,\'Share_Row_Exclusive\', 
		    6,\'Exclusive\',null) lmode, 
	    decode(L.REQUEST,1,\'No_Lock\', 
		    2,\'Row_Share\', 
		    3,\'Row_Exclusive\', 
		    4,\'Share\', 
		    5,\'Share_Row_Exclusive\', 
		    6,\'Exclusive\',null) request 
    from    V$LOCK L,  
	    V$SESSION S, 
	    SYS.USER$ U1, 
	    SYS.OBJ$ T1 
    where	L.SID = S.SID  
    and	T1.OBJ# = decode(L.ID2,0,L.ID1,L.ID2)  
    and	U1.USER# = T1.OWNER# 
    and	S.TYPE != \'BACKGROUND\' 
    order by 2,3,6;';
  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "User         Term     Sess,Ser   Table                        Lock Held           Lock Reqd.\n";
  print "------------ -------- ---------- ---------------------------- ------------------- -----------------\n";
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words = split;
    $verb = $words[0];
    if ($verb eq "X")
    {
      $user = $words[1];
      $term = $words[2];
      $sess = $words[3];
      $tabl = $words[4];
      $held = $words[5];
      $reqd = $words[6];

      printf "%-12s %-8s %-10s %-28s %-20s %-20s\n",
        $user,$term,$sess,$tabl,$held,$reqd;
      
      $count++;
    }
  }
  print "\n";
  if ($count < 1) { &msg("No table locks found"); }
  exit;
}

# --------------------------------------------------------------------
# -ls
# --------------------------------------------------------------------
sub ls_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ls" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($oraver_num == 8.0)
  {
    system("lsnrctl80 status 2>&1");
    system("lsnrctl80 dbsnmp_status 2>&1");
  }
  elsif ($oraver_num < 9)
  {
    system("lsnrctl status 2>&1");
    system("lsnrctl dbsnmp_status 2>&1");
  }
  else
  {
    system("lsnrctl status 2>&1");
  }
  exit;
}

# --------------------------------------------------------------------
# -lv
# --------------------------------------------------------------------
sub lv_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lv" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  @srvinfo = `srvinfo -ns`;
  foreach (@srvinfo)
  {
      chomp;
      if ( $_ =~ /^Drive/ )
      {
          print "\n$_  [ Pct Used]\n";
      }
      elsif ( $_ =~ /^\s\s\w\$\s/ )
      {
          ( undef, undef, $total, $free ) = split( ' ', $_ , 4 );
          $used = $total - $free;
          $pct_used = ( $used / $total ) * 100;
          print "$_";
          printf ("      %3.1f%\n",$pct_used);
      }
      elsif ( $_ =~ /^Network / )
      {
          last;
      }
  }

  print "\n";
  exit;
}

# --------------------------------------------------------------------
# -lw
# --------------------------------------------------------------------
sub lw_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-lw" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ( $^O =~ /^(MS)?Win/ )
  {
      print "Sorry - Task Not Supported On Windows Platform\n";
      exit 1;
  }
  
  # Save BDF list
  @unusedvols = `$P.pl -lv | grep '(Unused)' | egrep -v 'vg00|vg01|vg02' | sort | awk '{ print \$1 }'`;
  #print "num=$#unusedvols\n";

  # Find 1st unused disknn name
  chop($l = `ll -d /var/opt/oracle/rdisk* 2>&1`);
  if ($l eq "/var/opt/oracle/rdisk* not found")
  {
    $highdisk = 0;
  }
  else
  {
    chop($l = `ll -d /var/opt/oracle/rdisk* | tail -1 | awk '{ print \$NF}'`);
    $highdisk = substr($l,length($l)-2);
  }
  #print "$P | l=$l highdisk=$highdisk\n";
  print "$P | Highest Existing rdisk Link=$highdisk\n\n";
  
  foreach $lv (@unusedvols)
  {
    chop $lv;
    ##print "$P | Found unused volume: $lv\n";
    $thisdisk = ++$highdisk;
    if (length($thisdisk) < 2)
    {
      $thisdisko = "0" . $thisdisk;
    }
    else
    {
      $thisdisko = $thisdisk;
    }
    $lv =~ s/lvol/rlvol/g; 
    print "ln -s $lv /var/opt/oracle/rdisk$thisdisko\n";
  }
  exit;
}

# --------------------------------------------------------------------
# -m -ma
# --------------------------------------------------------------------
sub m_or_ma_task
{
  my( $package, $file, $line ) = caller;
  unless ( $arg eq "-m" or $arg eq "-ma" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( -e $alrtfile )
  {
       print "$P | Unable to locate alert log file!\n";
       exit;
  }

  @alrtlog = `type $alrtfile`;

  print "$P | Alert Log: $alrtfile\n";
  print "$P | ================================================================\n";
  if ($arg eq "-m")
  {
      $cmd = "tail -20 $alrtfile";
      $startat = ( scalar( @alrtlog ) > 20 ) ? scalar( @alrtlog ) - 20 : 0;
      for( $i = $startat; $i < scalar( @alrtlog ); $i++ )
      {
          print $alrtlog[ $i ];
      }
  }
  else
  {
      $cmd = "type $alrtfile";
      for( $i = 0; $i < scalar( @alrtlog ); $i++ )
      {
          print $alrtlog[ $i ];
      }
  }

  print "$P | ================================================================\n";
  print "$P | Command=$cmd\n";

  exit;
}

# --------------------------------------------------------------------
# -mf
# --------------------------------------------------------------------
sub mf_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-mf" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( -e $alrtfile )
  {
       print "$P | Unable to locate alert log file!\n";
       exit;
  }

  print "$P | Alert Log: $alrtfile\n";
  print "$P | ================================================================\n";
  $cmd = "tail -90f $alrtfile";
  system($cmd);
  print "$P | ================================================================\n";
  print "$P | Command=$cmd\n";

  exit;
}

# --------------------------------------------------------------------
# -mr
# --------------------------------------------------------------------
sub mr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-mr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( -e $alrtfile )
  {
       print "$P | Unable to locate alert log file!\n";
       exit;
  }

  $howmuch = $arg2;

  &msg("Alert Log: $alrtfile Reformatted");
  &msg("================================================================");

  $howmuch = 20 if ! $howmuch;

  &alertLogReformat( $alrtfile, $howmuch );

  &msg("================================================================");
  exit;
}

#------------------------------------------------------------------------------
sub alertLogReformat
#------------------------------------------------------------------------------
{
  $alr_file  = $_[0];
  $alr_lines = $_[1];

  open(ALR, "$alr_file") || die "Cannot open input $alr_file - $!";

  $alr_buffer = "";
  $alr_seq = 0;

  %alr_month_hash = (
    "Jan" => "01",
    "Feb" => "02",
    "Mar" => "03",
    "Apr" => "04",
    "May" => "05",
    "Jun" => "06",
    "Jul" => "07",
    "Aug" => "08",
    "Sep" => "09",
    "Oct" => "10",
    "Nov" => "11",
    "Dec" => "12"
  );

  while(<ALR>)
  {
    my @words = split;
    if ($#words == 4 && $words[4] =~ /^2/)
    {
      $alr_ts = $words[4] . "/" . $alr_month_hash{$words[1]} . "/" . $words[2] . "-" . $words[3];
      $alr_seq = 0;
      next;
    }

    $alr_seq++;
    $alr_line    = "$alr_ts [$alr_seq] $_";
    $alr_buffer .= $alr_line;
    push (@alr_array,$alr_line);
  }
  close(ALR);

  if (uc $alr_lines eq "A")
  {
    &msg("Displaying entire alert log ($#alr_array+1 lines)");
    print "\n"; 
    print $alr_buffer; 
    print "\n"; 
  }
  if (uc $alr_lines eq "V")
  {
    &msg("vi-ing entire alert log ($#alr_array+1 lines)");
    $editor_cmd = ( defined $ENV{"EDITOR"} ) ? $ENV{"EDITOR"} : $default_editor;
    open(OUTDATA, ">$tmpfile") || die "Cannot open output: $tmpfile - $!";
    print OUTDATA $alr_buffer; 
    close(OUTDATA);
    system("$editor_cmd $tmpfile"); 
  }
  elsif ($alr_lines > 0 && $alr_lines < 99999999)
  {
    &msg("Displaying the last $alr_lines lines of alert log");
    print "\n"; 
    for ($i = ($#alr_array - $alr_lines); $i <= $#alr_array; $i++)
    {
      ##print "i=$i cnt=$#alr_array\n";
      print $alr_array[$i]; 
    }
    print "\n"; 
  }
}

# --------------------------------------------------------------------
# -mt
# --------------------------------------------------------------------
sub mt_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-mt" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    SELECT 
      \'X\',
      type, 
      paddr,
      queued, 
      decode(totalq,0,0,wait / totalq) avg_wait 
    FROM 
      v\$queue;
    order by
      2, 3
      ';

  print "\n";
  &msg("MTS Statistics"); 
  print "\n";

  print "Queue        PAddr            Ses Q'd  Avg Wait\n";
  print "------------ ---------------- -------- --------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $mt_ptype    = shift @words;
      $mt_paddr    = shift @words;
      $mt_queued   = shift @words;
      $mt_avgwait  = sprintf("%.2f",shift @words);

      $mt_paddr = "" if $mt_paddr eq "00";

      printf "%-12s %-16s %8s %8s\n",$mt_ptype, $mt_paddr, $mt_queued, $mt_avgwait;
    }
  }

  $sqlstmt = '
    SELECT 
      \'X\',
      d.name,
      s.username, 
      c.status,
      c.queue,
      s.sid,
      s.serial# 
    from
      v\$circuit c,
      v\$dispatcher d,
      v\$session s
    where 
      c.dispatcher = d.paddr and
      c.saddr = s.saddr
    order by 
      d.name, 
      s.username;
  ';

  print "\n";
  &msg("Connected MTS Users"); 
  print "\n";

  print "Disp     User             Session# Serial#  Status       Queue\n";
  print "-------- ---------------- -------- -------- ------------ -----------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $mt_dname    = shift @words;
      $mt_user     = shift @words;
      $mt_status   = shift @words;
      $mt_queue    = shift @words;
      $mt_sid      = shift @words;
      $mt_ser      = shift @words;

      printf "%-8s %-16s %8s %8s %-12s %-12s\n",$mt_dname, $mt_user,  $mt_sid, $mt_ser, $mt_status, $mt_queue;
    }
  }

  $sqlstmt = '
    SELECT 
      \'X\',
      name,
      busy / (busy + idle) * 100 
    from
      v\$shared_server 
    order by 
      2;
  ';

  print "\n";
  &msg("Shared Server Process Utilization"); 
  print "\n";

  print "Server       Busy\n";
  print "-------- --------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $mt_sname    = shift @words;
      $mt_busy     = sprintf("%.2f",shift @words);

      printf "%-8s %8s%%\n",$mt_sname, $mt_busy;
    }
  }

  $sqlstmt = '
    SELECT 
      \'X\',
      name,
      busy / (busy + idle) * 100 
    from
      v\$dispatcher 
    order by 
      2;
  ';

  print "\n";
  &msg("Dispatcher Process Utilization"); 
  print "\n";

  print "Server       Busy\n";
  print "-------- --------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $mt_sname    = shift @words;
      $mt_busy     = sprintf("%.2f",shift @words);

      printf "%-8s %8s%%\n",$mt_sname, $mt_busy;
    }
  }

  exit;
}

# --------------------------------------------------------------------
# -mv
# --------------------------------------------------------------------
sub mv_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-mv" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( -e $alrtfile )
  {
       print "$P | Unable to locate alert log file!\n";
       exit;
  }

  $editor_cmd = ( defined $ENV{"EDITOR"} ) ? $ENV{"EDITOR"} : $default_editor;
  print "$P | Alert Log: $alrtfile\n";
  print "$P | ================================================================\n";
  system( "$editor_cmd $alrtfile" );
  print "$P | ================================================================\n";
  print "$P | Command=$editor_cmd\n";
  exit;
}

# --------------------------------------------------------------------
# -nf
# --------------------------------------------------------------------
sub nf_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-nf" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
 
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  #
  # Look for tables where the next extent will not fit in the largest
  #   freespace hole for their tablespace
  #
  $sqlstmt = '
    select
      \'X\',
      segment_owner,
      segment_name,
      segment_type,
      ts1,
      max_chunk,
      chunk_needed
    from
      (select
	 tablespace_name ts1,
	 max(bytes) as max_chunk
       from dba_free_space
       group by tablespace_name),
      (select
	 owner segment_owner,
	 segment_name,
	 segment_type,
	 tablespace_name ts2,
	 next_extent chunk_needed
	 from dba_segments)
    where
      ts1=ts2
      and chunk_needed> max_chunk
    order by
      2,3
  ';
  &oraSQL($sqlstmt);

  $foundtbls = 0;
  $titleflag = 0;

  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X") 
    { 
      $owner           = $words[1];
      $sname           = $words[2];
      $stype           = $words[3];
      $ts              = $words[4];
      $ts_max_free     = $words[5];
      $s_next_size     = $words[6];

      $foundtbls++;
      $tbl = $owner . "." . $sname;

      if (!$titleflag)
      {
	print "\n"; 
	print "                          Object(s) Whose Next_Extent Will Not Fit\n"; 
	print "\n"; 
	print "Segment                                  Segment Tablespace               Object Next  TS Max Contig.\n"; 
	print "Name                                     Type    Name                     Extent Size      Free Space\n"; 
	print "---------------------------------------- ------- -------------------- --------------- ---------------\n"; 
	$titleflag = 1;
      }
      printf "%-40s %-7s %-20s %15i %15i\n",$tbl,$stype,$ts,$s_next_size,$ts_max_free; 
      ##$newnext = int ( $ts_max_free / 1000000 ); 
      ##$newnext-- if $newnext > 2;
      $newnext = $ts_max_free - 1024;
      $newnext = 1024 if $newnext <= 0;
      $ddl .= "ALTER $stype $tbl STORAGE ( NEXT ${newnext} PCTINCREASE 0 );\n";
    }
  }
  print "\n";
  if ($foundtbls == 0)
  {
    &msg("No objects found whose NEXT EXTENT will not fit");
  }
  else
  {
    &msg("Found $foundtbls object(s) whose NEXT EXTENT will not fit");
    &msg("DDL to correct:\n");
    print "$ddl \n";
  }
  exit;

}

# --------------------------------------------------------------------
# -nl
# --------------------------------------------------------------------
sub nl_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-nl" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $found = 0;
  $sqlstmt = '
    select \'X\', 
      t.owner, 
      t.table_name, 
      nvl(t.logging,\'(Null-L)\'),
      nvl(t.tablespace_name,\'(Null-TS)\'),
      o.object_type
    from 
      dba_tables t,
      dba_objects o
    where
      t.table_name = o.object_name    
      and t.owner  = o.owner
      and t.logging = \'NO\'
      and t.tablespace_name is not null
    order by 2,3;
    select \'X\', 
      i.owner, 
      i.index_name, 
      nvl(i.logging,\'(Null-L)\'),
      nvl(i.tablespace_name,\'(Null-TS)\')
      -- o.object_type
    from 
      dba_indexes i
      -- dba_objects o
    where
      -- i.index_name = o.object_name    
      --and i.owner  = o.owner
      i.logging = \'NO\'
      and i.tablespace_name is not null
    order by 2,3;
  ';

  print "\n";
  print "Owner.Object                             Object Type          TS                   LOGGING \n";
  print "---------------------------------------- -------------------- -------------------- --------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $owner   = $words[1];
      $obj     = $words[2];
      $logging = $words[3];
      $ts      = $words[4];

      $objtype = "";
      for ($i = 5; $i <= $#words; $i++)
      {
        $objtype .= $words[$i];
        $objtype .= " ";
      }
      $objtype = "INDEX" if $objtype eq "";

      $both = $owner . "." . $obj;
      if ($owner ne "SYS" && $owner ne "SYSTEM")
      {
        printf("%-40s %-20s %-20s %-8s\n",$both,$objtype,$ts,$logging);
        $found++;
      }
    }
  }
  print "\n";
  &msg("Found $found NOLOGGING Objects");
  exit;
}

# --------------------------------------------------------------------
# -ob
# --------------------------------------------------------------------
sub ob_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ob" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
select
  \'X\',
  owner,
  object_name,
  object_id,
  status
from
  dba_objects
where 
  owner not like \'SYS%\'
  and owner != \'PUBLIC\'
order by
  2,3;';

  &msg("-ob: Objects");
  print "Object                                   Status     ObjId-Dec  Objid-Hex \n";
  print "---------------------------------------- ---------- ---------- ----------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];

    if ($verb eq "X")
    {
      $owner  = $words[1];
      $name   = $words[2];
      $id     = $words[3];
      $status = $words[4];
      $obj = "$owner.$name";
      printf "%-40s %-10s %10s %-10X\n",$obj,$status,$id,$id;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -op
# --------------------------------------------------------------------
sub op_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-op" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
       exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( $oraver_num >= 8 )
  {
      &msg( "Not Supported On Oracle $oraver" );
      exit;
  }

  print "Name                       File      Class        XNC\n";
  print "-------------------- ---------- ---------- ----------\n";

  $sqlstmt = '
    select 
       \'Z\', 
       name, file#, class#,
       max(xnc)
       from v$ping 
       group by 
       \'Z\', name, file#, class# order by 5;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Z")
    {
      $name  = $words[1];
      $file  = $words[2];
      $class = $words[3];
      $xnc   = $words[4];
      printf("%-20s %10s %10s %10s\n",$name,$file,$class,$xnc);
      $total += $xnc
    }
  }
  print  "==================== ========== ========== ==========\n";
  printf("Total:                                      %9d\n",$total);             
  exit;
}

# --------------------------------------------------------------------
# -pb
# --------------------------------------------------------------------
sub pb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  &msg("A 20 second sample will be taken now");
  $sqlstmt = "
    spool $tmpfile;
    set echo on;
    WHENEVER SQLERROR CONTINUE;
    drop table buffer_hit_metrics_tab;
    WHENEVER SQLERROR EXIT 1;
    create table buffer_hit_metrics_tab
    (sample_date date,
     pr1 number,
     pr2 number,
     delta_pr number,
     cg1 number,
     cg2 number,
     delta_cg number,
     bg1 number,
     bg2 number,
     delta_bg number,
     cum_ratio number,
     delta_ratio number);

    CREATE OR REPLACE procedure buffer_hit_metrics
    (sample_rate number, iterations number)
    as

    pr1 number;
    pr2 number;
    delta_pr number;
    cg1 number;
    cg2 number;
    delta_cg number;
    bg1 number;
    bg2 number;
    delta_bg number;

    procedure  wait (sample_rate number)
    is

    now        date;
    end_time   date;

    begin
      end_time := sysdate + sample_rate/(24*60*60);
      loop
      now := sysdate;
      if now > end_time then
	return;
      end if;
      end loop;
    end;

    begin

    select pr.value,
	   cg.value,
	   bg.value
      into pr1, cg1, bg1
      from v\$sysstat pr, v\$sysstat bg, v\$sysstat cg
     where pr.name = 'physical reads'
       and bg.name = 'db block gets'
       and cg.name = 'consistent gets';

    insert into buffer_hit_metrics_tab
       (sample_date,pr1,cg1,bg2,cum_ratio)
       values
       (sysdate,pr1,cg1,bg1,round((1-(pr1/(bg1+cg1)))*100,2));

    wait (sample_rate);

    for i in 1 .. iterations loop

    pr2 := pr1;
    cg2 := cg1;
    bg2 := bg1;

    select pr.value,
	   cg.value,
	   bg.value
      into pr1, cg1, bg1
      from v\$sysstat pr, v\$sysstat bg, v\$sysstat cg
     where pr.name = 'physical reads'
       and bg.name = 'db block gets'
       and cg.name = 'consistent gets';

    delta_pr := pr1-pr2;
    delta_cg := cg1-cg2;
    delta_bg := bg1-bg2;

    insert into buffer_hit_metrics_tab
       values
       (sysdate, pr1, pr2, delta_pr,
		 cg1, cg2, delta_cg,
		 bg1, bg2, delta_bg,
		 round((1-(pr1/(bg1+cg1)))*100,2),
		 round((1-(delta_pr/(delta_bg+delta_cg)))*100,2));
    commit;

    wait (sample_rate);
    end loop;
    end;
/

   -- Execute samples - 10 samples of 2 seconds
   execute buffer_hit_metrics (5, 4);
   -- Run query of results
   alter session set nls_date_format='mm/dd/yy:hh24:mi:ss';
   set linesize 5000;
   select 'X', 
pr1 ,
pr2 ,
delta_pr ,
cg1 ,
cg2 ,
delta_cg ,
bg1 ,
bg2 ,
delta_bg ,
cum_ratio ,
delta_ratio 
     from buffer_hit_metrics_tab 
     order by sample_date;
   -- Drop objects 
   WHENEVER SQLERROR CONTINUE;
   drop table buffer_hit_metrics_tab;
   drop procedure buffer_hit_metrics;
   WHENEVER SQLERROR EXIT 1;
    ";

  &oraSQL($sqlstmt);
  $ratio = "?";
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $ratio  = $words[10];
    }
  }

  if ( $ratio ne '?' )
  {
      printf "\nCurrent_db_block_buffer_Hit_Ratio: %-8s\n",$ratio;
  }
  else
  {
      print "$P | Unable to determine Current Data Block Buffer Hit Ratio Now\n";
      print "$P | <SQL Commands>\n";
      system( "type $tmpfile" );
      print "$P | <SQL Output>\n";
      foreach ( @oraSQLout ) { print "$P | $_"; }
  }
  
  exit;
}

# --------------------------------------------------------------------
# -pc
# --------------------------------------------------------------------
sub pc_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pc" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', username
    from dba_users
    order by 2;';
  
  print "\nUser accounts with username=password\n";
  print "------------------------------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words      = split();
    $verb       = $words[0];
    $user       = $words[1];
    if ($verb eq 'X')
    {
      @mySQLout = `echo SELECT * FROM DUAL; | $sqlpluscmd -s $user/$user`;
      foreach (@mySQLout)
      {
        @string = split();
        $result = $string[0];
        if ($result eq 'X')
        {
          print "$user\n";
        }
      }
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -pd
# --------------------------------------------------------------------
sub pd_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pd" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
select \'X\',
       sum(decode(name, \'consistent gets\', value, 0)) consistent,
       sum(decode(name, \'db block gets\',   value, 0)) db_blockgets,
       sum(decode(name, \'physical reads\',  value, 0)) physical,
       round(((sum(decode(name, \'consistent gets\', value, 0)) +
               sum(decode(name, \'db block gets\',   value, 0)) -
               sum(decode(name, \'physical reads\',  value, 0))) /
              (sum(decode(name, \'consistent gets\', value, 0)) +
               sum(decode(name, \'db block gets\',   value, 0)))) * 100, 2)
  hitratio
from v\$sysstat;
  select \'Y\',
    name, value 
    from v\$parameter 
    where 
    name in (\'db_block_size\',\'db_block_buffers\',\'db_cache_size\');
 select
    \'V\', round((sum(gethits) / sum(gets) * 100),2) from v\$librarycache;
 select
    \'U\', round((sum(gets) - sum(getmisses)) / sum(gets) * 100,2) from v\$rowcache;
 select
    \'W\', 
    name, 
    100* (1-(physical_reads / (DB_BLOCK_GETS + CONSISTENT_GETS))),
    set_msize
  from 
    v\$buffer_pool_statistics 
  where
    (db_block_gets + consistent_gets) > 0
  order by 1; 
 select
    \'Z\', 
    100* (1-(sum(physical_reads) / (sum(DB_BLOCK_GETS) + sum(CONSISTENT_GETS))))
  from 
    v\$buffer_pool_statistics 
  where
    (db_block_gets + consistent_gets) > 0
    and name != \'RECYCLE\'
  order by 1; 
  ';

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Z")
    {
      $cachehit_x_recycle = $words[1];
    }
    if ($verb eq "V")
    {
      printf "Library Cache Hit Ratio:    %7.2f%\n",$words[1];
    }
    if ($verb eq "U")
    {
      printf "Dictionary Cache Hit Ratio: %7.2f%\n\n",$words[1];
    }
    if ($verb eq "X")
    {
      &msg("Data Block Buffer Hit Ratio Data");
      print "\n";
      ##print "$words[1]\n";
      printf "Consistent Gets:   %15s\n",$words[1];
      printf "DB Block Gets:     %15s\n",$words[2];
      printf "Physical Reads:    %15s\n",$words[3];
      printf "\nData Block Buffer Hit Ratio:%7.2f%\n",$words[4];

      $db_block_gets = $words[2];
      if ($db_block_gets =~ /1844/)
      {
        print "\n";
        print "*****************************************************************************\n";
        print "  NOTE: db_block_gets value of $db_block_gets is invalid due to an\n";
        print "    Oracle bug. Therefore, the above Data Block Buffer Hit Ratio is\n";
        print "    invalid. The 8i+ Pool Hit Ratio numbers below are more accurate.\n";
        print "*****************************************************************************\n";
        print "\n";
      }
    }
    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        $bs = $words[2];
      }
      if ($words[1] eq "db_block_buffers")
      {
        $bb = $words[2];
      }
      if ($words[1] eq "db_cache_size")
      {
        $bc = $words[2];
      }
    }
    if ($verb eq "W")
    {
      $pl = $words[1];
      $ph = sprintf("%6.2f",$words[2]);
      $ps = int(($words[3]*$bs) / 1000000);
      printf "DB_BLOCK_BUFFER Read Hit Ratio - Pool: %-8s  Hit_Ratio: %6.2f  Pool_Size: %5s (MB)\n",$pl,$ph,$ps;
    }
  }
  print "\n";

  printf "DB Block Size:          %10d\n",$bs;

  if ($bb > 0)
  {
    printf "DB Block Buffers:       %10d\n\n",$bb;
    $cache_mem = $bs*$bb/1000000;
  }
  elsif ($bc > 0)
  {
    printf "db_cache_size:          %10.0f\n\n",$bc;
    $cache_mem = $bc/1000000;
  }

  printf "Total DB Block Buffer Size: %6d(MB) (This Oracle Instance)\n",$cache_mem;

  $shmused = 0;
  open (SM,"C:\\dbatools\\pstat.exe |");
  while (<SM>)
  {
     s/\n//;
     s/^\s+//;
     $line = $_;
     @words = split;
     if (uc(${words[-1]}) eq "ORACLE.EXE" &&
        ! (${words[0]} =~ "pid:"))
     {
        $shmused += @words[2];
     }
     if ($shmfree eq "")
     {
        $w_cnt = -1;
        foreach $word (@words)
        {
           $w_cnt+=1;
           if (lc($word) =~ "avail:")
           {
              $_ = $word;
              s/Avail://;
              $shmfree = $_;
              if ($shmfree eq "" && !(${words[$w_cnt + 1]} =~ "TotalWs"))
              {
                 $shmfree = @words[$w_cnt + 1];
              }
           }
        }
     }
  }
  close (SM);

  $shmused = $shmused / 1024.0;
  $shmfree_label = substr($shmfree,-1,1);
  if (uc($shmfree_label) eq "K")
  {
     chop($shmfree);
     $shmfree = $shmfree / 1024.0;
  }

  printf " Server Shared Memory Used: %6d(MB) (All Oracle Instances)\n",$shmused;
  printf " Server Shared Memory Free: %6d(MB) (Approximate)\n\n",$shmfree;

  exit;
}

# --------------------------------------------------------------------
# -pf
# --------------------------------------------------------------------
sub pf_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pf" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select
    'X',
    (a.count/sum(b.value))*100 free_list
  from
    v\$waitstat a,
    v\$sysstat b
  where
    a.class = 'free list'
    and b.name in ('db block gets','consistent gets')
  group by
    a.count;";

  print "\n";
  print "% Of Data Requests That\n";
  print "  Waited for a Freelist\n";
  print "-----------------------\n";
  
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $pct = $words[1] + 0;
      $pct = sprintf("%7.2f",$pct);
      printf "%23s\n", $pct;
    }
  }
  print "\n";

  if ($pct < 1)
  {
    &msg("Since the percent of total request that encountered a freelist");
    &msg("wait is less than 1, you don't need to add freelists.");
  }
  if ($pct >= 1)
  {
    &msg("Since the percent of total request that encountered a freelist");
    &msg("wait is >= l, consider adding freelists.");
  }
  exit;
}

# --------------------------------------------------------------------
# -ph
# --------------------------------------------------------------------
sub ph_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ph" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  unless( $oraver_num >= 8 )
  {
      &msg( "Not Supported On Oracle $oraver" );
      exit;
  }

  $num = 200;
  $sqlstmt = "select /*+ ordered */ 'X',
      e.owner ||'.'|| e.segment_name  segment_name,
      e.extent_id  extent#,
      x.dbablk - e.block_id + 1  block#,
      x.tch,
      l.child#,
      l.sleeps
    from
      sys.v\$latch_children  l,
      sys.x\$bh  x,
      sys.dba_extents  e
    where
      l.name    = 'cache buffers chains' and
      l.sleeps  > $num and
      x.tch     > $num and
      x.hladdr  = l.addr and
      e.file_id = x.file# and
      x.dbablk between e.block_id and e.block_id + e.blocks - 1
    order by
      7 desc, 5 desc;";

  ##print "sqlstmt=$sqlstmt\n";
  &oraSQL("$sqlstmt");

  print "\n";
  print "  Sleeps     Hits Extent   Block# Children Segment\n";
  print "-------- -------- ------ -------- -------- ------------------------------\n";

  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $segment  = $words[1];
      $extent   = $words[2];
      $blocknum = $words[3];
      $tch      = $words[4];
      $childnum = $words[5];
      $sleeps   = $words[6];
      printf("%8s %8s %6s %8s %8s %-50s\n",$sleeps,$tch,$extent,$blocknum,$childnum,$segment);
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -pi
# --------------------------------------------------------------------
sub pi_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pi" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "\n";

  $stars = "**********";
  $gt = 0;

  $sqlstmt = "
    select 'Y1',
      sum(phyblkrd + phyblkwrt)
      from
      v\$filestat;
    select 'Y2',
      sum(phyblkrd + phyblkwrt)
      from
      v\$tempstat;
    select 'X',
      d.name filename,
      f.phyblkrd blocks_r,
      f.phyblkwrt block_w,
      f.phyblkrd + f.phyblkwrt total_io
      from
      v\$filestat f,
      v\$datafile d
    where 
      f.file# = d.file#
    union all
    select 'X',
      d.name filename,
      f.phyblkrd blocks_r,
      f.phyblkwrt block_w,
      f.phyblkrd + f.phyblkwrt total_io
      from
      v\$tempstat f,
      v\$tempfile d
    where 
      f.file# = d.file#
    order by
      5 desc;
    ";

  print "                                                                                  Pct Of\n";
  print "                                                             1000's 1000's 1000's DB I/O\n";
  print "Datafile                                                      Reads Writes    R+W  Total\n";
  print "------------------------------------------------------------ ------ ------ ------ ------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y1" || $verb eq "Y2")
    {
      $gt += int($words[1]/1000);
    }
    if ($verb eq "X")
    {
      $df = $words[1];
      $r  = int($words[2]/1000);
      $w  = int($words[3]/1000);
      $t  = int($words[4]/1000);
      $pct = $t / $gt * 100;
      $pct1 = int($t / $gt * 10) + 1;
      $hist = substr($stars,0,$pct1);
      printf "%-60s %6d %6d %6d %5.2f%% %-10s\n",$df,$r,$w,$t,$pct,$hist;
      $t_r += $r;
      $t_w += $w;
      $t_t += $t;
    }
  }
  printf "                                                             ====== ====== ======\n";
  printf "(Total):                                                     %6d %6d %6d\n",$t_r, $t_w, $t_t;
  exit;
}

# --------------------------------------------------------------------
# -pj
# --------------------------------------------------------------------
sub pj_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pj" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $len = $arg2;
  $len = 22 if $len eq "";

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - len=$len");
  print "\n";

  $stars = "**********";

  $sqlstmt = "
    select 'Y1',
      sum(phyblkrd + phyblkwrt)
      from
      v\$filestat;
    select 'Y2',
      sum(phyblkrd + phyblkwrt)
      from
      v\$tempstat;
    select 'X',
      substr(d.name,1,$len),
      sum(f.phyblkrd),
      sum(f.phyblkwrt),
      sum(f.phyblkwrt + f.phyblkrd),
      'DF'
    from
      v\$filestat f,
      v\$datafile d
    where 
      f.file# = d.file#
    group by 
      substr(d.name,1,$len)
    union all
    select 'X',
      substr(d.name,1,$len),
      sum(f.phyblkrd),
      sum(f.phyblkwrt),
      sum(f.phyblkwrt + f.phyblkrd),
      'TF'
    from
      v\$tempstat f,
      v\$tempfile d
    where 
      f.file# = d.file#
    group by 
      substr(d.name,1,$len)
    order by 
      5 desc;
    ";

  print "                                                                                                 Pct Of\n";
  print "                                                                      1000's    1000's    1000's DB I/O\n";
  print "Type  Filesystem                                                       Reads    Writes       R+W  Total\n";
  print "------------------------------------------------------------------ --------- --------- --------- ------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y1" || $verb eq "Y2")
    {
      $gt += int($words[1]/1000);
    }

    if ($verb eq "X")
    {
      $fs = $words[1];
      $r  = int($words[2]/1000);
      $w  = int($words[3]/1000);
      $t  = int($words[4]/1000);
      $ft = $words[ 5 ];

      $pct = $t / $gt * 100;
      $pct1 = int($t / $gt * 10) + 1;

      $hist = substr($stars,0,$pct1);

      printf "%-6s%-60s %9d %9d %9d %5.2f%% %-10s\n",$ft,$fs,$r,$w,$t,$pct,$hist;
      $t_r += $r;
      $t_w += $w;
      $t_t += $t;
    }
  }

  printf "                                                                   ========= ========= =========\n";
  printf "(Total):                                                           %9d %9d %9d\n",$t_r, $t_w, $t_t;
  exit;
}

# --------------------------------------------------------------------
# -pk
# --------------------------------------------------------------------
sub pk_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pk" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "\n";

  $stars = "**********";
  $sqlstmt = "
  select 'Y',
       sum(phyblkrd + phyblkwrt),
       sum(phyblkrd),
       sum(phyblkwrt)
       from
       v\$filestat;

  select 'Z',
       sum(phyblkrd + phyblkwrt),
       sum(phyblkrd),
       sum(phyblkwrt)
       from
       v\$tempstat;

  select 'X',
       t.name,
       sum(f.phyblkrd),
       sum(f.phyblkwrt),
       sum(f.phyblkwrt + f.phyblkrd)
    from
       v\$filestat f,
       v\$datafile d,
       v\$tablespace t
    where 
      f.file# = d.file# 
      and d.ts# = t.ts#
    group by 
      t.name
  union all
  select 'X',
       t.name,
       sum(f.phyblkrd),
       sum(f.phyblkwrt),
       sum(f.phyblkwrt + f.phyblkrd)
    from
       v\$tempstat f,
       v\$tempfile d,
       v\$tablespace t
    where 
      f.file# = d.file# 
      and d.ts# = t.ts#
    group by 
      t.name
    order by 
      5 desc
  ;";

  print "                                                     % Of  % Of    % Of\n";
  print "                        1000's    1000's    1000's   Read  Write    R+W\n";
  print "Tablespace               Reads    Writes       R+W  Total  Total  Total\n";
  print "-------------------- --------- --------- --------- ------ ------ ------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      $gt   = int($words[1]/1000);
      $gt_r = int($words[2]/1000);
      $gt_w = int($words[3]/1000);
    }
    if ($verb eq "Z")
    {
      $gt   += int($words[1]/1000);
      $gt_r += int($words[2]/1000);
      $gt_w += int($words[3]/1000);
    }
    if ($verb eq "X")
    {
      ##print "gt_r=$gt_r gt_w=$gt_w gt=$gt\n";

      $ts = $words[1];
      $r  = int($words[2]/1000);
      $w  = int($words[3]/1000);
      $t  = int($words[4]/1000);

      $pct = $t / $gt * 100;
      $pct1 = int($t / $gt * 10) + 1;
      $hist = substr($stars,0,$pct1);

      $pct_r = $r / $gt_r * 100;
      $pct_w = $w / $gt_w * 100;


      printf "%-20s %9d %9d %9d %5.2f%% %5.2f%% %5.2f%% %-10s\n",$ts,$r,$w,$t,$pct_r,$pct_w,$pct,$hist;
      $t_r += $r;
      $t_w += $w;
      $t_t += $t;
    }
  }
  printf "                     ========= ========= =========\n";
  printf "(Total):             %9d %9d %9d\n",$t_r, $t_w, $t_t;
  exit;
}

# --------------------------------------------------------------------
# -pp
# --------------------------------------------------------------------
sub pp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "                                                   MR I/O   \n";
  print "Datafile                                           Time (MS)\n";
  print "-------------------------------------------------- ---------\n";
  $sqlstmt = "
    select
      'X', 
      d.name,
      f.lstiotim*10 
    from 
       v\$filestat f,
       v\$datafile d
    where 
       f.file# = d.file#
       and lstiotim > 0
       and lstiotim != 65535
    union all
    select
      'X', 
      d.name,
      f.lstiotim*10 
    from 
       v\$tempstat f,
       v\$tempfile d
    where 
       f.file# = d.file#
       and lstiotim > 0
       and lstiotim != 65535
    order by
       3
     ;";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $df  = $words[1];
      $num = $words[2];
      printf "%-50s %9s\n",$df,$num;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -ps
# --------------------------------------------------------------------
sub ps_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ps" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "
select 'X', status, parallel from v\$instance;
select 'Y', inst_number, inst_name from v\$active_instances;
select 'Z', value, name from v\$dlm_misc;";

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $status  = $words[1];
      $par     = $words[2];
      print "\nParallel Server On: $par  Instance Status: $status\n";
      print "\nInstances in Cluster:\n\n";
      print "Node     Name\n";
      print "======== ==================\n";
    }
    if ($verb eq "Y")
    {
      $num     = "Node:" . $words[1];
      $name    = $words[2];
      printf "%-7s  %-30s\n",$num,$name;
    }
    if ($verb eq "Z")
    {
      if (!$zflag)
      {
	print "\nDLM Statistics:\n\n";
	print "Name                                 Value\n";
	print "==================================== =========\n";
        $zflag = 1;
      } 
      $value    = $words[1];
      $name     = $words[2] . " " . $words[3] . " " . $words[4] . " " . $words[5] . " " . $words[6];
      printf "%-36s %9d\n",$name,$value;
    }
  }
  exit; 
}

# --------------------------------------------------------------------
# -pw
# --------------------------------------------------------------------
sub pw_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-pw" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select distinct 'X', p1, p2 
  from v\$session_wait
  where 
    event = 'buffer busy waits';";

  &oraSQL("$sqlstmt");
  @saveSQLout1 = @oraSQLout;
  foreach (@saveSQLout1)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $filenum  = $words[1];
      $blocknum = $words[2];
      &msg("Found BUFFER BUSY WAIT - filenum=$filenum blocknum=$blocknum");
      $foundit = 1;
      
      # Find the table name
      &oraSQL("select 'X', file_name, tablespace_name from dba_data_files where file_id = $filenum;");
      @saveSQLout2 = @oraSQLout;
      foreach (@saveSQLout2)
      {
	##print;
	print if /^ORA/;
	@words   = split();
	$verb    = $words[0];
	if ($verb eq "X")
	{
          $filename  = $words[1];
          $tsname   = $words[2];
          &msg("  Datafile=$filenum Tablespace=$tsname Filename=$filename");
        }
      }

      # Find the table name

      # TEST 
      ##$filenum = 6; $blocknum = 35707;

      $sql = "select /*+ RULE */ 
        'X',
        owner, 
        segment_name, 
        segment_type,
        tablespace_name,
        block_id,
        blocks
      from 
        dba_extents 
      where 
        file_id = $filenum 
        and block_id <= $blocknum
        and $blocknum <= ((block_id + blocks) - 1);";
      ##print "sql=$sql\n";
      &oraSQL("$sql");
      @saveSQLout3 = @oraSQLout;
      foreach (@saveSQLout3)
      {
	##print;
	print if /^ORA/;
	@words   = split();
	$verb    = $words[0];
	if ($verb eq "X")
	{
          $owner   = $words[1];
          $segment = $words[2];
          $type    = $words[3];
          $ts      = $words[4];
          $blockid = $words[5];
          $blocks  = $words[6];
          &msg("  Segment=$owner.$segment Type=$type TS=$ts - block_id=$blockid blocks=$blocks");
        }
      }
    }
  }
  if (!$foundit) { &msg("No BUFFER BUSY WAITS found"); }
  exit;
}

# --------------------------------------------------------------------
# -qs
# --------------------------------------------------------------------
sub qs_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-qs" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $sqlstmt = '
    select 
      \'Y\', 
      sum(executions) 
    from 
      v\$sqlarea;
    select 
      \'X\', 
      command_type ,
      sum(users_executing) ,
      sum(executions) 
    from 
      v\$sqlarea 
    group by \'X\', command_type 
    order by 4 desc;';

  #&msg("sqlstmt=$sqlstmt");

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " Since Instance Start");
  print "SQL Statement Type    Executing Now     Total Executions\n";
  print "-------------------- -------------- -------------- -----\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb     = $words[0];
    if ($verb eq "Y")
    {
      $grant_tot = $words[1];
    }
    if ($verb eq "X")
    {
      $cmd_type = $words[1];
      $exec_now = $words[2];
      $exec_tot = $words[3];
      if    ($cmd_type ==  2) { $sql_type = "INSERT"; }
      elsif ($cmd_type ==  3) { $sql_type = "SELECT"; }
      elsif ($cmd_type ==  6) { $sql_type = "UPDATE"; }
      elsif ($cmd_type ==  7) { $sql_type = "DELETE"; }
      elsif ($cmd_type ==  9) { $sql_type = "CREATE_INDEX"; }
      elsif ($cmd_type == 10) { $sql_type = "DROP_INDEX"; }
      elsif ($cmd_type == 11) { $sql_type = "ALTER_INDEX"; }
      elsif ($cmd_type == 12) { $sql_type = "DROP_INDEX"; }
      elsif ($cmd_type == 15) { $sql_type = "ALTER_TABLE"; }
      elsif ($cmd_type == 16) { $sql_type = "DROP_SEQUENCE"; }
      elsif ($cmd_type == 17) { $sql_type = "GRANT"; }
      elsif ($cmd_type == 18) { $sql_type = "REVOKE"; }
      elsif ($cmd_type == 19) { $sql_type = "CREATE_SYNONYM"; }
      elsif ($cmd_type == 26) { $sql_type = "LOCK_TABLE"; }
      elsif ($cmd_type == 28) { $sql_type = "RENAME"; }
      elsif ($cmd_type == 35) { $sql_type = "ALTER_DATABASE"; }
      elsif ($cmd_type == 39) { $sql_type = "CREATE_TABLESPACE"; }
      elsif ($cmd_type == 40) { $sql_type = "ALTER_TABLESPACE"; }
      elsif ($cmd_type == 41) { $sql_type = "DROP_TABLESPACE"; }
      elsif ($cmd_type == 42) { $sql_type = "ALTER_SESSION"; }
      elsif ($cmd_type == 44) { $sql_type = "COMMIT"; }
      elsif ($cmd_type == 47) { $sql_type = "PL/SQL_EXECUTE"; }
      elsif ($cmd_type == 48) { $sql_type = "SET_TRANSACTION"; }
      elsif ($cmd_type == 49) { $sql_type = "ALTER_SYS_SW_LOG"; }
      elsif ($cmd_type == 62) { $sql_type = "ANALYZE_TABLE"; }
      elsif ($cmd_type == 63) { $sql_type = "ANALYZE_INDEX"; }
      else                    { $sql_type = "OTHER-$cmd_type"; }
      $pctof = 100 * ($exec_tot / $grant_tot);
      printf "%-20s %14.0f %14.0f %4.1f%%\n", $sql_type,$exec_now,$exec_tot,$pctof;
      $t_now += $exec_now;
      $t_tot += $exec_tot;
    }
  }
  
  print "-------------------- -------------- -------------- -----\n";
  printf "%-20s %14.0f %14.0f %4.1f%%\n", $zzzz,$t_now,$t_tot,$zzzz;
  exit;
}

# --------------------------------------------------------------------
# -ra
# --------------------------------------------------------------------
sub ra_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ra" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $sqlstmt = "select 'X', 
rn.NAME,
rs.EXTENTS,
rs.EXTENDS,
rs.RSSIZE,
rs.XACTS,
rs.GETS,
rs.HWMSIZE,
rs.WRAPS,
rs.AVEACTIVE,
rs.STATUS,
rs.WAITS,
rs.optsize
 from v\$rollstat rs, v\$rollname rn where rn.usn=rs.usn;";

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "RSSize  : Current rollback segment size in bytes\n";
  print "A Xn's  : Number of current transactions using the rollback segment\n"; 
  print "HWMSize : High Water Mark size in bytes\n"; 
  print "Wraps   : Cumulative # of times a X'n continues writing from one\n"; 
  print "          extents to another existing extent\n"; 
  print "Avg Actv: Avg # of bytes in active extents in the rbs measured over time\n\n"; 

  print "Name          Extents  Extends         RSSize   A Xn's      Gets         HWMSize  Wraps   Avg Actv  Status   Waits Optimal\n";
  print "------------ -------- -------- -------------- -------- ---------- -------------- ------  --------- -------- ------ -------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words     = split();
    $verb      = $words[0];
    $name      = $words[1];
    $extents   = $words[2];
    $extends   = $words[3];
    $rssize    = $words[4];
    $xacts     = $words[5];
    $num_gets  = $words[6];
    $hwmsize   = $words[7];
    $wraps     = $words[8];
    $aveactive = $words[9];
    $status    = $words[10];
    $waits     = $words[11];
    $optimal   = $words[12];
    if ($verb eq "X")
    {
      printf "%-12s %8d %8d %14d %8d %10d %14d %6s %10d %-8s %6d %-8s\n",
        $name,$extents,$extends,$rssize,$xacts,$num_gets,$hwmsize,$wraps,$aveactive,$status,$waits,$optimal;
      
      $t_waits += $waits;
      $t_gets  += $num_gets;
    }
    #if ($xacts > 1) { $xacts_flag = 1; }
    #if ($waits > 1) { $waits_flag = 1; }
  }

  print "\n";

  #if ($xacts_flag)
  #{
  #  &msg("Since at least one segment has >1 transaction running against it,"); 
  #  &msg("try increasing the number of rollback segments."); 
  #}
  #if ($waits_flag)
  #{
  #  &msg("Since at least one segment has >=1 transaction(s) waiting for it,"); 
  #  &msg("try increasing the number of rollback segments."); 
  #}

  $waits_gets_r = 100 * ($t_waits/$t_gets);
  $waits_gets_r = sprintf("%8.5f",$waits_gets_r);

  &msg("Total Waits:            $t_waits");
  &msg("Total Gets:             $t_gets");
  &msg("Ratio of Waits to Gets: $waits_gets_r%");

  if ($waits_gets_r > 1)
  {
    &msg("Because the ratio is > 1, add more rollback segments");

  }

  exit;
}

# --------------------------------------------------------------------
# -rb
# --------------------------------------------------------------------
sub rb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select 'X', 
s.SEGMENT_NAME,
s.OWNER,
s.TABLESPACE_NAME,
s.BYTES,
s.INITIAL_EXTENT,
s.NEXT_EXTENT,
s.MIN_EXTENTS,
s.MAX_EXTENTS,
s.PCT_INCREASE,
s.EXTENTS,
r.status,
r.owner
 from dba_segments s, dba_rollback_segs r
 where
 s.segment_type='ROLLBACK' and
 s.segment_name=r.segment_name;";

  print "Name               Tablespace    Extents   X-Init   X-Next    X-Min      X-Max Status   PUBLIC/SYS\n";
  print "------------------ ------------ -------- -------- -------- -------- ---------- -------- ----------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    $name    = $words[1];
    $owner   = $words[2];
    $ts      = $words[3];
    $bytes   = $words[4];
    $xinit   = $words[5];
    $xnext   = $words[6];
    $xmin    = $words[7];
    $xmax    = $words[8];
    $xpct    = $words[9];
    $extents = $words[10];
    $status  = $words[11];
    $owner   = $words[12];
    if ($verb eq "X")
    {
      printf "%-18s %-12s %8d %8d %8d %8d %10d %-8s %-10s\n",
        $name,$ts,$extents,$xinit,$xnext,$xmin,$xmax,$status,$owner;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -rc -rd
# --------------------------------------------------------------------
sub rc_or_rd_task
{
  my( $package, $file, $line ) = caller;
  unless ( $arg eq "-rc" or $arg eq "-rd" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select 'X', 
    SEGMENT_NAME,
    OWNER,
    TABLESPACE_NAME,
    SEGMENT_ID,
    INITIAL_EXTENT,
    NEXT_EXTENT,
    MIN_EXTENTS,
    MAX_EXTENTS,
    PCT_INCREASE,
    STATUS
    from dba_rollback_segs;";

  &msg("Rollback Segments");
  print "Name     Owner    Tablespace     X-Init   X-Next    X-Min    X-Max Status\n";
  print "-------- -------- ------------ -------- -------- -------- -------- ----------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    $name    = $words[1];
    $owner   = $words[2];
    $ts      = $words[3];
    $id      = $words[4];
    $xinit   = $words[5];
    $xnext   = $words[6];
    $xmin    = $words[7];
    $xmax    = $words[8];
    $xpct    = $words[9];
    $status  = $words[10];
    if ($verb eq "X")
    {
      printf "%-8s %-8s %-12s %8d %8d %8d %8d %-10s\n",
        $name,$owner,$ts,$xinit,$xnext,$xmin,$xmax,$status;
      if ($arg eq "-rc")
      {
        &msg("(Shrinking rollback segment: $name)");
        &oraSQLold( "alter rollback segment $name shrink;", $os_conn );
      }

      if ($arg eq "-rd")
      {
        &msg("(Altering rollback segment: $name MAXEXTENTS UNLIMITED)");
        &oraSQLold( "alter rollback segment $name storage ( maxentents unlimited );", $os_conn );
      }
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -rf
# --------------------------------------------------------------------
sub rf_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rf" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  
  $rows = 0;
  $min_time_c = "?";
  $min_time_n = "?";

  $sqlstmt = "
    SELECT 
      'X',
      min(to_char(time,'YYYY/MM/DD-HH24:MI:SS')) 
    FROM 
      v\$recover_file;";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      next if not defined $words[ 0 ];
      $min_time_c = shift @words;
      $min_time_n = &timeToNum($min_time_c);
     
      $rows++;
    }
  }
  &msg("Minimum TIME Value From V\$RECOVER_FILE (Character): $min_time_c");
  &msg("Minimum TIME Value From V\$RECOVER_FILE (Numeric):   $min_time_n");
  exit;
}

# --------------------------------------------------------------------
# -rg
# --------------------------------------------------------------------
sub rg_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rg" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  
  $sqlstmt = "
    SELECT 
      'X',
      version, 
      status, 
      comp_name
    FROM 
      dba_registry;
    order by
      1, 3
      ";

  print "\n";
  print "DB Component                             Version      Status\n";
  print "---------------------------------------- ------------ ------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;
      $rg_component = "";

      $rg_version   = shift @words;
      $rg_status    = shift @words;
      foreach (@words)
      {
        $rg_component .= "$_ ";
      }

      printf "%-40s %-12s %-12s\n",$rg_component,$rg_version,$rg_status;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -rh
# --------------------------------------------------------------------
sub rh_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rh" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "
    alter session set sort_area_size          = 100000000;
    alter session set sort_area_retained_size = 100000000;
    select 'X', 
    THREAD#, 
    SEQUENCE#, 
    FIRST_CHANGE#, 
    switch_change#,
    to_char(first_time,'MM/DD/YYYY-HH24:MI:SS')
    from v\$loghist
    order by 5 ;";

  print "\n";
  print "Thread   ArcLog  First Chg. Num   Last Chg. Num Timestamp      \n";
  print "------ -------- --------------- --------------- ------------------------\n";
  
  $diag = "";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    $diag .= $_;
    @words = split();
    $verb   = $words[0];
    $thread = $words[1];
    $seq    = $words[2];
    $fstchg = $words[3];
    $swchg  = $words[4];
    $fsttime= $words[5];
    if ($verb eq "X")
    {
      ##print "fsttime=$fsttime\n";
      printf "%6d %8d %15.0f %15.0f %-24s\n",
        $thread,$seq,$fstchg,$swchg,$fsttime;
      #last if ++$counter > 500;
      ++$counter;
    }
  }
  if ($counter < 1)
  {
    &msg("ERROR: No rows found");
    &msg("diag=$diag");
  }
  exit;
}

# --------------------------------------------------------------------
# -ro
# --------------------------------------------------------------------
sub ro_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ro" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select 'X', role from dba_roles order by 1;";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $role = $words[1];
      print "$role\n";
      $count++;
    }
  }
  &msg("Found $count roles");
  exit;
}

# --------------------------------------------------------------------
# -rp
# --------------------------------------------------------------------
sub rp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($oraver_num < 8.1)
  {
    $sqlstmt = "
    select 'Y', value from v\$parameter where name = 'db_block_size';
    select 'Z',
      sid,
      serial#,
      context,
      sofar,
      totalwork,
      current_time - start_time,
      to_char(start_time,'YYYY/MM/DD-HH24:MI:SS') ,
      to_char(current_time,'YYYY/MM/DD-HH24:MI:SS') ,
      compnam
    from
      v\$session_longops
    where
      compnam like 'RMAN%'
      and compnam not like 'RMAN: aggregate%'
    order by 8, 4;";
  }
  else
  {
      $sqlstmt = "
      select 'Y', value from v\$parameter where name = 'db_block_size';
      select 'Z', 
        sid, 
        serial#, 
        context, 
        sofar, 
        totalwork, 
        last_update_time - start_time, 
        --round(sofar/totalwork*100,2) 'PctComplete',
        to_char(start_time,'YYYY/MM/DD-HH24:MI:SS') ,
        to_char(last_update_time,'YYYY/MM/DD-HH24:MI:SS') ,
        opname
      from
        v\$session_longops 
      where 
        opname like 'RMAN%'
        and opname not like 'RMAN: aggregate%'
        --and sofar <> totalwork
      order by 8, 4;";
  }

  print "SID-Ser#  Start               Last Updated        El. Mins  GB/Hour Work Sofar Work Total Pct. Done  Desc.\n";
  print "--------- ------------------- ------------------- -------- -------- ---------- ---------- ---------- ------------------------\n";
  
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    $sid     = $words[1];
    $ser     = $words[2];
    #$context = $words[3];
    $sofar   = $words[4];
    $totalwk = $words[5];
    $etime   = sprintf("%5.1f",$words[6] * (24*60));
    $stime   = $words[7];
    $utime   = $words[8];
    $opname  = $words[9] . $words[10]  . $words[11]  . $words[12];

    if ($totalwk > 0) { $pctcomp = sprintf("%7.3f",($sofar/$totalwk)*100); }
    else              { $pctcomp = "*"; }

    $totalsize = $blocksize * $sofar;
    $gbrate = int((($totalsize / $etime) * 60) / (1024 * 1024 * 1024)) if $etime > 0;

    $sidser = "$sid-$ser";
    if ($verb eq "Y")
    {
      $blocksize = $words[1];
    }
    if ($verb eq "Z")
    {
      printf "%-9s %-19s %-19s %8s %8s %10s %10s %10s %-10s \n",$sidser,$stime,$utime,$etime, $gbrate, $sofar,$totalwk,$pctcomp,$opname;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -rr
# --------------------------------------------------------------------
sub rr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select \'X\',
    a.GROUP#,
    a.FIRST_CHANGE#,
    b.member,
    a.SEQUENCE#
  from
    v\$log a,
    v\$logfile b
  where a.group# = b.group#
    and a.status = \'CURRENT\';";

  &oraSQL($sqlstmt);

 
  print "Group 1stChange# Name                                                 Seq.\n";
  print "----- ---------- -------------------------------------------------- ------\n";

  foreach (@oraSQLout)
  {
    @words = split();
    $verb   = @words[0];
    $group  = @words[1];
    $fstchg = @words[2];
    $name   = @words[3];
    $seq    = @words[4];
    if ($verb eq "X")
    {
      ##print "fsttime=$fsttime\n";
      printf "%5d %10s %-50s %6d\n",
        $group,$fstchg,$name,$seq;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -rs
# --------------------------------------------------------------------
sub rs_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-rs" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', 
    GROUP#, THREAD#, SEQUENCE#, BYTES, MEMBERS, ARCHIVED, STATUS, FIRST_CHANGE#, FIRST_TIME
    from v\$log order by 3, 2;';

  print "Group Thread  Seq.      Bytes Mbrs. Archived Status     1st Chg.   1st Date        \n";
  print "----- ------ ----- ---------- ----- -------- ---------- ---------- ----------------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb   = $words[0];
    $group  = $words[1];
    $thread = $words[2];
    $seq    = $words[3];
    $bytes  = $words[4];
    $members= $words[5];
    $arched = $words[6];
    $status = $words[7];
    $fstchg = $words[8];
    $fsttime= $words[9];
    if ($verb eq "X")
    {
      ##print "fsttime=$fsttime\n";
      printf "%5d %6d %5d %10d %5d %-8s %-10s %10s %-16s\n",
        $group,$thread,$seq,$bytes,$members,$arched,$status,$fstchg,$fsttime;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -ru
# --------------------------------------------------------------------
sub ru_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ru" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "\n";

  $sqlstmt = 'select 
    \'X\',
    s.username, 
    s.sid,
    s.serial#,
    t.used_ublk,
    translate(t.start_time,\' \',\'-\')
  from 
    v\$transaction t,
    v\$session s
  where
    t.addr = s.taddr
  order by 5 desc;';

  print "DBUserid     SID    Serial  Undo Blks Trans. Started   \n";
  print "------------ ------ ------ ---------- -----------------\n";

  &oraSQL($sqlstmt);
  $cnt = 0;
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = shift(@words);
    if ($verb eq "X")
    {
      $user     = shift(@words);
      $sid      = shift(@words);
      $ser      = shift(@words);
      $blks     = shift(@words);
      $strttime = shift(@words);

      printf("%-12s %-6s %-6s %10s %-19s\n",$user,$sid,$ser,$blks,$strttime);

      $cnt++;
    }
  }
  print "\n";
  &msg("$cnt Total Active Transactions");
  exit;
}

# --------------------------------------------------------------------
# -sa 
# --------------------------------------------------------------------
sub sa_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sa" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
	   sum(vss.value),
           count(*), 
	   --substr(vsn.name,1,20) 
	   vsn.name
    from   
	   v\$sesstat vss,
	   v\$statname vsn
    where (vss.statistic#=vsn.statistic#) 
	  AND vss.value <> 0
    --group by substr(vsn.name,1,20)
    group by vsn.name
    order by 4;';
  
  print "Session Statistic                                                      Value Sessions     Per Session\n";
  print "------------------------------------------------------------ --------------- -------- ---------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words      = split();
    $verb       = $words[0];
    if ($verb eq 'X')
    {
      $value      = $words[1];
      $count      = $words[2];

      ##print "count=$count\n";

      $perses     = "";
      $perses     = int($value / $count) if $count > 0;

      $text = "";
      for ($i = 3; $i <= $#words; $i++)
      {
	$text .= $words[$i];
	$text .= " ";
      }
      printf "%-60s %15s %8s %15s\n",$text,$value,$count,$perses;
    }

  }
  exit;
}

# --------------------------------------------------------------------
# -sb
# --------------------------------------------------------------------
sub sb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    SELECT \'X\', high.thread#, "LowGap#", "HighGap#"FROM(
	SELECT thread#, MIN(sequence#)-1 "HighGap#"    FROM    (
	    SELECT a.thread#, a.sequence#        FROM        (            SELECT
    *
		FROM v\$archived_log        ) a,        (
		SELECT thread#, MAX(next_change#)gap1            FROM
    v\$log_history
		GROUP BY thread#        ) b        WHERE a.thread# = b.thread#
	    AND a.next_change# > gap1    )    GROUP BY thread#) high,(
	SELECT thread#, MIN(sequence#) "LowGap#"    FROM    (
	    SELECT thread#, sequence#        FROM v\$log_history, v\$datafile
	    WHERE checkpoint_change# <= next_change#
	    AND checkpoint_change# >= first_change#    )    GROUP BY thread#)
    low
    WHERE low.thread# = high.thread# AND "LowGap#" < "LowGap#";';

  print "       <=Missing Log Range=>\n";
  print "Thread        Low       High\n";
  print "------ ---------- ----------\n";
  
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $thread  = $words[1];
      $lo      = $words[2];
      $hi      = $words[3];
      printf "%6s %10s %10s\n", $thread, $lo, $hi;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -sc 
# --------------------------------------------------------------------
sub sc_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sc" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
       vs.sid,
      vs.serial#, 
      vs.type,
      vs.osuser, 
      sysdate - logon_time logon_time, 
      vs.process,
      vs.machine, 
      vs.username,
      vss.value,
      vs.program
from   v\$session vs,
       v\$sesstat vss,
       v\$statname vsn
where (vss.statistic#=vsn.statistic#) AND
      (vs.sid=vss.sid) AND
      (vsn.name = \'CPU used by this session\')
AND vss.value <> 0
order by 10;';

  print "                                                           ------------------------------------------Session------------------------------------------\n";
  print "SID-Ser#  OS:Userid-PID-Server     DB:User    Program       Hours  CPUTm Commit PFBlks TScans  MemSorts DiskSorts Con+BuGet PhysReads RHRat  LokWaitTm\n";
  print "--------- ------------------------ ---------- ------------ ------ ------ ------ ------ ------ --------- --------- --------- --------- ------ ---------\n";

  &oraSQL($sqlstmt);
  @saveSQLout = @oraSQLout;
  foreach (@saveSQLout)
  {
    @words      = split();
    $verb       = $words[0];
    $sid        = $words[1];
    $serial     = $words[2];
    $type       = $words[3];
    $osuser     = $words[4];
    $sessdur    = $words[5];
    $process    = $words[6];
    $machine    = $words[7];
    $username   = $words[8];
    $sesscpu    = $words[9];
    $program = $words[10] . $words[11] . $words[12] . $words[13];

    $sesser = $sid . "," . $serial;
    $comb = "$osuser-$process-$machine";
    $comb = substr($comb,0,24) if length($comb) > 24;
    $sdur = $sessdur * 24;
    $program = substr($program,0,12) if length($program) > 12;

    ##print "$sessdur $sdur\n";

    if ($verb eq "X")
    {
      # convert stat names into statistic#'s
      if ($statnumbers eq "")
      {
	$substmt = "
	  select distinct 'X', statistic#, name from v\$statname where name in (
	    'user commits',
	    'prefetched blocks',
	    'transaction lock foreground wait time',
	    'db block gets',
	    'consistent gets',
	    'physical reads',
	    'sorts (memory)',
	    'sorts (disk)',
	    'table scans (short tables)',
	    'table scans (long tables)'
	  ) order by 2;";
	&oraSQL($substmt);
	foreach (@oraSQLout)
	{
	  @words      = split();
	  $verb = $words[0];
	  if ($verb eq "X")
	  {
	    $statnumbers .= "$words[1],";

            $commits_code  = $words[1] if (/user commits/);
            $prefetch_code = $words[1] if (/prefetched blocks/);
            $locktime_code = $words[1] if (/transaction lock foreground wait time/);
            $blkgets_code  = $words[1] if (/db block gets/);
            $congets_code  = $words[1] if (/consistent gets/);
            $preads_code   = $words[1] if (/physical reads/);
            $sorts_m_code  = $words[1] if (/sorts \(memory\)/);
            $sorts_d_code  = $words[1] if (/sorts \(disk\)/);
            $scans_s_code  = $words[1] if (/table scans \(short tables\)/);
            $scans_l_code  = $words[1] if (/table scans \(long tables\)/);
	  }
	}
	chop $statnumbers;
	#print "statnumbers=$statnumbers\n";
	#print "blkgets_code=$blkgets_code\n";
	#print "prefetch_code=$prefetch_code\n";
	#print "locktime_code=$locktime_code\n";
	#print "congets_code=$congets_code\n";
	#print "preads_code=$preads_code\n";
	#print "sorts_m_code=$sorts_m_code\n";
	#print "sorts_d_code=$sorts_d_code\n";
	#print "scans_s_code=$scans_s_code\n";
	#print "scans_l_code=$scans_l_code\n";
      }
 
      # Get other stats
      #&scSub2($sid,"4,84,38,40,180,181,150,151");
      ##print "statnumbers=$statnumbers\n";
      &scSub2($sid,"$statnumbers");
      foreach(@scArray)
      {
        @subwords = split;
        #print "stat=$subwords[0] value=$subwords[1]\n";

	$s_locktime = $subwords[1] if $subwords[0] == $locktime_code;
	$s_commits  = $subwords[1] if $subwords[0] == $commits_code;
	$s_prefetch = $subwords[1] if $subwords[0] == $prefetch_code;
	$s_blkgets  = $subwords[1] if $subwords[0] == $blkgets_code;
	$s_congets  = $subwords[1] if $subwords[0] == $congets_code;
	$s_phreads  = $subwords[1] if $subwords[0] == $preads_code;
 	$s_sorts_m  = $subwords[1] if $subwords[0] == $sorts_m_code;
 	$s_sorts_d  = $subwords[1] if $subwords[0] == $sorts_d_code;
 	$s_scans_s  = $subwords[1] if $subwords[0] == $scans_s_code;
 	$s_scans_l  = $subwords[1] if $subwords[0] == $scans_l_code;

        #print "sc=$s_commits\n";
      }
      $s_scans = $s_scans_s + $s_scans_l;
      if ($s_congets + $s_blkgets > 0)
      {
        $rhrat = (($s_congets + $s_blkgets - $s_phreads) / ($s_congets + $s_blkgets) * 100); 
      }
      else
      {
        $rhrat = ".";
      }
      $rhrat = "-1" if $rhrat < 1;

      $gets = $s_congets + $s_blkgets;
      ##print "cg=$s_congets bg=$s_blkgets \n";;

      #$s_commits  = &scSub($sid,4);
      #$s_prefetch = &scSub($sid,84);
      #$s_blkgets = &scSub($sid,38);
      #$s_phreads = &scSub($sid,40);
      #$s_sorts_m = &scSub($sid,180);
      #$s_sorts_d = &scSub($sid,181);
      #$s_scans_s = &scSub($sid,150);
      #$s_scans_l = &scSub($sid,151);
      #$s_scans = $s_scans_s + $s_scans_l;

      printf "%-9s %-24s %-10s %-12s %6.1f %6.0f %6.0f %6.0f %6.0f %9.0f %9.0f %9.0f %9.0f %6.1f %9.0f\n",
        $sesser, $comb, $username, $program, $sdur, $sesscpu, $s_commits, $s_prefetch, $s_scans, $s_sorts_m, $s_sorts_d, $gets, $s_phreads, $rhrat, $s_locktime;
    }
  }
  exit;
}

sub scSub
{
  $mysid  = $_[0];
  $mystat = $_[1];

  $sqlsub = "select \'X\', value 
    from  v\\\$sesstat 
    where sid = $mysid and statistic# = $mystat;";
  &oraSQL($sqlsub);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $value = $words[1];
    }
  }
  return $value;
}

sub scSub2
{
  $mysid   = $_[0];
  $mystats = $_[1];

  $sqlsub = "select \'X\', statistic#, value 
    from  v\\\$sesstat 
    where sid = $mysid and statistic# in ($mystats);";

  ##print "sqlsub=$sqlsub\n";

  &oraSQL($sqlsub);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $stat  = $words[1];
      $value = $words[2];
      push(@scArray,"$stat $value");
    }
  }
  return;
}

# --------------------------------------------------------------------
# -sd -se
# --------------------------------------------------------------------
sub sd_or_se_task
{
  my( $package, $file, $line ) = caller;
  unless ( $arg eq "-sd" or $arg eq "-se" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";

  $jobcnt = 0;
  $sesscnt = 0;

  sub seTitle 
  {
    print "SID-Ser#   Sess. Type Status   OS:Userid-PID-Server       SPID  DB Userid        Session Start Time  Program\n";
    print "---------- ---------- -------- -------------------------- ----- ---------------- ------------------- --------------------\n";
  }

  &seTitle() if $arg eq "-se";

  $sqlstmt = '
    select distinct \'Y\', sid from v\$mystat;
    select \'X\', 
      s.sid, 
      s.serial#, 
      nvl(p.spid,-1),
      s.type,
      s.osuser, 
      to_char(logon_time,\'YYYY/MM/DD-HH24:MI:SS\') logon_time, 
      s.status,
      s.process, 
      s.machine,
      s.username,
      s.program
	from v\$session s, v\$process p
	where osuser is not null and
	p.addr = s.paddr order by 2;
    select /*+ rule */ \'Z\', 
      s.sid, 
      s.serial#, 
      s.type,
      to_char(logon_time,\'YYYY/MM/DD-HH24:MI:SS\') logon_time, 
      s.status,
      s.username, 
      j.job,
      o.what
	from 
          v\$session s, 
          dba_jobs_running j,
          dba_jobs o
	where 
            j.sid = s.sid
        and j.job = o.job  order by 2;
  ';

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();

    $verb    = $words[0];

    if ($verb eq "Y")
    {
      $mysid = $words[1];
    } 

    if ($verb eq "Z")
    {
      $sid     = $words[1];
      $ser     = $words[2];
      $type    = "JOB";
      $logtime = $words[4];
      $status  = $words[5];
      $dbuser  = $words[6];
      $jobnum  = "JobNum=$words[7]";
      $jobname = "JobName=$words[8]";

      $sidser = "$sid,$ser";
      $jobinfo = "$jobname $jobnum";

      $jobcnt++;

      ##print "dbuser=$dbuser\n";

      printf "%-10s %-10s %-8s %-26s %-5s %-16s %-19s %-20s\n",
        $sidser,$type,$status,"","",$dbuser,$logtime,$jobinfo;
    }

    if ($verb eq "X")
    {
      $sid     = $words[1];
      $ser     = $words[2];
      $shadow  = $words[3];
      $type    = $words[4];
      $user    = $words[5];
      $logtime = $words[6];
      $status  = $words[7];
      $process = $words[8];
      $machine = $words[9];
      $dbuser  = $words[10];
      $program = $words[11] . $words[12] . $words[13] . $words[14];
      $comb = "$user-$process-$machine";
      $comb = substr($comb,0,26) if length($comb) > 26;
      $sidser = "$sid,$ser";

      if ($arg eq "-sd")
      {
        print "==========================================================================================================\n";
        &seTitle;
      }
      $sesscnt++;
      $sescount{$status}++ if $type =~ /USER/;
      if ($user eq "") { $user = "."; }
      if ($sid == $mysid) { $type = "USER*"; }
      ###print "user=$user\n";
      printf "%-10s %-10s %-8s %-26s %-5s %-16s %-19s %-20s\n",
        $sidser,$type,$status,$comb,$shadow,$dbuser,$logtime,$program;

      if ($arg eq "-sd" && $type ne "BACKGROUND")
      {
        print "\n$P | SQL For Session $sid \n";
        &srSub($sid);
        print "\n$P | Waits For Session $sid \n\n";
        &swSub($sid);
        print "==========================================================================================================\n\n";
      }
    }
  }
  print "\n";
  @sorted = sort keys %sescount;
  foreach (@sorted)
  {
    &msg("Count of USER session(s) in $_ status: $sescount{$_}");
  }
  &msg("There are $sesscnt total sessions (mysid=$mysid - denoted by *)");
  &msg("There are $jobcnt total jobs running");
  exit;
}

# --------------------------------------------------------------------
# -sg
# --------------------------------------------------------------------
sub sg_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sg" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";
  $sqlstmt = "
    select 'X', pool, name, bytes from v\$sgastat order by 2;
    select 'Y', value from v\$parameter where name = 'shared_pool_size';";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    $all    = $_; 
    @words  = split();
    $verb   = $words[0];
    $pool1  = $words[1];
    $pool2  = $words[2];
    $word1  = $words[3];
    $word2  = $words[4];
    $word3  = $words[5];
    #$word4  = $words[6];
    $wordlast = $words[$#words];

    $ndx    = index($_,"()");
    $desc   = substr($_,$ndx+3,40);
    if ($verb eq "Y")
    {
      $shared_pool_size = $words[1];
      if ($shared_pool_size =~ /M$/)
      {
        chop($shared_pool_size);
        $shared_pool_size *= 2**20;
      }
    }

    if ($verb eq "X")
    {
      if ($pool1 eq "shared" && $pool2 eq "pool" && $word1 eq "free" && $word2 eq "memory")
      {
        $freemem = $word3;
      }
      elsif ($pool1 eq "large" && $pool2 eq "pool" && $word1 eq "free" && $word2 eq "memory")
      {
        $freememlarge = $word3;
      }
      elsif ($pool1 eq "java" && $pool2 eq "pool" && $word1 eq "free" && $word2 eq "memory")
      {
        $freememjava = $word3;
      }
      elsif ($pool1 eq "shared" && $pool2 eq "pool" && $word1 eq "sql" && $word2 eq "area")
      {
        $sqlarea = $word3;
      }
      elsif ($pool1 eq "shared" && $pool2 eq "pool" && $word1 eq "library" && $word2 eq "cache")
      {
        $libcache = $word3;
      }
      elsif ($all =~ /db_block_buffers/) { $dbblock = $wordlast; }
      elsif ($all =~ /buffer_cache/) { $dbblock = $wordlast; }
      elsif ($all =~ /log_buffer/) { $logbuf = $wordlast; }
      elsif ($all =~ /fixed_sga/) { $fixedsize = $wordlast; }
      elsif ($pool1 eq "shared" && $pool2 eq "pool")
      {
        $sharedtot += $wordlast;
      }
    }
  }

  $total = $sharedtot + $dbblock + $logbuf + $fixedsize + $sqlarea + $libcache + $freememlarge + $freememjava;
  $freesharedpct = 100 - ($freemem / $shared_pool_size * 100);

  ##print "sps=$shared_pool_size\n";

  print  "SGA Pool                 Size(Bytes)   Size(MB)\n";
  print  "-------------------- --------------- ----------\n";

  printf "Shared Pool (Misc.)   %14s  %9.2f\n",$sharedtot,$sharedtot/1000000;
  printf "SQL Area              %14s  %9.2f\n",$sqlarea,$sqlarea/1000000;
  printf "Library Cache         %14s  %9.2f\n",$libcache,$libcache/1000000;
  printf "DB Block Buffer       %14s  %9.2f\n",$dbblock,$dbblock/1000000;
  printf "Redo Log Buffer       %14s  %9.2f\n",$logbuf,$logbuf/1000000;
  printf "Fixed SGA             %14s  %9.2f\n",$fixedsize,$fixedsize/1000000;
  printf "Large Pool Free       %14s  %9.2f   \n",$freememlarge,$freememlarge/1000000;
  printf "Java Pool Free        %14s  %9.2f   \n",$freememjava,$freememjava/1000000;

  print  "==================== =============== ==========\n";
  printf "SGA Total Used        %14s  %9.2f\n",$total,$total/1000000;
  printf "Shared Pool Free      %14s  %9.2f\n",$freemem,$freemem/1000000;
  print  "==================== =============== ==========\n";
  printf "Total Shared Memory   %14s  %9.2f\n",$total+$freemem,($total+$freemem)/1000000;

  print "\n";
  printf "shared_pool_size      %14s \n",$shared_pool_size;
  printf "Shared Pool Free      %14d \n",$freemem;
  printf "Shared Pool %% Full         %9.2f%% \n",$freesharedpct;

  exit;
}

# --------------------------------------------------------------------
# -sh
# --------------------------------------------------------------------
sub sh_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sh" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
  select \'Y\',
    name, value 
    from v\$parameter 
    where 
    name in (\'db_block_size\',\'db_block_buffers\');
  select
    \'T\', sum(reloads) / sum(pins)  from v\$librarycache;
  select
    \'V\', sum(gethits) / sum(gets)  from v\$librarycache;
  select
    \'U\', sum(getmisses) / sum(gets) from v\$rowcache;
  ';

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "T")
    {
      $pin_miss_ratio = $words[1] * 100;
      $pin_miss_ratio = sprintf("%6.2f",$pin_miss_ratio);
      if ($pin_miss_ratio >= 1)
      {
        $rec .= " (LC) Since the Library Cache reloads/pins ratio of $pin_miss_ratio exceeds 1%, you should increase the init.ora shared_pool_size parameter.\n";
      }
    }

    if ($verb eq "V")
    {
      $lc_hit_ratio = $words[1] * 100;
      $lc_hit_ratio = sprintf("%6.2f",$lc_hit_ratio);
      if ($lc_hit_ratio < 90)
      {
        $rec .= " (LC) Since the Library Cache hit ratio $lc_hit_ratio is < 90%, you should increase the init.ora shared_pool_size parameter.\n";
      }
    }

    if ($verb eq "U")
    {
      $dc_miss_ratio = $words[1] * 100;
      $dc_miss_ratio = sprintf("%6.2f",$dc_miss_ratio);
      if ($dc_miss_ratio >= 15)
      {
        $rec .= " (DC) Since the Dictionary Cache hit ratio $dc_miss_ratio is >= 15%%, you should increase the init.ora shared_pool_size parameter.\n";
      }
    }

    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        #$db_block_size = $words[2];
      }
      if ($words[1] eq "db_block_buffers")
      {
        #$db_block_buffers = $words[2];
      }
    }

  }
  print "\n";
  printf "Library Cache Reloads/Pins Ratio: %6s%% (S/B < 1%)\n",$pin_miss_ratio;
  printf "Library Cache Hit Ratio:          %6s%% (S/B >= 90%)\n",$lc_hit_ratio;
  printf "Dictionary Cache Miss Ratio:      %6s%% (S/B < 15%)\n",$dc_miss_ratio;

  print "\n";
  print "Recommendations:\n";
  print $rec;
  print "(None)" if $rec eq "";
  print "\n";
  exit;
}

# --------------------------------------------------------------------
# -sl
# --------------------------------------------------------------------
sub sl_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sl" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($arg2 ne "")
  {
    &oraSQL("select * from $arg2;");
    foreach (@oraSQLout)
    {
      print "$_";
    }
  }
  else
  {
    &msg("USAGE: $P -sl OWNER.TABLE_NAME");
    exit 1;
  }
  exit;
}

# --------------------------------------------------------------------
# -sn
# --------------------------------------------------------------------
sub sn_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sn" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select \'X\',value from v\$parameter where name='user_dump_dest'";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb = @words[0];
    if ($verb eq "X")
    {
      $user_dump_dest = $words[1];
    }
  }
  print "     \"$user_dump_dest\"\n";
  print "     This will take from 60 to 90 seconds.\n";
  print "     Executing...\n";
  if (substr($oraver,0,3) ge "8.1")
  {
    $maxsize = "alter session set max_dump_file_size=unlimited;";
  }
  else
  {
    $maxsize = "";
  }
  $sqlstmt = "$maxsize
              alter session set events 'immediate trace name systemstate level 10';
              !echo \"$P Sleeping 60...\"
              !sleep 60
              alter session set events 'immediate trace name systemstate level 10';
              !echo \"$P Sleeping 60...\"
              !sleep 60
              alter session set events 'immediate trace name systemstate level 10';";
  &oraSQLold ($sqlstmt,$os_conn);
  foreach (@oraSQLout)
  {
  }
  &msg ("$P Done");

  exit;
}

# --------------------------------------------------------------------
# -so 
# --------------------------------------------------------------------
sub so_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-so" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
	   vss.value,
	   substr(vsn.name,1,20) "Type of Sort",
	   substr(vs.osuser,1,20) "os user",
    substr(vs.username,1,20) "db user"
    from   v\$session vs,
	   v\$sesstat vss,
	   v\$statname vsn
    where (vss.statistic#=vsn.statistic#) AND
	  (vs.sid=vss.sid) AND
	  (vsn.name like \'%sort%\')
	  AND vss.value <> 0
    order by 2,3;';
  
  print "DB User    OS User          Sort Type            Value\n";
  print "---------- ---------------- --------------- ----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words      = split();
    $verb       = $words[0];
    $value      = $words[1];
    $s_type1    = $words[2];
    $s_type2    = $words[3];
    $osuser     = $words[4];
    $dbuser     = $words[5];

    $s_type="$s_type1 $s_type2";
    $_=$dbuser;
    s/ //g;
    $dbuser=$_;
    $_=$osuser;
    s/ //g;
    $osuser=$_;
    if ($verb eq "X")
    {
      if ($dbuser eq "") { $dbuser = "."; }
      if ($osuser eq "") { $osuser = "."; }
      printf "%-10s %-16s %-15s %10s\n",
        $dbuser,$osuser,$s_type,$value;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -sp
# --------------------------------------------------------------------
sub sp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
       exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select 'X',
    to_char(snap_time, 'YYYY/MM/DD-HH24:MI:SS'),
    snap_id,
    snap_level,
    sysdate - snap_time 
    from perfstat.stats\$snapshot
    order by 2;";

  print " Snap ID Snap Timestamp      Snap Lvl Snap Age (Days)\n";
  print "-------- ------------------- -------- ---------------\n";

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $ts  = $words[1];
      $num = $words[2];
      $lvl = $words[3];
      $age = $words[4];
      $age = sprintf("%5.3f",$age);
      printf "%8s %18s %8s %15.3f\n", $num, $ts, $lvl, $age;
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -sq
# --------------------------------------------------------------------
sub sq_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sq" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
       exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  shift @ARGV;
  $sql = join( ' ', @ARGV );
  $sql =~ s/^"//; $sql =~ s/"$//;

  if ($sql eq "")
  {
     print "USAGE: $P -sq \"<sql_statement>\"\n";
     exit 1;
  }
  else
  {
     $sql .= ";";
     &msg("$P | Executing SQL Statement: $sql");
  }

  &oraSQL("$sql");
  foreach (@oraSQLout)
  {
     print "$_";
  }
  exit;
}

# --------------------------------------------------------------------
# -sr
# --------------------------------------------------------------------
sub sr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  if ($arg2 eq "" )
  {
    print "$P | Usage: $P -sr SESSION_NUMBER\n";
    exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " $arg2");
  &msg("SQL Statement:");
  $z = "-" x 80;
  print "$z \n";

  &srSub($arg2);

  $z = "-" x 80;
  print "$z \n";
  exit; 
}

sub srSub
{
  $sessionid = $_[0];

  $sql = "select 
  'Z',  
  sql_text 
from
  --v\$sqltext_with_newlines 
  v\$sqltext
where
  address =
  (select sql_address from v\$session where sid = $sessionid)
order by
  piece;
";
  #print "sql=$sql \n";

  &oraSQL("$sql");
  $sql  = "";
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = shift(@words);
    if ($verb eq "Z")
    {
      foreach (@words) 
      { 
        ##print "inword >$_< \n";
        $sql .= $_ . " "; 
        $inword = $_;
      }
      chop $sql;
      # In some cases, pad with a blank to prevent the last word being joined with the first word of the next line
      $z = uc $inword;
      ##print "z=$z \n";
      if ($z eq "FROM" || $z eq "AND")
      {
        $sql .= " ";
      }
      ##print "sql    >$sql< \n";
    }
  }
  # Massage
  $_ = $sql;
  s/\n//g;
  $sql = $_;

  $_ = $sql;
  $sqlout = "";
  @sqlwords = split;
  foreach (@sqlwords)
  {
    $word = $_;
    ##print "word=$word \n";
    $worduc = uc $word;
    $newlineflag = 0;
    $newlineflag = 1 if $worduc eq "SELECT" || $worduc eq "FROM" || $worduc eq "ORDER" || $worduc eq "WHERE" || $worduc eq "GROUP" || $worduc eq "FROM"; 
    $sqlout .= "\n" if $newlineflag;
    $sqlout .= $word;
    $sqlout .= "\n " if $newlineflag;
    $sqlout .= " ";
  }

  print "$sqlout";
  print "\n";
}

# --------------------------------------------------------------------
# -ss
# --------------------------------------------------------------------
sub ss_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ss" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
         tablespace_name,
         initial_extent,
         next_extent,
         min_extents,
         max_extents,
         pct_increase
    from dba_tablespaces
    order by tablespace_name;';

  print "\n";
  print "Tablespace              Initial       Next        Min         Max Pct Increase\n";
  print "-------------------- ---------- ---------- ---------- ----------- ------------\n";
  
  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $found = 1;
      $ts = $words[1];
      $initial = $words[2];
      $next = $words[3];
      $min = $words[4];
      $max = $words[5];
      $pct = $words[6];
      printf("%-20s %10d %10d %10d %11d %12d\n",$ts,$initial,$next,$min,$max,$pct);
    }
  }

  if (!$found)
  {
    &msg("Something is wrong - query returned no rows");
    exit 1;
  }
  exit;
}

# --------------------------------------------------------------------
# -st
# --------------------------------------------------------------------
sub st_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-st" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "select 'X', class, value, name from v\$sysstat order by 2 desc, 4;";
  print "\n";
  print "Class Statistic Name                                               Value\n";
  print "----- -------------------------------------------------- ---------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split();
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $class = $words[1];
      $value = $words[2];
      $namet = $words[3];
      ##print "z=$z \n";
      chop($name = substr($_,index($_,$namet),50));

      if ($name =~ /table scans \(short tables/) { $scans_short = $value; }
      if ($name =~ /table scans \(long tables/)  { $scans_long  = $value; }
      if ($name =~ /sorts \(disk/)               { $sorts_disk  = $value; }
      if ($name =~ /sorts \(memory/)             { $sorts_mem   = $value; }
     
      printf "%5d %-50s %15s\n",$class,$name,$value;
    }
  }

  print "\nCalculated Statistics:\n";

  $scans_total = $scans_short + $scans_long;
  $scans_pct   = ($scans_long / $scans_total) * 100;
  $scans_pct   = sprintf("%8.5f",$scans_pct);
  printf "Long Table Scans / Total Table Scans:  %9s%%\n",$scans_pct;

  $sorts_total = $sorts_disk + $sorts_mem;
  $sorts_pct   = ($sorts_disk / $sorts_total) * 100;
  $sorts_pct   = sprintf("%8.5f",$sorts_pct);
  printf "Disk Sorts / Total Sorts:              %9s%%\n",$sorts_pct;

  exit;
}

# --------------------------------------------------------------------
# -su
# --------------------------------------------------------------------
sub su_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-su" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  ( $numcpus, $cpubusy, $cpuidle ) = &cpustats();
  &msg("#CPUs($numcpus) Busy: $cpubusy%");

  exit;
}

# --------------------------------------------------------------------
# sw
# --------------------------------------------------------------------
sub sw_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sw" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
     exit 1;
  }

  sub swTitle
  {
    print "SID    DB-User          WaitType St WTSeq  Tmr WTTimSec Wait Event Details\n";
    print "------ ---------------- -------- -- ------ --- -------- -------------------------------------\n";
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";
  &swTitle;
  &swSub;
  exit;
}

sub swSub
{
  $sqlstmt = "select 'X', 
    sw.sid, 
    sw.seq#,  
    nvl(sw.wait_time, 0), 
    nvl(sw.seconds_in_wait,0),
    nvl(se.username,'.'),
    nvl(se.status,'.'),
    sw.p1raw, 
    sw.p2raw, 
    sw.p3raw,
    sw.event, 
    sw.state, 
    sw.p1text, 
    sw.p1, 
    sw.p2text, 
    sw.p2, 
    sw.p3text, 
    sw.p3
    from 
      v\$session_wait sw,
      v\$session      se
    where 
      sw.sid = se.sid
    order by 2;";
  &oraSQL($sqlstmt);
  @saveSQLout = @oraSQLout;

  $wait_idle_count = 0;
  $wait_real_count = 0;
  $total_sessions  = 0;

  foreach (@saveSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    $sidno   = $words[1];
    $seq     = $words[2];
    $waittime= $words[3];
    $sectimer= $words[4];
    $se_user = $words[5];
    $se_stat = substr($words[6],0,1);
    $p1raw   = $words[7];
    $p2raw   = $words[8];
    $p3raw   = $words[9];

    ##print "words=$#words\n";

    $rawtext = "P1Raw=$p1raw P2Raw=$p2raw P3Raw=$p3raw ";
    $text = $rawtext;

    for ($i = 10; $i <= $#words; $i++)
    {
      $text .= $words[$i];
      $text .= " ";
    }

    if ($verb eq "X")
    {
      $total_sessions++;

      $wait_type = "NoWait";

      if ($sectimer > 0)
      {
        if (
             $text =~ /SQL\*Net/ ||
             $text =~ /slave wait/ ||
             $text =~ /pmon timer/ ||
             $text =~ /smon timer/ ||
             $text =~ /rdbms ipc message/ 
           )
        {
          $wait_type = "IdleWait"; 
          $wait_idle_count++; 
        }
        else
        {
          $wait_type = "RealWait"; 
          $wait_real_count++; 
        }
      }

      ##print "sidno=$sidno sid=$sid \n";
      next if $sidno ne $sid && $arg eq "-sd";
      &swTitle if $arg eq "-sd";
      if ($user eq "") { $user = "."; }
      ##print "sectimer=$sectimer\n";
      printf "%-6d %-16s %-8s %2s %6d %3d %8d %-40s\n",
        $sidno,$se_user,$wait_type,$se_stat,$seq,$waittime,$sectimer,$text;

      #---- FUTURE
##	SELECT
##	  SID,
##	  USERNAME,
##	  TERMINAL,
##	  PROGRAM
##	FROM
##	  V$SESSION
##	WHERE
##	  SADDR in
##	  (
##	    SELECT
##	      KGLLKSES
##	    FROM
##	      X$KGLLK LOCK_A
##	    WHERE KGLLKREQ = 0
##	      AND EXISTS
##	    (SELECT LOCK_B.KGLLKHDL FROM X$KGLLK LOCK_B
##	       WHERE KGLLKSES = 'C000000089485210' /* BLOCKED SESSION */
##	       AND LOCK_A.KGLLKHDL = LOCK_B.KGLLKHDL
##	       AND KGLLKREQ > 0)   );

      #---- Analyze selected waits
      if ($text =~ /library cache pin/ && $p1raw)
      {
        print "       -------------------------------------------------------------------\n"; 
        print "       Wait analysis for [library cache pin] wait:\n"; 

        &oraSQL("
          SELECT
            'Z',
            kglnaown \"Owner\", 
            kglnaobj \"Object\"
          FROM 
            x\$kglob
          WHERE
            kglhdadr='$p1raw';
        ");
        @save2SQLout = @oraSQLout;
        foreach (@save2SQLout)
        {
	  @words2   = split();
	  $verb2    = $words2[0];
          if ($verb2 eq "Z")
          {
            $diag_obj_owner = $words2[1];
            $diag_obj_name  = $words2[2];

            print "         Object Being Latched: $diag_obj_owner.$diag_obj_name\n";
          }
        }

        &oraSQL("
          SELECT
            'Z',
            s.sid, 
            kglpnmod \"Mode\", 
            kglpnreq \"Req\"
          FROM 
            x\$kglpn p, 
            v\$session s
          WHERE
            p.kglpnuse=s.saddr
            AND kglpnhdl='$p1raw';
        ");
        @save2SQLout = @oraSQLout;
        foreach (@save2SQLout)
        {
	  @words2   = split();
	  $verb2    = $words2[0];
          if ($verb2 eq "Z")
          {
            $diag_sid  = $words2[1];
            $diag_mode = $words2[2];
            $diag_req  = $words2[3];

            if ($diag_mode == 0) { $diag_mode_long = "WAITING"; }  
            if ($diag_mode == 2) { $diag_mode_long = "SHARED"; }  
            if ($diag_mode == 3) { $diag_mode_long = "EXCLUSIVE"; }  
            if ($diag_req  == 0) { $diag_req_long  = "NOT_WAITING"; }  
            if ($diag_req  == 2) { $diag_req_long  = "SHARED"; }  
            if ($diag_req  == 3) { $diag_req_long  = "EXCLUSIVE"; }  

            print "         - Session $diag_sid is holding this resource in mode $diag_mode_long - Requesting mode $diag_req_long\n";
          }
        }

        print "       -------------------------------------------------------------------\n"; 
      }
    }
  }
  print "\n";
  &msg("Total Sessions=$total_sessions - Idle Waits=$wait_idle_count - Real Waits=$wait_real_count");
}

# --------------------------------------------------------------------
# -sy
# --------------------------------------------------------------------
sub sy_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-sy" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }
  
  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $rows = 0;

  $sqlstmt = "
    SELECT 
      'X',
      grantee,
      translate(privilege,' ','_'),
      admin_option
    FROM 
      dba_sys_privs 
    order by
      2, 3;";

  print "\n";
  print "Grantee                  System Privilege                         Adm\n";
  print "------------------------ ---------------------------------------- ---\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $grantee = shift @words;
      $priv    = shift @words;
      $adm     = shift @words;

      printf "%-24s %-40s %-3s\n",$grantee,$priv,$adm;
      $rows++;
    }
  }
  print "\n";
  &msg("$rows Rows found");
  exit;
}

# --------------------------------------------------------------------
# -ta
# --------------------------------------------------------------------
sub ta_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ta" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $tablekey = '{TABLE}' if $arg2 eq '-key'; 

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";
  print "Owner.Table                               TableSpace   Blocks     MB Xtents  MaxXtnt\n";
  print "----------------------------------------- ------------ ------ ------ ------ --------\n";
  $sqlstmt = 'select \'Y\',
    name, value
    from v\$parameter
    where
    name in (\'db_block_size\',\'db_block_buffers\');
  select \'X\', owner, segment_name, tablespace_name, blocks, extents, max_extents 
    from dba_segments where segment_type = \'TABLE\'
    order by 2,3;';
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        $bs = $words[2];
        ##print "bs=$bs\n";
      }
    }
    if ($verb eq "X")
    {
      $owner = $words[1];
      $table = $words[2];
      $ts    = $words[3];
      $blks  = $words[4];
      $exts  = $words[5];
      $maxx  = $words[6];
      $mb   = ($blks * $bs) / (1024 * 1024);

      if ($maxx == 2147483645)
      {
        $maxx = "(Unlim)";
      }

      $t = $owner . "." . $table;
      #if (substr($owner,0,3) ne "SYS")
      if (($owner eq "SYS" || $owner eq "SYSTEM") && $arg2 ne "ALL") { next; } 

      printf("%-41s %-12s %6d %6d %6d %8s $tablekey\n",$t,$ts,$blks,$mb,$exts,$maxx);
      $found++;
      $tot_mb += $mb;
      $tot_xt += $exts;
    }
  }
  print  "========================================= ============ ====== ====== ====== ========\n";
  printf "(%6s Total Tables)                                         %6d %6d\n\n",$found,$tot_mb,$tot_xt;
  exit;
}

# --------------------------------------------------------------------
# -tb
# --------------------------------------------------------------------
sub tb_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tb" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "Schema             Table Count\n";
  print "------------------ -----------\n";
  $sqlstmt = '
    select 
       \'Z\', 
       owner,
       count(*) 
       from dba_tables 
       group by owner;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Z")
    {
      $owner = $words[1];
      $tbls  = $words[2];
      printf("%-20s %9d\n",$owner,$tbls);
      $total += $tbls;
    }
  }
  print "================== ===========\n";
  printf("Total:               %9d\n",$total);             
  exit;
}

# --------------------------------------------------------------------
# -tc
# --------------------------------------------------------------------
sub tc_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tc" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
  select \'X\', owner, segment_name, tablespace_name, blocks, extents, max_extents 
    from dba_segments where segment_type = \'TABLE\'
    order by 2,3;';
  &oraSQL($sqlstmt);
  $count = 0;
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];

    if ($verb eq "X")
    {
      $owner = $words[1];
      $table = $words[2];
      $obj   = "$owner.$table";

      $outdata .= "select \'$obj\', count(*), to_char(sysdate,\'YYYY/MM/DD-HH24:MI:SS\') from $obj ;\n";
      $count++;
    }
  }

  open(OUTDATA, ">$tmpfile") || die "Cannot open output $tmpfile - $!";
  print OUTDATA "$outdata"; 
  close(OUTDATA);

  &msg("Wrote $count SQL statements to $tmpfile");
  exit;
}

# --------------------------------------------------------------------
# -td
# --------------------------------------------------------------------
sub td_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-td" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  if ($arg2 eq "")
  {
    print "$P | Usage: $P -td TABLE-OWNER.TABLE-NAME\n";
    exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " $arg2");

  $sqlstmt = "
set linesize 80
desc $arg2;";

  &oraSQLold($sqlstmt, $os_conn);
  foreach (@oraSQLout)
  {
    print;
  }
  exit;
}

# --------------------------------------------------------------------
# -ti
# --------------------------------------------------------------------
sub ti_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ti" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  @words = split(/\./,$arg2);
  $owner = uc $words[0];
  $table = uc $words[1];

  if ($owner eq "" || $table eq "")
  {
    print "$P | Usage: $P -ti TABLE-OWNER.TABLE-NAME\n";
    exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " $owner.$table");

  # optionally count rows
  if ($arg3 eq "-count") 
  {
    $sqlstmt = "select \'X\', count(*) from $owner.$table;";
    ##print "sqlstmt=$sqlstmt\n";
    &oraSQL($sqlstmt);
    foreach (@oraSQLout)
    {
      @words = split;
      $verb  = $words[0];
      if ($verb eq "X")
      {
	$rows_counted = $words[1];
      }
    }
  }

  $sqlstmt = "
      select \'Y\', value 
        from v\\\$parameter where name = \'db_block_size\';
      select 
      \'X\', 
      tablespace_name,
      nvl(to_char(     num_rows),\'(Null)\'),
      nvl(to_char(last_analyzed,\'YYYY/MM/DD-HH24:MI:SS\'),\'(Null)\'),
      initial_extent,
      next_extent,
      min_extents,
      max_extents,
      pct_increase,
      buffer_pool,
      nvl(to_char(blocks),\'(Null)\'),
      freelists,
      pct_free,
      ini_trans,
      max_trans,
      nvl(to_char(freelist_groups),\'(Null)\'),
      nvl(logging,\'(Null)\'),
      avg_row_len
    from 
      dba_tables 
    where 
      table_name = \'$table\'
      and owner  = \'$owner\';";
  ##print "sqlstmt=$sqlstmt\n";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = $words[0];
    if ($verb eq "Y")
    {
      $blocksize = $words[1];
    }
    if ($verb eq "X")
    {
      $tblflag = 1;
      $ts       = $words[1];
      $numrows  = $words[2];
      $lastanal = $words[3];
      $extinit  = $words[4];
      $extnext  = $words[5];
      $extmin   = $words[6];
      $extmax   = $words[7];
      $pctincr  = $words[8];
      $bufpool  = $words[9];
      $blocks   = $words[10];
      $freel    = $words[11];
      $pctfree  = $words[12];
      $initrans = $words[13];
      $maxtrans = $words[14];
      $flgrps   = $words[15];
      $logging  = $words[16];
      $avgrowl  = $words[17];

      $mb       = int(($blocks * $blocksize) / 1000000);

      print "\nTable Data (from dba_tables)\n";
      print   "----------------------------\n";
      printf "Tablespace:      %-40s\n",$ts;
      printf "RowCount:        %-19s (From ANALYZE)\n",$numrows;
      printf "LastAnalyzed:    %-19s (From ANALYZE)\n",$lastanal;
      printf "Blocks:          %-19s (From ANALYZE)\n",$blocks;
      printf "TableMegBytes:   %-19s (From ANALYZE)\n",$mb;
      printf "AvgRowLen:       %-19s (From ANALYZE)\n",$avgrowl;
      printf "ActualRows:      %-19s (From SELECT COUNT(*))\n",$rows_counted if $rows_counted ne "";
      printf "InitialExtents:  %-15s\n",$extinit;
      printf "NextExtents:     %-15s\n",$extnext;
      printf "MinExtents:      %-15s\n",$extmin;
      printf "MaxExtents:      %-15s\n",$extmax;
      printf "PercentIncr:     %-15s\n",$pctincr;
      printf "BufferPool:      %-15s\n",$bufpool;
      printf "Freelists:       %-15s\n",$freel;
      printf "FreelistGroups:  %-15s\n",$flgrps;
      printf "PctFree:         %-15s\n",$pctfree;
      printf "IniTrans:        %-15s\n",$initrans;
      printf "MaxTrans:        %-15s\n",$maxtrans;
      printf "Logging:         %-15s\n",$logging;
    }
  }

  if (!$tblflag)
  {
    &msg("Table $owner.$table does not exist");
    exit 1;
  }

  # Column data

  print "\nColumn Data (from dba_tab_cols)\n";
  print   "-------------------------------\n";
  $sqlstmt = "select 
      \'X\', 
      column_id,
      column_name,
      data_type,
      data_length,
      nvl(to_char(data_precision),\'(Null)\'),
      nvl(to_char(data_scale),\'(Null)\'),
      nvl(nullable,\'(Null)\'),
      avg_col_len
    from 
      dba_tab_columns
    where 
      table_name = \'$table\'
      and owner  = \'$owner\'
    order by 2;";
  ##print "sqlstmt=$sqlstmt\n";
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = $words[0];
    if ($verb eq "X")
    {
      #$colflag = 1;
      $colid    = $words[1];
      $colname  = $words[2];
      $coltype  = $words[3];
      $collen   = $words[4];
      $colprec  = $words[5];
      #$colscale = $words[6];
      $colnull  = $words[7];
      #$avg_len  = $words[8];
      printf "Num: %-2s Name: %-12s Type: %-12s Length: %-4s Precision: %-3s Null: %-3s \n",$colid,$colname,$coltype,$collen,$colprec,$colnull;
    }
  }

  # Index data

  $index_counter= 0;

  print "\nIndex Data (from dba_indexes and dba_ind_columns)\n";
  print   "----------------------------\n";
  $sqlstmt = "select 
      \'X\', 
      owner,
      index_name,
      tablespace_name,
      nvl(to_char(last_analyzed, \'YYYY/MM/DD-HH24:MI:SS\'),\'(Null)\'),
      initial_extent,
      next_extent,
      min_extents,
      max_extents,
      pct_increase,
      buffer_pool,
      nvl(to_char(leaf_blocks),\'(Null)\'),
      status,
      freelists,
      pct_free,
      ini_trans,
      max_trans,
      nvl(to_char(freelist_groups),\'(Null)\'),
      nvl(logging,\'(Null)\')
    from 
      dba_indexes 
    where 
          table_name = \'$table\'
      and table_owner  = \'$owner\';";
  ##print "sqlstmt=$sqlstmt\n";
  &oraSQL($sqlstmt);
  @oraSQLoutsave1 = @oraSQLout;
  $ndx_count = 0;
  foreach (@oraSQLoutsave1)
  {
    @words = split;
    $verb = $words[0];
    if ($verb eq "X")
    {
      $ndx_count++;
      $ndx_owner  = $words[1];
      $ndx_name   = $words[2];
      $ts         = $words[3];
      $lastanal   = $words[4];
      $extinit    = $words[5];
      $extnext    = $words[6];
      $extmin     = $words[7];
      $extmax     = $words[8];
      $pctincr    = $words[9];
      $bufpool    = $words[10];
      $leafblks   = $words[11];
      $ndxstatus  = $words[12];
      $freelists  = $words[13];
      $pctfree    = $words[14];
      $initrans   = $words[15];
      $maxtrans   = $words[16];
      $flgrps     = $words[17];
      $logging    = $words[18];
     
      $ndx = "$ndx_owner.$ndx_name";
      $index_counter++;
      print "Index $index_counter: $ndx\n";
      printf "  IndexStatus:     %-40s\n",$ndxstatus;
      printf "  Tablespace:      %-40s\n",$ts;
      printf "  LastAnalyzed:    %-19s (From ANALYZE)\n",$lastanal;
      printf "  LeafBlocks:      %-19s (From ANALYZE)\n",$leafblks;
      printf "  InitialExtents:  %-15s\n",$extinit;
      printf "  NextExtents:     %-15s\n",$extnext;
      printf "  MinExtents:      %-15s\n",$extmin;
      printf "  MaxExtents:      %-15s\n",$extmax;
      printf "  PercentIncr:     %-15s\n",$pctincr;
      printf "  BufferPool:      %-15s\n",$bufpool;
      printf "  Freelists:       %-15s\n",$freelists;
      printf "  FreelistGroups:  %-15s\n",$flgrps;
      printf "  PctFree:         %-15s\n",$pctfree;
      printf "  IniTrans:        %-15s\n",$initrans;
      printf "  MaxTrans:        %-15s\n",$maxtrans;
      printf "  Logging:         %-15s\n",$logging;
      $sqlstmt = "select 
	  \'X\', 
	  column_position,
	  column_name,
	  column_length,
	  descend
	from 
	  dba_ind_columns 
	where 
	  index_name = \'$ndx_name\'
	  and index_owner  = \'$ndx_owner\'
        order by 2;";
      ##print "sqlstmt=$sqlstmt\n";
      print "\n";
      &oraSQL($sqlstmt);
      @oraSQLoutsave2 = @oraSQLout;
      foreach (@oraSQLoutsave2)
      {
	@words = split;
	$verb  = $words[0];
	if ($verb eq "X")
	{
	  $colpos   = $words[1];
	  $colname  = $words[2];
	  $collen   = $words[3];
	  $coldesc  = $words[4];
	  printf "  Index Column - Number: %-4s  Name: %-15s  Length: %-4s  Descending: %-4s\n",$colpos,$colname, $collen, $coldesc;
	}
      }
    print "\n";
    }
  }
  &msg("Found $ndx_count Indexes");
  exit;
}

# --------------------------------------------------------------------
# -tj
# --------------------------------------------------------------------
sub tj_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tj" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $len = $arg2;
  $len = 22 if $len eq "";

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} . " - len=$len");
  print "\n";

  $sqlstmt = "
  WHENEVER SQLERROR CONTINUE;
  drop table orastat_tj;
  WHENEVER SQLERROR EXIT 1;
  create global temporary table orastat_tj as 
    select name, bytes from v\$datafile;
  insert into orastat_tj 
    select name, bytes from v\$datafile;
  insert into orastat_tj 
    select name, bytes from v\$tempfile;
  select 'X',
       substr(name,1,$len),
       sum(bytes)
    from
       orastat_tj
    group by 
      substr(name,1,$len)
    order by 
      2 ;";

  ##print "sqlstmt=$sqlstmt\n";

  print "Filesystem                                                   mB Alloc.\n";
  print "------------------------------------------------------------ ---------\n";

  &oraSQLold($sqlstmt,$os_conn);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $fs  = $words[1];
      $mb  = int($words[2]/1000000);
      printf "%-60s %9d\n",$fs,$mb;
      $t_mb += $mb;
    }
  }
  printf "                                                             =========\n";
  printf "(Total):                                                     %9d\n",$t_mb;
  exit;
}

# --------------------------------------------------------------------
# -tm
# --------------------------------------------------------------------
sub tm_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tm" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  $rows = 0;

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "
    SELECT 
      'X',
      owner,
      segment_name,
      count(*) 
    FROM 
      dba_extents
    where
      segment_type = 'TEMPORARY'
    GROUP BY 
      owner,
      segment_name;";


  print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "Segment Name                                  Count\n";
  print "---------------------------------------- ----------\n";

  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $owner   = shift @words;
      $segment = shift @words;
      $count   = shift @words;

      $objname = "$owner.$segment";
  
      printf "%-40s %10s \n",$objname,$count;
      $rows++;
      $total += $count;
    }
  }
  print "======================================== ==========\n";
  printf "%-40s %10s \n","Total:",$total;
  exit;
}

# --------------------------------------------------------------------
# -tp
# --------------------------------------------------------------------
sub tp_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tp" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  &msg("Non-SYS Objects (From DBA_SEGMENTS)");

  if ($arg2 eq "top")
  {
    $orderby = "6 desc";
  }
  else
  {
    $orderby = "2,3,4";
  }

  $sqlstmt = "
    select 'W',
      sum(blocks) 
      from dba_segments
      where owner not in ('SYS','SYSTEM');
    select 'Y',
      name, value
      from v\$parameter
      where
      name in ('db_block_size','db_block_buffers');
    select 'X', 
      owner,  
      segment_name, 
      translate(segment_type,' ','_'), 
      tablespace_name, 
      nvl(blocks,-1), 
      buffer_pool
    from dba_segments
      where owner not in ('SYS','SYSTEM')
    order by $orderby ;";

  ##print "sqlstmt=$sqlstmt\n";

  print "\n";
  print "Owner.Object                             ObjectType TableSpace     Buf-Pool    Blocks   MBytes %OfTot\n";
  print "---------------------------------------- ---------- -------------- -------- --------- -------- ------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];

    if ($verb eq "W")
    {
      $sumblocks = $words[1];
      ##print "sumblocks=$sumblocks\n";
    }

    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        $bs = $words[2];
        ##print "bs=$bs\n";
      }
    }

    if ($verb eq "X")
    {
      $owner = $words[1];
      $name  = $words[2];
      $type  = $words[3];
      $ts    = $words[4];
      $blks  = $words[5];
      $bp    = $words[6];

      $type = "INDXPART" if $type eq "INDEX_PARTITION";
      $type = "TABLPART" if $type eq "TABLE_PARTITION";

      $obj = $owner . "." . $name;

      $found_obj++;

      $mb = int(($bs * $blks) / 1000000);

      $total_blks += $blks;

      $pct = "";
      if ($blks > 0 && $sumblocks > 0)
      {
	$pct = ($blks / $sumblocks) * 100;
	$pct = sprintf("%4.1f",$pct);
	$total_pct += $pct;
      }

      printf("%-40s %-10s %-14s %-8s %9d %8d %6s\n",$obj,$type,$ts,$bp,$blks,$mb,$pct);
      if    ($bp eq "DEFAULT") 
      {
        $def_total += $mb; 
        $def_blks  += $blks; 
      }
      elsif ($bp eq "KEEP")    
      { 
        $kep_total += $mb; 
        $kep_blks  += $blks; 
      }
      elsif ($bp eq "RECYCLE") 
      { 
        $rec_total += $mb; 
        $rec_blks  += $blks; 
      }

      $total_mb += $mb;

      if ($arg2 eq "top" && $found_obj >= $arg3)
      {
	last;
      }
    }
  }

  print "======================================== ========== ============== ======== ========= ======== ======\n";
  printf("%-40s %-10s %-14s %-8s %9d %8d %6s\n","Total:","","","",$total_blks,$total_mb,$total_pct);

  print "\n";
  &msg("$found_obj Objects");
  print "\n";

  $total_blks = $def_blks + $kep_blks + $rec_blks;

  &msg("Total Size By Buffer Pool (Sum of Sizes of Above Tables And Indexes):");
  print "\n";
  print "Buffer Pool Total Size (mB) Total Blks \n";
  print "----------- --------------- -----------\n";
  printf("%-11s %15d %11d\n","DEFAULT", $def_total, $def_blks);
  printf("%-11s %15d %11d\n","KEEP",    $kep_total, $kep_blks);
  printf("%-11s %15d %11d\n","RECYCLE", $rec_total, $rec_blks);
  print "=========== =============== ===========\n";
  printf("%-11s %15d %11d\n","Total:", $total_mb,$total_blks);
  exit;
}

# --------------------------------------------------------------------
# tq
# --------------------------------------------------------------------
sub tq_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tq" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
      TABLESPACE_NAME           ,
      INITIAL_EXTENT            ,
      nvl(NEXT_EXTENT,\'-1\')   ,
      MIN_EXTENTS               ,
      nvl(MAX_EXTENTS,\'-1\')   , 
      nvl(PCT_INCREASE,\'-1\')  ,
      MIN_EXTLEN                ,
      STATUS                    ,
      CONTENTS                  ,
      LOGGING                   ,
      EXTENT_MANAGEMENT         ,
      ALLOCATION_TYPE           ,
      PLUGGED_IN                
    from
      dba_tablespaces 
    order by
      2;';

  &oraSQL("$sqlstmt");

  print "\n";
  print "                        ExtIni ExtNxt                          ExtMin \n";
  print "Tablespace       Status     (K)   (K)   ExtMin   MaxExt PctInc Ln (K) Contents   Logging    ExtMgmt-AllocType\n";
  print "---------------- ------ ------ ------ -------- -------- ------ ------ ---------- ---------- -----------------\n";

  @saveSQLout1 = @oraSQLout;
  foreach (@saveSQLout1)
  {
    @words   = split();
    $verb    = shift @words;
    if ($verb eq "X")
    {
      $ts       = shift @words;
      $extini   = shift @words;
      $extnex   = shift @words;
      $extmin   = shift @words;
      $extmax   = shift @words;
      $pctinc   = shift @words;
      $extminln = shift @words;
      $status   = shift @words;
      $contents = shift @words;
      $logging  = shift @words;
      $extmgmt  = shift @words;
      $alloctyp = shift @words;
      $plugged  = shift @words;

      $extmax = "(Unlim.)" if $extmax == 2147483645;
      $extmax = "(N/A)"    if $extmax == -1;

      $extini /= 1024;
      $extnex /= 1024;
      $extminln /= 1024;

      $extnex = "(N/A)"    if $extnex < 1;
      $pctinc = "(N/A)"    if $pctinc == -1;

      $extmgmt .= "-$alloctyp" if $alloctyp;

      printf("%-16s %-6s %6s %6s %8s %8s %6s %6s %-10s %-10s %-10s\n",
	$ts , 
        $status   , 
        $extini   , 
        $extnex   , 
        $extmin   , 
        $extmax   , 
        $pctinc   , 
        $extminln , 
        $contents , 
        $logging  , 
        $extmgmt  , 
        $alloctyp, 
        $plugged);
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -tr
# --------------------------------------------------------------------
sub tr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $rows = 0;

  $sqlstmt = '
    select \'Y\',
      value
      from v\$parameter
      where
      name in (\'db_block_size\');
    SELECT 
      \'X\',
      tr.addr, 
      tr.start_time,
      tr.status, 
      tr.used_ublk, 
      tr.used_urec,
      se.sid,
      se.serial#,
      rn.name,
      se.username
    FROM 
      v\$transaction tr,
      v\$session se,
      v\$rollname rn
    WHERE
      tr.xidusn = rn.usn
      and tr.ses_addr = se.saddr
    ORDER BY  
      6;';


  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "Session,Serial#  Start Timestamp   Status   UndoBlks UndoRecs   UndoMB RBS Name     DB User\n";
  print "---------------- ----------------- -------- -------- -------- -------- ------------ ------------\n";

  $total = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "Y")
    {
      $blksize = shift @words;
    }

    if ($verb eq "X")
    {
      $addr   = shift @words;
      $start  = shift @words;
      $start .= "-" . shift @words;
      $status = shift @words;
      $ublk   = shift @words;
      $urec   = shift @words;
      $sid    = shift @words;
      $ser    = shift @words;
      $rbs    = shift @words;
      $user   = shift @words;

      $undo_mb = ($ublk * $blksize) / 1000000;
      $undo_mb = sprintf("%.2f",$undo_mb);

      $sidser = "$sid,$ser";
      $addr = "";

      printf "%-16s %-17s %-8s %8s %8s %8s %-12s %-12s\n",$sidser, $start, $status, $ublk, $urec, $undo_mb, $rbs, $user;
      $rows++;
      $totalrecs += $urec;
      $totalblks += $ublk;
      $totalmb   += $undo_mb;
    }
  }
  print "================ ================= ======== ======== ======== ======== ============ ============\n";
  $totalmb = sprintf("%.2f",$totalmb);
  printf "%-16s %-17s %-8s %8s %8s %8s \n\n","Total:", "", "", $totalblks, $totalrecs, $totalmb;
  &msg("Found $rows transactions");
  
  exit;
}

# --------------------------------------------------------------------
# -ts
# --------------------------------------------------------------------
sub ts_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ts" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ($oraver_num >= 8.1)
  {
    $where = ", d.extent_management, d.allocation_type"
  }
  else
  {
    $where = "";
  }
  ##print "w=$where \n";

  $sqlstmt = "
   select 'X',
          a.tablespace_name as tablespace_name,
          d.contents,
          a.sum_bytes as bytes_avail,
          a.max_id,
          a.sum_bytes - nvl(c.sum_bytes,0) as bytes_used,
          nvl(c.sum_bytes,0) as bytes_free,
          round((a.sum_bytes - nvl(c.sum_bytes,0.00)) / a.sum_bytes * 100.0, 2) as pct_used,
          d.status $where
     from 
          ( select 
              tablespace_name,
              sum(bytes) sum_bytes,
              count(1) max_id
            from 
              dba_data_files
            group by 
              tablespace_name
          ) a,
          ( select 
              tablespace_name,
	      sum(bytes) sum_bytes
	    from 
              dba_free_space
	    group by 
              tablespace_name
          ) c,
          dba_tablespaces d 
    where 
          a.tablespace_name = c.tablespace_name (+)
      and a.tablespace_name = d.tablespace_name (+) order by 2;
    select 'Y', count(distinct tablespace_name) from dba_temp_files; ";

  print "\n";
  print "PermanentTablespace  Type Status   ExtentMgmt AllocTyp NumDF  Size(MB)  Free(MB)  Used(MB)   Pct\n";
  print "-------------------- ---- -------- ---------- -------- ----- --------- --------- --------- -----\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/ && !/^ORA-00942/;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      $found = 1;
      if ($words[1] > 0) 
      { 
        $foundtemp = 1; 
        $counttemp = $words[1];
      } 
    }
    if ($verb eq "X")
    {
      $tscount++;
      $found = 1;
      $ts = $words[1];
      $type = substr($words[2],0,4);
      $size = $words[3]/1048576;
      $max = $words[4];
      $used = $words[5]/1048576;
      $free = $words[6]/1048576;
      $pct = $words[7];
      $status = $words[8];
      $extmgmt  = $words[9];
      $alloctyp = $words[10];

      $extmgmt  = "." if $extmgmt eq "";
      $alloctyp = "." if $alloctyp eq "";

      $used = $size - $free;

      if ($arg2) { $z = "$date "; }
      else       { $z = ""; }

      printf("${z}%-20s %-4s %-8s %-10s %-8s %5d %9d %9d %9d %5.1f \n",$ts,$type,$status,$extmgmt,$alloctyp,$max,$size,$free,$used,$pct);

      $tot_size += $size;
      $tot_free += $free;
      $tot_used += $used;
      $tot_df   += $max;
    }
  }
  if (!$found)
  {
    print "$P | Something is wrong - query returned no rows\n";
    exit 1;
  }

  print "==================== ==== ======== ========== ======== ===== ========= ========= ========= =====\n";
  $ts = "Total:";
  $pct = ($tot_used / $tot_size * 100);
  printf("%-20s                                   %5d %9d %9d %9d %5.1f\n",$ts,$tot_df,$tot_size,$tot_free,$tot_used,$pct);

  print "\n";
  &msg("Found $tscount tablespace(s)\n");

  if ($foundtemp)
  {
    print "Note: This DB also has $counttemp 8i+ Temporary TS(s) listed in dba_temp_files - invoking orastat -tt\n\n";
    $arg = "-tt";            #<<<<<<<<<<<<<< MODIFYING $arg On The Fly
    &tt_task();              #<<<<<<<<<<<<<< INVOKING tt_task On The Fly
  }

  exit;
}

# --------------------------------------------------------------------
# -tt
# --------------------------------------------------------------------
sub tt_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tt" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ( $oraver_num < 8.1 )
  {
      print "\n$P | No locally managed temporary tablespaces were found.\n";
      exit;
  }
  elsif ($oraver =~ /^8.1.7/ or $oraver_num > 8.2 )
  {
    $where = ", d.extent_management, d.allocation_type"
  }
  else
  {
    $where = "";
  }
  ##print "w=$where \n";

  $tot_size = 0;
  $tot_free = 0;
  $tot_used = 0;
  $tot_df   = 0;

  $sql = "
    select \'Y\',
      value
      from v\\\$parameter
      where
      name in (\'db_block_size\');
    select \'X\',
          a.tablespace_name as tablespace_name,
          a.size_blocks, 
          nvl(a.max_id,0),
          nvl(c.used_blocks,0),
          d.status,
          d.extent_management,
          d.allocation_type
     from 
          ( select 
              tablespace_name,
              sum(blocks) as size_blocks,
              count(1) max_id
            from 
              dba_temp_files
            group by 
              tablespace_name
          ) a,
          ( select 
              tablespace_name,
	      sum(used_blocks) as used_blocks
	    from 
              v\\\$sort_segment
	    group by 
              tablespace_name
          ) c,
          dba_tablespaces d 
    where 
          a.tablespace_name = c.tablespace_name (+)
      and a.tablespace_name = d.tablespace_name (+); ";

  print "\n";
  print "TemporaryTablespace  Type Status   ExtentMgmt AllocTyp NumDF  Size(MB)  Free(MB)  Used(MB)   Pct\n";
  print "-------------------- ---- -------- ---------- -------- ----- --------- --------- --------- -----\n";

  &oraSQL($sql);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print;
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Y")
    {
      $block_size = $words[1];
    }
    if ($verb eq "X")
    {
      &error("Something is wrong - block_size is NULL") if ! $block_size;

      $found = 1;

      $ts          = $words[1];
      $size_blocks = $words[2];
      $num_df      = $words[3];
      $used_blocks = $words[4];
      $status      = $words[5];
      $extmgmt     = $words[6];
      $alloctyp    = $words[7];

      $size = ($size_blocks * $block_size) / 1048576;
      $used = ($used_blocks * $block_size) / 1048576;
      $free = $size - $used;

      if ($size == 0) 
      {
        $pct  = 0;
      }
      else
      {
        $pct  = ($used / $size) * 100;
      }

      $type = "TEMP"; 

      printf("%-20s %-4s %-8s %-10s %-8s %5d %9d %9d %9d %5.1f \n",$ts,$type,$status,$extmgmt,$alloctyp,$num_df,$size,$free,$used,$pct);

      $tot_size += $size;
      $tot_free += $free;
      $tot_used += $used;
      $tot_df   += $num_df;
    }
  }
  if (!$found)
  {
    print "$P | Something is wrong - query returned no rows\n";
    exit 1;
  }

  print "==================== ==== ======== ========== ======== ===== ========= ========= ========= =====\n";
  $ts = "Total:";
  $pct = ($tot_used / $tot_size) * 100;
  printf("%-20s                                   %5d %9d %9d %9d %5.1f\n",$ts,$tot_df,$tot_size,$tot_free,$tot_used,$pct);

  print "\n";
  &msg("NOTE: These numbers are COMPLETELY ACCURATE. They show");
  &msg("      v\$sort_usage and dba_temp_file data.");

  exit;
}

# --------------------------------------------------------------------
# -tu
# --------------------------------------------------------------------
sub tu_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tu" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "
    select 'Y',
      value
      from v\$parameter
      where
      name in ('db_block_size');
    select 
      'X', 
      sid, 
      serial#, 
      v\$session.username, 
      extents, 
      contents, 
      blocks
    from 
      v\$session, 
      v\$sort_usage 
    where 
      v\$session.saddr = v\$sort_usage.session_addr
      ;
    ";

  print "SID-Ser#     User       Extents  Blocks   MB       Contents\n";
  print "------------ ---------- -------- -------- -------- -------------\n";
  

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = $words[0];
    $sid     = $words[1];
    $ser     = $words[2];
    $user    = $words[3];
    $ex      = $words[4];
    $contents = $words[5];
    $blocks  = $words[6];
    $sidser = "$sid-$ser";

    if ($verb eq "Y")
    {
      $blocksize = $words[1];
      ##print "blocksize=$blocksize\n";
      next;
    }

    if ($verb eq "X")
    {
      if ($user eq "") { $user = "."; }
      $mb = int(($blocks * $blocksize) / 1000000);
      ###print "user=$user\n";
      printf "%-12s %-10s %8s %8s %8s %-20s\n",
        $sidser,$user,$ex,$blocks,$mb,$contents;
      $t_blocks += $blocks;
      $t_mb += $mb;
    }
  }
  
  print "\n";
  &msg("Total TEMP Blocks In Use: $t_blocks");
  &msg("Total TEMP mB In Use:     $t_mb");
  exit;
}

# --------------------------------------------------------------------
# -tz
# --------------------------------------------------------------------
sub tz_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-tz" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  if ($arg2 eq "top")
  {
    $orderby = "5 desc";
  }
  else
  {
    $orderby = "2,3";
  }

  &msg("Non-SYS Tables (From DBA_TABLES)");
  &msg("Note: Totals are only valid if ANALYZE/dbms_stats Has Been Run");

  $sqlstmt = "
    select \'W\',
      sum(blocks) 
      from dba_tables;
    select \'Y\',
      name, value
      from v\\\$parameter
      where
      name in (\'db_block_size\',\'db_block_buffers\');
    select \'X\', 
      owner,  
      table_name, 
      nvl(tablespace_name,\'?\'), 
      nvl(blocks,-1), 
      nvl(num_rows,-1), 
      nvl(buffer_pool,\'?\'), 
      nvl(to_char(last_analyzed,\'MM/DD/YYYY-HH24\'),-1)
    from dba_tables
    order by $orderby ;";

  ##print "sqlstmt=$sqlstmt\n";

  print "\n";
  print "Owner.Table                         TableSpace     Rows      Last Analyzed Buf-Pool Blocks     MBytes %OfTot\n";
  print "----------------------------------- -------------- --------- ------------- -------- -------- -------- ------\n";

  $total_objs = 0;
  $total_anal = 0;
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];

    if ($verb eq "W")
    {
      $sumblocks = $words[1];
      ##print "sumblocks=$sumblocks\n";
    }

    if ($verb eq "Y")
    {
      if ($words[1] eq "db_block_size")
      {
        $bs = $words[2];
        ##print "bs=$bs\n";
      }
    }

    if ($verb eq "X")
    {
      $owner = $words[1];
      $table = $words[2];
      $ts    = $words[3];
      $blks  = $words[4];
      $rows  = $words[5];
      $bp    = $words[6];
      $la    = $words[7];

      $t = $owner . "." . $table;
      if ($owner ne "SYS" && $owner ne "SYSTEM")
      {
        $found_tables++;

	$total_objs++; 
	$total_anal++ if $rows > -1; 

        $mb = int(($bs * $blks) / 1000000);

        $pct = "";
        if ($blks > 0 && $sumblocks > 0)
        {
	  $pct = ($blks / $sumblocks) * 100;
	  $pct = sprintf("%4.1f",$pct);
          $total_pct += $pct;
        }

        printf("%-35s %-14s %9d %13s %-8s %8d %8d %6s\n",$t,$ts,$rows,$la,$bp,$blks,$mb,$pct);
	if    ($bp eq "DEFAULT") {  $def_total += $mb; }
	elsif ($bp eq "KEEP")    {  $kep_total += $mb; }
	elsif ($bp eq "RECYCLE") {  $rec_total += $mb; }
        $total_mb += $mb;

	if ($arg2 eq "top" && $found_tables >= $arg3)
	{
	  last;
	}
      
      }
    }
  }
  print "=================================== ============== ========= ============= ======== ======== ======== ======\n";
  printf("%-35s %-14s %9s %13s %-8s %8s %8d %6s\n","Total:","","","","","",$total_mb,$total_pct);

  print "\n";
  &msg("$found_tables Tables");
  print "\n";

  &msg("Non-SYS Indexes (From DBA_INDEXES)");
  &msg("Note: Totals are only valid if ANALYZE/dbms_stats Has Been Run");

  $sqlstmt = "
    select \'W\',
      sum(leaf_blocks) 
      from dba_indexes;
    select \'X\', 
      owner,  
      index_name, 
      nvl(tablespace_name,\'?\'), 
      nvl(leaf_blocks,-1), 
      nvl(num_rows,-1), 
      nvl(buffer_pool,\'?\'), 
      nvl(to_char(last_analyzed,\'MM/DD/YYYY-HH24\'),-1)
    from dba_indexes
    order by $orderby ;";

  ##print "sqlstmt=$sqlstmt\n";

  print "\n";
  print "Owner.Index                         TableSpace     Rows      Last Analyzed Buf-Pool Blocks     MBytes %OfTot\n";
  print "----------------------------------- -------------- --------- ------------- -------- -------- -------- ------\n";
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    ##print;
    print if /^ORA/;
    @words = split;
    $verb    = $words[0];

    if ($verb eq "W")
    {
      $sumblocks = $words[1];
      ##print "sumblocks=$sumblocks\n";
    }

    if ($verb eq "X")
    {
      $owner = $words[1];
      $table = $words[2];
      $ts    = $words[3];
      $blks  = $words[4];
      $rows  = $words[5];
      $bp    = $words[6];
      $la    = $words[7];

      $t = $owner . "." . $table;
      if ($owner ne "SYS" && $owner ne "SYSTEM")
      {
        $found_indexes++;

	$total_objs++; 
	$total_anal++ if $rows != -1; 

        $mb = int(($bs * $blks) / 1000000);

        $pct = "";
        if ($blks > 0 && $sumblocks > 0)
        {
	  $pct = ($blks / $sumblocks) * 100;
	  $pct = sprintf("%4.1f",$pct);
          $total_pct += $pct;
        }

        printf("%-35s %-14s %9d %13s %-8s %8d %8d %6s\n",$t,$ts,$rows,$la,$bp,$blks,$mb,$pct);
	if    ($bp eq "DEFAULT") {  $def_total += $mb; }
	elsif ($bp eq "KEEP")    {  $kep_total += $mb; }
	elsif ($bp eq "RECYCLE") {  $rec_total += $mb; }
        $total_mb += $mb;

	if ($arg2 eq "top" && $found_indexes >= $arg3)
	{
	  last;
	}
      
      }
    }
  }
  print "=================================== ============== ========= ============= ======== ======== ======== ======\n";
  printf("%-35s %-14s %9s %13s %-8s %8s %8d %6s\n","Total:","","","","","",$total_mb,$total_pct);

  print "\n";
  &msg("$found_indexes Indexes");
  print "\n";

  $def_blks = int(($def_total*1000000) / $bs);
  $kep_blks = int(($kep_total*1000000) / $bs);
  $rec_blks = int(($rec_total*1000000) / $bs);

  $total_blks = $def_blks + $kep_blks + $rec_blks;

  &msg("Total Size By Buffer Pool (Sum of Sizes of Above Tables And Indexes):");
  print "\n";
  print "Buffer Pool Total Size (mB) Total Blks \n";
  print "----------- --------------- -----------\n";
  printf("%-11s %15d %11d\n","DEFAULT", $def_total, $def_blks);
  printf("%-11s %15d %11d\n","KEEP",    $kep_total, $kep_blks);
  printf("%-11s %15d %11d\n","RECYCLE", $rec_total, $rec_blks);
  print "=========== =============== ===========\n";
  printf("%-11s %15d %11d\n","Total:", $total_mb,$total_blks);

  print "\n";
  $total_pct = sprintf("%.2f",($total_anal/$total_objs)*100);
  &msg("Analyzed Objects: $total_anal/$total_objs (${total_pct}%)"); 

  exit;
}

# --------------------------------------------------------------------
# -ua
# --------------------------------------------------------------------
sub ua_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ua" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\', 
      username, 
      default_tablespace,
      temporary_tablespace
    from  
      dba_users 
    where 
      temporary_tablespace = \'SYSTEM\' or
      default_tablespace = \'SYSTEM\' 
      and username not in ( \'SYSTEM\', \'SYS\', \'DBSNMP \', \'OUTLN \')
    order by 2;';

  print "\n";
  print "Userid       Dflt. TS Temp TS \n"; 
  print "------------ -------- --------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $user = shift @words;
      $dflt = shift @words;
      $temp = shift @words;

      next if $user eq "DBSNMP";
      next if $user eq "OUTLN";

      if ($dflt eq "SYSTEM") { $sqltorun .= "ALTER USER $user DEFAULT   TABLESPACE USERS;\n"; }
      if ($temp eq "SYSTEM") { $sqltorun .= "ALTER USER $user TEMPORARY TABLESPACE TEMP;\n"; }

      printf "%-12s %-8s %-8s \n",$user,$dflt,$temp;
    }
  }
  
  print "\n";
  &msg("SQL To alter these users:");
  print "\n";
  print "$sqltorun\n";
  exit;
}

#--------------------------------------------------------------------
# -ub
# --------------------------------------------------------------------
sub ub_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ub" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\', owner, sum(bytes)
    from dba_segments group by owner order by 3 desc;';

  #&msg("sqlstmt=$sqlstmt");
  print "Owner                 MBytes\n";
  print "---------------- -----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $owner = $words[1];
      $bytes = $words[2];
      $mb = $bytes / 1000000;
      printf "%-16s %11.1f\n",
              $owner,$mb;
      $mb_total += $mb;
    }
  }
  print "                 ===========\n";
  printf "%-16s %11.1f\n",$null,$mb_total;
  exit;
}

# --------------------------------------------------------------------
# -ud
# --------------------------------------------------------------------
sub ud_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ud" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\', grantee from dba_role_privs where granted_role = \'ORADBA\' order by 1;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $user = $words[1];
      print "$user\n";
    }
  }
  exit;
}

# --------------------------------------------------------------------
# -ug
# --------------------------------------------------------------------
sub ug_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ug" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $date = scalar( localtime() );
  if ($arg2 ne "")
  {
    $where = "where grantee = '$arg2'";
  }
  &msg("Date=$date Where=$where");

  $sqlstmt = "
  select 
    \'X\',
    grantee,
    owner,
    table_name,
    privilege,
    grantable
  from dba_tab_privs
  ${where}
  order by 1,2,3;";

  print "Grantee                Table                                    Privilege    Grantable?\n";
  print "---------------------- ---------------------------------------- ------------ ----------\n";
  
  $count = 0;

  &oraSQL("$sqlstmt");
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $found = 1;
      $grantee = $words[1];
      $owner   = $words[2];
      $tab     = $words[3];
      $priv    = $words[4];
      $gable   = $words[5];

      $table = $owner . "." . $tab;
      $count++;

      printf("%-22s %-40s %-12s %-8s\n",$grantee,$table,$priv,$gable);
    }
  }
  print "\n";
  &msg("$count Grants found");
  exit;
}

# --------------------------------------------------------------------
# -un
# --------------------------------------------------------------------
sub un_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-un" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  if ( $oraver_num < 9 )
  {
      print "This Task Is Not Available For This Version $oraver Of Oracle\n";
      exit 1;
  }

  $rows = 0;

  $sqlstmt = '
    SELECT 
      \'X\',
      to_char(begin_time,\'YYYY/MM/DD-HH24:MI:SS\') ,
      to_char(end_time,\'YYYY/MM/DD-HH24:MI:SS\') ,
      undotsn,
      undoblks,
      txncount,
      maxconcurrency,
      ssolderrcnt,
      nospaceerrcnt
    FROM 
      v\$undostat
    where
      begin_time >= sysdate - 1
    order by
      2, 4;';


  ##print "sqlstmt=$sqlstmt\n";
  print "\n";
  print "Start Time          End Time            TS#  Blks Used  TXN Count Max Con Tx SS Old Err No Spc Err\n";
  print "------------------- ------------------- --- ---------- ---------- ---------- ---------- ----------\n";

  $spaceflag = 0;
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $begin_time = shift @words;
      $end_time   = shift @words;
      $undotsn    = shift @words;
      $undoblks   = shift @words;
      $txncount   = shift @words;
      $maxcon     = shift @words;
      $ssolerr    = shift @words;
      $nospaceerr = shift @words;

      ##print "sf=$spaceflag\n";

      $spaceflag = 1 if $nospaceerr gt "0";

      printf "%19s %19s %3s %10s %10s %10s %10s %10s \n",
	$begin_time ,
	$end_time  ,
	$undotsn  ,
	$undoblks   ,
	$txncount  ,
	$maxcon     ,
	$ssolerr    ,
	$nospaceerr;

      $rows++;
      $total += $count;
    }
  }
  print "\n";

  if ($spaceflag == 1)
  {
    &msg("Since there were >=1 'SPACE ERROR' events, you need to add space to the UNDO TS");
  }
  else
  {
    &msg("Since there were no 'SPACE ERROR' events, you do not need to add space to the UNDO TS");
  }
  
  exit;
}

# --------------------------------------------------------------------
# -us
# --------------------------------------------------------------------
sub us_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-us" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "\n";

  if ( $oraver_num < 8.0 )
  {
      $sqlstmt = "select 'X',
                         username,
                         password,
                         default_tablespace,
                         temporary_tablespace,
                         to_char( created, 'YYYY/MM/DD-HH24:MI:SS' ),
                         profile  
                  from dba_users
                  order by 2;";
  }
  else
  {
      $sqlstmt = "select 'X',
                         username,
                         password,
                         default_tablespace,
                         temporary_tablespace,
                         to_char( created, 'YYYY/MM/DD-HH24:MI:SS' ),
                         profile,
                         translate( account_status, ' ', '_' )
                  from dba_users
                  order by 2;";
  }

  print "User             Password   Profile    Dflt. TS             Temp. TS   Created             Account Status\n";
  print "---------------- ---------- ---------- -------------------- ---------- ------------------- --------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb  = $words[0];
    if ($verb eq "X")
    {
      $user    = $words[1];
      $pw      = $words[2];
      $ts      = $words[3];
      $temp    = $words[4];
      $crdate  = $words[5];
      $profile = $words[6];
      $status  = ( defined $words[ 7 ] ) ? $words[ 7 ] : ' ';
      if ($pw ne "EXTERNAL")
      {
        $pw = "(Set)";
      }

      printf("%-16s %-10s %-10s %-20s %-10s %-19s %-s\n",$user,$pw,$profile,$ts,$temp,$crdate, $status);
      $found++;
    }
  }
  print "\n";
  &msg("$found Total Users");
  exit;
}

#--------------------------------------------------------------------
# -ut
# --------------------------------------------------------------------
sub ut_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ut" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\', owner, tablespace_name, sum(bytes)
    from dba_segments group by owner, tablespace_name order by 4 desc;';

  #&msg("sqlstmt=$sqlstmt");
  print "Owner      Tablespace                MBytes\n";
  print "---------- -------------------- -----------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb    = $words[0];
    if ($verb eq "X")
    {
      $owner = $words[1];
      $ts    = $words[2];
      $bytes = $words[3];
      $mb = $bytes / 1000000;
      printf "%-10s %-20s %11.1f\n",
              $owner,$ts, $mb;
      $mb_total += $mb;
    }
  }
  print "                                ===========\n";
  printf "%-10s %-20s %11.1f\n",$null,$null,$mb_total;
  exit;
}

# --------------------------------------------------------------------
# -v 
# --------------------------------------------------------------------
sub v_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-v" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "Oracle Version: $oraver\n";
}

# --------------------------------------------------------------------
# -vs
# --------------------------------------------------------------------
sub vs_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-vs" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  print "----------------------------------\n";
  $sqlstmt = '
    select 
       \'Z\', 
       owner,
       view_name 
       from dba_views 
       order by 1, 2;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Z")
    {
      $owner = $words[1];
      $view  = $words[2];
      $name = "$owner.$view"; 
      print "$name \n";
      $total++;
    }
  }
  &msg("Total Number of Views: $total");
  exit;
}

# --------------------------------------------------------------------
# -vw
# --------------------------------------------------------------------
sub vw_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-vw" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );
  print "Schema             View Count\n";
  print "------------------ -----------\n";
  $sqlstmt = '
    select 
       \'Z\', 
       owner,
       count(*) 
       from dba_views 
       group by owner;';
  
  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words = split;
    $verb    = $words[0];
    if ($verb eq "Z")
    {
      $owner = $words[1];
      $tbls  = $words[2];
      printf("%-20s %9d\n",$owner,$tbls);
      $total += $tbls;
    }
  }
  print "================== ===========\n";
  printf("Total:               %9d\n",$total);             
  exit;
}

# --------------------------------------------------------------------
# -wr 
# --------------------------------------------------------------------
sub wr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-wr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = 'select \'X\',
      se.username,
      count(*),
      sw.event
    from
      v\$session se,
      v\$session_wait sw
    where
      sw.sid = se.sid
      and se.username > \'A\'
      and sw.event not like \'SQL*Net%\'
      and sw.event not like \'PX Deq: Table Q Normal%\'
      and sw.event not like \'PX Deq: Execution Msg%\'
      and sw.event not like \'pipe get%\'
    group by
      se.username, sw.event;
  ';
  
  print "\n";
  print "DB User    Wait Event Type                          Sessions Waiting\n";
  print "---------- ---------------------------------------- ----------------\n";

  $count_tot = 0;

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words      = split();
    $verb       = $words[0];
    if ($verb eq "X")
    {
      $user      = $words[1];
      $count     = $words[2];
      $text = "";
      for ($i = 3; $i <= $#words; $i++)
      {
	$text .= $words[$i];
	$text .= " ";
      }

      printf "%-10s %-40s %16s\n", $user,$text,$count;

      $count_tot += $count;
    }
  }
  print "\n";
  &msg("Total Session(s) Waiting: $count_tot");
  exit;
}

# --------------------------------------------------------------------
# -ws
# --------------------------------------------------------------------
sub ws_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-ws" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = '
    select \'X\', count, time, class 
    from v\$waitstat order by 2;';

  #&msg("sqlstmt=$sqlstmt");
  &msg("Wait Stats (from v\$waitstat)\n");
  print "Wait       Wait Time  Class\n";
  print "Count      1/100 Sec. Name\n";
  print "---------- ---------- --------------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    @words  = split();
    $verb   = $words[0];
    if ($verb eq "X")
    {
      $count = $words[1];
      $time  = $words[2];
      $z = index($_,"$words[3]");
      $class = substr($_,$z);
      chop $class;
      printf "%10d %10d %-s\n",$count,$time,$class;
      $tot_count += $count;
      $tot_time  += $time;
    }
  }
  print "========== ==========\n";
  printf "%10d %10d %-s\n",$tot_count,$tot_time;
  exit;
}

# --------------------------------------------------------------------
# -zr
# --------------------------------------------------------------------
sub zr_task
{
  my( $package, $file, $line ) = caller;
  if ( $arg ne "-zr" )
  {
      &msg( "Task Table Corrupt! \$package=$package, \$line=$line, \$arg=$arg" );
      exit 1;
  }

  &msg( "$arg: " . $TaskTbl{$arg}->{DESC} );

  $sqlstmt = "
    WHENEVER SQLERROR CONTINUE;
    SELECT 
      'X',
      ts, 
      prihost,
      prisid,
      sbyhost,
      sbysid,
      tool
    FROM 
      dbamon.standby_rebuild
    order by
      1, 2, 3;
      ";

  print "\n";
  print "DB Component                             Version      Status\n";
  print "---------------------------------------- ------------ ------------\n";

  &oraSQL($sqlstmt);
  foreach (@oraSQLout)
  {
    print if /^ORA/;
    ##print; 
    @words   = split();
    $verb    = shift @words;

    if ($verb eq "X")
    {
      $rows++;

      $ts      = shift @words;
      $prihost = shift @words;
      $prisid  = shift @words;
      $sbyhost = shift @words;
      $sbysid  = shift @words;
      $tool    = shift @words;

      printf "%s %-12s %-12s %-12s %-12s %s\n",$ts, $prihost, $prisid, $sbyhost, $sbysid, $tool;
    }
  }
}
